# Copyright (C) 2006-2016 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.

import errno
import logging
import os
import re
import signal
import stat
import subprocess
import threading
import time

from agent.api.base import Agent
from agent.lib.constants import LOCAL_APROC_DB, LOCAL_EXPORTS_DB
from agent.lib.daemon import close_fds, init_log, reload_xmlrpc_server
from agent.lib.db import read_item, write_item, delete_item, dump_db
from agent.lib.files import ovs_download_file
from agent.lib.notification import notify_manager, async_proc_prog
from agent.lib.password import encrypt_password, read_password, update_password
from agent.target.host import (get_file_size, get_copy_cmd, run_cmd, set_nice,
                               do_export_nfs, do_unexport_nfs)


class Host(Agent):
    """
    Class for host management.
    """

    def package_get_repositories(self):
        raise NotImplementedError()

    def package_set_repositories(self, repositories):
        raise NotImplementedError()

    def package_list(self, scope):
        raise NotImplementedError()

    def package_update(self, package=''):
        raise NotImplementedError()

    def package_uninstall(self, packages):
        raise NotImplementedError()

    def get_datetime(self):
        raise NotImplementedError()

    def set_datetime(self, year, month, day, hour, min, sec):
        raise NotImplementedError()

    def get_timezone(self):
        raise NotImplementedError()

    def set_timezone(self, timezone, utc=False):
        raise NotImplementedError()

    def get_ntp(self):
        raise NotImplementedError()

    def set_ntp(self, servers, local_time_source=False, allow_query=False):
        raise NotImplementedError()

    def enable_ntp(self):
        raise NotImplementedError()

    def disable_ntp(self):
        raise NotImplementedError()

    def get_log(self, loglist=None):
        raise NotImplementedError()

    def get_support_files(self):
        raise NotImplementedError()

    def notify_manager(self, notification, data=None):
        notify_manager(notification, data)

    def update_agent_password(self, username, old_password, new_password):
        password_dict = read_password()
        method, stored_passwd = password_dict.get(username, ('CRYPT', '!'))
        if stored_passwd != encrypt_password(old_password, method, stored_passwd):
            raise Exception('Authorization Failed')
        update_password(username, new_password)
        reload_xmlrpc_server()

    def discover_server(self):
        """
        Discover the identity, version and other properties of the server.

        @return     Discovery data as an XML document.
                    Raises an exception on failure.
        """
        raise NotImplementedError()

    def discover_hardware(self):
        """
        Discover information about the hardware present on this server.
        This includes CPU, I/O Adapters, Memory, and BIOS info.
        No information about storage devices is returned.

        @return    Returns the discovery data as an XML document.
                   Raises an exceptionon failure.
        """
        raise NotImplementedError()

    def copy_file(self, src, dst, sparse=True, update_period=20):
        """
        Copy src to dst with O_DIRECT.  Either may be a file or a block device.
        """
        set_nice()
        args = get_copy_cmd(src, dst, sparse)
        src_size = get_file_size(src)
        proc = subprocess.Popen(args, stderr=subprocess.PIPE, close_fds=True)
        dd_stat = ''

        def on_stop(sig, frame):
            try:
                proc.terminate()
            except Exception, err:
                logging.debug('Failed to terminate copy_cmd pid %d. error: %s', proc.pid, err)
            try:
                if os.path.exists(dst):
                    mode = os.stat(dst).st_mode
                    if not stat.S_ISBLK(mode):
                        os.unlink(dst)
            except Exception, err:
                logging.debug('Failed to unlink %s error: %s', dst, err)
            raise Exception('Task stopped by user. copy_file process terminated.')

        signal.signal(signal.SIGUSR2, on_stop)

        def progress_reporter():
            try:
                logging.debug('progress reporter for (%s) started',
                              ' '.join(args))
                dd_stat_re = re.compile(r'^(?P<copied>\d+) bytes? (.*) copied.*$')
                time.sleep(update_period)
                proc.poll()
                while proc.returncode is None:
                    try:
                        os.kill(proc.pid, signal.SIGUSR1)
                    except OSError, err:
                        if err[0] == errno.ESRCH:
                            logging.debug('progress_reporter: no such process')
                            break
                        else:
                            raise
                    while True:
                        dd_stat = proc.stderr.readline()
                        m = dd_stat_re.match(dd_stat)
                        if m:
                            copied = m.group('copied')
                            status = {'OPERATION': 'copy',
                                      'CURRENT_BYTES': long(copied),
                                      'TOTAL_BYTES': long(src_size),
                                      'OBJECT': 'file',
                                      'CURRENT_OBJECT_COUNT': 1,
                                      'TOTAL_OBJECT_COUNT': 1}
                            async_proc_prog(status)
                            break
                    time.sleep(update_period)
                    proc.poll()
                logging.debug('progress reporter for (%s) finished',
                              ' '.join(args))
                logging.debug('return code: %s', proc.returncode)
            except Exception, err:
                logging.debug('progress reporter for (%s) error: %s',
                              ' '.join(args), err)

        threading.Thread(target=progress_reporter).start()

        proc.wait()
        if proc.returncode != 0:
            stderr = dd_stat + proc.stderr.read(1024)
            try:
                if os.path.exists(dst):
                    mode = os.stat(dst).st_mode
                    if not stat.S_ISBLK(mode):
                        os.unlink(dst)
            except Exception, err:
                logging.debug('Failed to unlink %s error: %s', dst, err)
            raise RuntimeError('Command: (%s) failed (%s): %s' %
                               (' '.join(args), proc.returncode, stderr))
        # sync file to disk
        if os.path.exists(dst):
            fd = os.open(dst, os.O_WRONLY)
            try:
                os.fsync(fd)
            finally:
                os.close(fd)

    def ovs_download_file(self, url, filename, option=None, obj='file',
                          obj_current=1, obj_total=1, update_period=20):
        """
        Download file from URL.

        @return: Dict of function name to function mappings.
        @rtype: C{dict}

        """
        return ovs_download_file(url, filename, option, obj, obj_current,
                                 obj_total, update_period)

    def ovs_async_proc(self, func, *args):
        """
        Runs func as a daemon, add it PID to the process DB table (aproc)

        @param func    The function to run
        @param args    A list of arguments to pass to the function

        @return        If successful, returns the process ID.
                       Raises an exception on failure.
        """
        for (key, pid) in dump_db(LOCAL_APROC_DB).items():
            if not os.path.exists("/proc/%s" % pid):
                delete_item(LOCAL_APROC_DB, key)

        from agent.lib.service import get_api_function, scrub_and_log
        function = get_api_function(func)
        scrub_and_log('async call start', func, *args)

        os.chdir("/")
        os.setsid()
        os.umask(022)

        pid = os.fork()
        if pid > 0:
            write_item(LOCAL_APROC_DB, str(pid), pid)
            return pid

        close_fds()
        init_log()

        try:
            pid = os.getpid()
            status = 0
            try:
                im_self = getattr(function, 'im_self', None)
                if im_self:
                    output = function(im_self, *args)
                else:
                    output = function(*args)
            except Exception, err:
                logging.exception("async call exception")
                status = 1
                output = str(err)

            delete_item(LOCAL_APROC_DB, str(pid))

            if output is None:
                output = ''

            logging.debug('async call complete: func: %s pid: %s status: %s '
                          'output: %s', func, pid, status, output)
            try:
                notify_manager('{ASYNC_PROC} exit PID %s' % pid,
                               {'status': status, 'output': output})
            except Exception, err:
                logging.warn('Error sending notification: pid(%s) status(%s) '
                             'output(%s) error(%s)', pid, status, output, err)
            os._exit(0)
        except Exception, err:
            logging.error('Error executing %s: %s', func, err)
            os._exit(1)

    def ovs_async_proc_status(self, pid):
        """
        Gets the status of an async process.

        Returns:

        - 0: when the process is running.
        - 1: when the process has terminated.

        Raise exception on errors.

        If the pid is not in aproc DB, we think the process has terminated.

        if the process is defunct ("zombie"), we think it has terminated.
        """
        if not read_item(LOCAL_APROC_DB, str(pid)):
            return 1
        try:
            state = run_cmd(['ps', '-o', 's=', '-p', pid])
        except Exception:
            return 1
        if state.strip() == 'Z':
            return 1
        return 0

    def ovs_async_proc_stop(self, pid):
        """
        Stop a running process.

        For a process to receive SIGUSR2, a processes must register a
        signal handler. The purpose of this signal is to tell the process
        to clean up and exit vs just killing it.

        PID must exist in the process DB.
        """
        if not read_item(LOCAL_APROC_DB, str(pid)):
            logging.warn('PID %s does not exist in DB.', pid)
            return
        try:
            os.kill(int(pid), signal.SIGUSR2)
        except OSError, err:
            if err.errno == errno.ESRCH:
                logging.debug('No such process: %s', pid)
            else:
                raise

    def discover_mounted_file_systems(self, args):
        """
        Get the mounted type file systems.

        Returns an empty document if the server is unassociated.

        @param      None

        @return     Returns the discovery data as an XML document.
                    Raises an exception on failure
        """
        raise NotImplementedError()

    def export_fs(self, export_uuid, export_type, client, path, options=None):
        """
        Export the given path.

        @param      export_uuid - uuid of the export
        @param      export_type - type of the export
        @param      client - client IP address or hostname
        @param      path - mount point to be exported
        @param      options - exportfs options

        @return     None
                    Raise an exception on failure.
        """
        if export_type == 'nfs':
            # Sleep for one second if editing an existing entry for this
            # client/path pair in export file.
            d = read_item(LOCAL_EXPORTS_DB, export_uuid)
            if d:
                logging.debug('export_fs for uuid (%s) already exist '
                              'in local exports DB', export_uuid)
                time.sleep(1)
            do_export_nfs(client, path, options)
            d = {'export_uuid': export_uuid,
                 'export_type': export_type,
                 'client': client,
                 'path': path,
                 'options': options}
            write_item(LOCAL_EXPORTS_DB, export_uuid, d)
        else:
            raise NotImplementedError()

    def unexport_fs(self, export_uuid):
        """
        Unexport the given path.

        @param      export_uuid - uuid of the export

        @return     None
                    Raise an exception on failure.
        """
        d = read_item(LOCAL_EXPORTS_DB, export_uuid)
        if d:
            if d['export_type'] == 'nfs':
                do_unexport_nfs(d['client'], d['path'])
                delete_item(LOCAL_EXPORTS_DB, export_uuid)
            else:
                raise NotImplementedError()

    def get_system_disk_space(self):
        raise NotImplementedError()
