# Copyright (C) 2006-2016 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.

import os
import os.path
import re
import glob
import random
import string
import logging
import shutil
import socket
import platform
import tempfile
import xml.dom.minidom
from iniparse.compat import ConfigParser

import agent.lib.constants
from agent.api.host.base import Host
from agent.extension import OVSDiscoverExtension
from agent.lib import filelock
from agent.lib.constants import LOCAL_SERVER_DB, LOCAL_EXPORTS_DB
from agent.lib.db import dump_db
from agent.lib.discover import SimpleDiscoveryXML
from agent.lib.linux import (get_root_dev_info, get_rpm_version,
                             rpm_install, rpm_uninstall,
                             service_enable, service_disable,
                             service_start, service_stop,
                             ovn_configure, ovn_unconfigure, ovn_status,
                             OVM_SYS_REPO_PART_PREFIX, get_dev_info,
                             get_rootdev, get_bootdev, get_swaps)
from agent.lib.settings import (get_server_pool_uuid, get_owning_manager_uuid,
                                get_membership_state, get_server_certificate,
                                get_signed_server_certificate)
from agent.target.cluster import get_cluster_state
from agent.target.host import run_cmd, get_server_unique_id, get_file_size, get_ovs_info
from agent.target.hypervisor import get_hv_info
from agent.lib.stats import (get_statistic_interval, set_statistic_interval,
                             get_status_interval, set_status_interval,
                             get_fs_statistic_interval, get_fs_stats_uuids,
                             set_fs_stats_interval, set_fs_stats_uuids,
                             get_last_boot_time, set_apparent_size_interval)
from agent.target.info import CAPABILITIES
from agent.utils.dateBackend import dateBackend
from agent.utils.timezoneBackend import timezoneBackend
from agent.utils import xmlrpclib
from agent.storage.st_discover.directives import add_all_block_data
from agent.storage.st_discover.storageXML import io_host_XML

SERVER_RESERVED_MARKER = "##SERVER_RESERVED##\n"
DISCOVERY_MARKER = "##STORAGE_DISCOVERY##\n"
REPOSITORY_OPTS = {
    'name': 'name',
    'enabled': 'enabled',
    'url': 'baseurl',
    'pkg-signature': 'gpgcheck',
    'pkg-signature-key': 'gpgkey',
}
yum_conf = '/etc/yum.conf'
yum_repo_conf = '/etc/yum.repos.d/ovm.repo'
yum_lock = '/var/run/ovs-agent/yum.lock'

XML_INDENT = 2    # Indent amount for XML documents


class Linux(Host):
    """
    Linux host management.
    """

    def get_services(self, version=None):
        services = {
            'package_get_repositories': self.package_get_repositories,
            'package_set_repositories': self.package_set_repositories,
            'package_list': self.package_list,
            'package_update': self.package_update,
            'package_uninstall': self.package_uninstall,
            'get_datetime': self.get_datetime,
            'set_datetime': self.set_datetime,
            'get_timezone': self.get_timezone,
            'set_timezone': self.set_timezone,
            'get_ntp': self.get_ntp,
            'set_ntp': self.set_ntp,
            'enable_ntp': self.enable_ntp,
            'disable_ntp': self.disable_ntp,
            'get_log': self.get_log,
            'get_support_files': self.get_support_files,
            'update_agent_password': self.update_agent_password,
            'discover_server': self.discover_server,
            'discover_hardware': self.discover_hardware,
            'discover_physical_luns': self.discover_physical_luns,
            'discover_mounted_file_systems': self.discover_mounted_file_systems,
            'copy_file': self.copy_file,
            'ovs_download_file': self.ovs_download_file,
            'ovs_async_proc': self.ovs_async_proc,
            'ovs_async_proc_status': self.ovs_async_proc_status,
            'ovs_async_proc_stop': self.ovs_async_proc_stop,
            'get_last_boot_time': get_last_boot_time,
            'notify_manager': self.notify_manager,
            'set_statistic_interval': set_statistic_interval,
            'set_status_interval': set_status_interval,
            'export_fs': self.export_fs,
            'unexport_fs': self.unexport_fs,
            'set_fs_stats_interval': set_fs_stats_interval,
            'set_fs_stats_uuids': set_fs_stats_uuids,
            'get_system_disk_space': self.get_system_disk_space,
            'rpm_install': rpm_install,
            'rpm_uninstall': rpm_uninstall,
            'service_enable': service_enable,
            'service_disable': service_disable,
            'service_start': service_start,
            'service_stop': service_stop,
            'ovn_configure': ovn_configure,
            'ovn_unconfigure': ovn_unconfigure,
            'ovn_status': ovn_status,
            'set_apparent_size_interval': set_apparent_size_interval,
        }
        return services

    def package_get_repositories(self):
        import yum
        lock = filelock.LockFile(yum_conf)
        lock.acquire(wait=5)
        try:
            result = {}
            yb = yum.YumBase()
            for repo in sorted(yb.repos.findRepos('*')):
                result[repo.id] = {}
                for option, value in repo.iteritems():
                    new_option = None
                    for opt, val in REPOSITORY_OPTS.items():
                        if option == val:
                            new_option = opt
                    if not new_option:
                        continue
                    result[repo.id][new_option] = value
            return result
        finally:
            lock.release()

    def package_set_repositories(self, repositories):
        import yum
        lock = filelock.LockFile(yum_conf)
        lock.acquire(wait=5)
        try:
            repos = {}
            for section, params in repositories.items():
                new_params = {}
                for option, value in params.items():
                    if option in ['ssl-key', 'ssl-certificate']:
                        continue
                    new_option = REPOSITORY_OPTS.get(option, option)
                    new_params[new_option] = value
                repos[section] = new_params

            parser = ConfigParser()
            for section, params in repos.items():
                parser.add_section(section)
                conf = yum.config.RepoConf()
                for option, value in params.items():
                    if option not in conf.iterkeys():
                        raise RuntimeError('Unknown option: %s' % option)
                    setattr(conf, option, value)
                for name, value in conf.iteritems():
                    if name not in params or value is None:
                        continue
                    option = conf.optionobj(name)
                    parser.set(section, name, option.tostring(value))

            tmpfd, tmpfn = tempfile.mkstemp(dir='/etc/yum.repos.d')
            try:
                tmpfp = os.fdopen(tmpfd, 'w')
                parser.write(tmpfp)
                tmpfp.close()
                shutil.move(tmpfn, yum_repo_conf)
            finally:
                try:
                    os.unlink(tmpfn)
                except:
                    pass
                try:
                    os.close(tmpfd)
                except:
                    pass
        finally:
            lock.release()

    def package_list(self, scope):
        import yum
        lock = filelock.FileLock(yum_lock)
        lock.acquire(wait=120)
        try:
            if scope not in ['installed', 'updates']:
                raise Exception('Unknown scope: %s' % scope)
            yum_clean()
            yb = yum.YumBase()
            ygh = yb.doPackageLists(scope, patterns=None, showdups=None, ignore_case=False)
            return [{'name': pkg.name, 'version': pkg.version, 'release': pkg.release}
                    for pkg in getattr(ygh, scope)]
        finally:
            lock.release()

    def package_update(self, package=''):
        yum_clean()
        cmd = ['/usr/bin/yum', 'update', '-y', '-q']
        if package:
            cmd += [package]
        run_cmd(cmd)

    def package_uninstall(self, packages):
        yum_clean()
        run_cmd(['/usr/bin/yum', 'remove', '-y', '-q'] + packages)

    def get_datetime(self):
        date_backend = dateBackend()
        return tuple(date_backend.getDate())

    def set_datetime(self, year, month, day, hour, min, sec):
        date_backend = dateBackend()
        date_backend.writeDateConfig((year, month - 1, day), (hour, min, sec))
        date_backend.syncHardwareClock()

    def get_timezone(self):
        timezone_backend = timezoneBackend()
        timezone_info = timezone_backend.getTimezoneInfo()
        if timezone_info[1] == 'false':
            return timezone_info[0], False
        else:
            return timezone_info[0], True

    def set_timezone(self, timezone, utc=False):
        if utc:
            utc = 1
        else:
            utc = 0
        timezone_backend = timezoneBackend()
        timezone_backend.writeConfig(timezone, utc)

    def get_ntp(self):
        date_backend = dateBackend()
        (ntp_servers, local_time_source) = date_backend.getNtpServers()
        is_ntp_running = bool(date_backend.isNtpRunning())
        return (ntp_servers, local_time_source, is_ntp_running)

    def set_ntp(self, servers, local_time_source=False, allow_query=False):
        if len(servers) == 0 and local_time_source == False:
            raise Exception('Please specify an NTP server to use or use local time source.')
        logging.debug('substitute manager src ip %s for %s',
                      agent.lib.constants.MANAGER_SRC_IP, servers)
        servers = [re.sub('0\.0\.0\.0', agent.lib.constants.MANAGER_SRC_IP, s)
                   for s in servers]
        date_backend = dateBackend()
        broadcast_client = date_backend.getNtpBroadcastClient()
        iburst = date_backend.getIburst()
        # Workaround to keep NTP servers in order: remove all servers and rewrite again.
        if date_backend.writeNtpConfig([], broadcast_client, local_time_source, iburst) == None:
            raise Exception('Could not clear NTP servers.')
        if date_backend.writeNtpConfig(servers, broadcast_client, local_time_source, iburst) == None:
            raise Exception('Could not write NTP configurations.')
        fd = open('/etc/sysconfig/ntpd', 'r+')
        lines = fd.readlines()
        fd.seek(0)
        fd.truncate(0)
        for line in lines:
            if re.match(r'^OPTIONS=.*', line) and not re.search(r'("| )-x( |")', line):
                line = re.sub('"$', ' -x"', line)
            fd.write(line)
        fd.close()
        fd = open('/etc/ntp.conf', 'w')
        for line in date_backend.ntpFile:
            if allow_query:
                line = re.sub('^restrict default kod nomodify notrap nopeer noquery$',
                              'restrict default kod nomodify notrap nopeer', line)
                line = re.sub('^restrict -6 default kod nomodify notrap nopeer noquery$',
                              'restrict -6 default kod nomodify notrap nopeer', line)
            else:
                line = re.sub('^restrict default kod nomodify notrap nopeer$',
                              'restrict default kod nomodify notrap nopeer noquery', line)
                line = re.sub('^restrict -6 default kod nomodify notrap nopeer$',
                              'restrict -6 default kod nomodify notrap nopeer noquery', line)
            fd.write(line)
        fd.close()

    def enable_ntp(self):
        date_backend = dateBackend()
        date_backend.stopNtpService()
        ntp_servers, _ = date_backend.getNtpServers()
        if ntp_servers:
            try:
                run_cmd(['ntpdate', ntp_servers[0]])
            except Exception, err:
                logging.warn('Error to run ntpdate: %s', err)
        date_backend.startNtpService(None)
        date_backend.chkconfigOn()

    def disable_ntp(self):
        date_backend = dateBackend()
        date_backend.stopNtpService()
        date_backend.chkconfigOff()

    def get_log(self, loglist=None):
        logs = {'messages': '/var/log/messages',
                'dmesg': '/var/log/dmesg',
                'xend': '/var/log/xen/xend.log',
                'ovs-agent': '/var/log/ovs-agent.log'}
        if loglist:
            if not set(loglist) <= set(logs):
                raise Exception('log list contains items not in %s: %s' % (logs.keys(), loglist))
            logs = dict([(k, v) for (k, v) in logs.items() if k in loglist])
        result = {}
        for (logname, filename) in logs.items():
            if os.path.exists(filename):
                result[logname] = ''
                for i in (5, 4, 3, 2, 1):
                    rotate_filename = '%s.%s' % (filename, i)
                    if os.path.exists(rotate_filename):
                        fd = open(rotate_filename)
                        try:
                            result[logname] += fd.read()
                        finally:
                            fd.close()
                fd = open(filename)
                try:
                    result[logname] += fd.read()
                finally:
                    fd.close()
                result[logname] = xmlrpclib.Binary(result[logname])
        return result

    def get_support_files(self):
        name = ''.join(random.sample(string.ascii_lowercase, 6))
        for filename in glob.glob('/tmp/sosreport-%s-*.tar.xz' % name):
            os.unlink(filename)
        blacklist = 'apache,ldap,samba,systemtap'
        cmd = ['/usr/sbin/sosreport', '--batch', '--skip-plugins', blacklist,
               '--name', name]
        run_cmd(cmd)
        filenames = glob.glob('/tmp/sosreport-%s-*.tar.xz' % name)
        if not filenames:
            raise RuntimeError('No sosreport is generated.')
        filename = filenames[0]
        fd = open(filename)
        data = xmlrpclib.Binary(fd.read())
        fd.close()
        os.unlink(filename)
        os.unlink('%s.md5' % filename)
        return data

    def discover_server(self):
        """
        Discover the identity, version and other properties of the server.

        @param      None

        @return     Discovery data as an XML document.
                    Raises an exception on failure.
        """

        lock = filelock.FileLock('/var/run/ovs-agent/discover_server.lock')
        lock.acquire(wait=120)
        try:
            doc = SimpleDiscoveryXML("Discover_Server_Result")
            doc.push("Server")
            response = get_server_unique_id()
            if not response:
                raise Exception("Unable to obtain server unique id")
            doc.add("Unique_Id", response)

            doc.add("Boot_Time", str(get_last_boot_time()['last_boot_time']))

            (timezone, utc) = self.get_timezone()
            doc.push("Date_Time")
            doc.add("Time_Zone", str(timezone))
            doc.add("UTC", str(utc))
            doc.pop()

            (ntp_servers, local_time_source, is_ntp_running) = self.get_ntp()
            doc.push("NTP")
            for ntp_server in ntp_servers:
                doc.add("Server", ntp_server)
            doc.add("Local_Time_Source", str(local_time_source))
            doc.add("Is_NTP_Running", str(is_ntp_running))
            doc.pop()

            # RPM and agent versions are currently the same
            agent_version = get_rpm_version('ovs-agent')
            doc.add("Agent_Version", agent_version)
            doc.add("RPM_Version", agent_version)

            ovs_info = get_ovs_info()
            release = ovs_info['release']
            build = ovs_info['build']
            doc.add("OVM_Version", '-'.join([release, build]))

            doc.add("CPU_Type", platform.processor())
            doc.add("OS_Type", platform.system())

            release = parse_oracle_release()
            doc.add("OS_Name", release["OS_Name"])
            doc.add("OS_Major_Version", release["OS_Major_Version"])
            doc.add("OS_Minor_Version", release["OS_Minor_Version"])

            doc.add("Hypervisor_Type", "xen")
            doc.add("Hypervisor_Name", "Xen")

            # FIXME: WE NEED TO SEPARATE HOST FROM HYPERVISOR
            self.discover_xen(doc)

            uuid = get_server_pool_uuid()
            if not uuid:
                uuid = "none"
            doc.add("Pool_Unique_Id", uuid)

            uuid = get_owning_manager_uuid()
            if not uuid:
                uuid = "none"
            doc.add("Manager_Unique_Id", uuid)
            doc.add("Hostname", socket.gethostname())

            d = dump_db(LOCAL_SERVER_DB)
            registered_ip = d.get("registered_ip")
            node_number = d.get("node_number")
            server_roles = d.get("roles")
            is_master = d.get("is_master")

            manager_ip = agent.lib.constants.MANAGER_SRC_IP

            if registered_ip:
                doc.add("Registered_IP", registered_ip)

            if node_number != None:
                doc.add("Node_Number", str(node_number))

            if not server_roles:
                server_roles = "unassociated"
            else:
                server_roles = ",".join(list(server_roles))

            doc.add("Server_Roles", server_roles)

            if is_master == True:
                master_str = "true"
            else:
                master_str = "false"
            doc.add("Is_Current_Master", master_str)

            if manager_ip:
                doc.add("Manager_IP", manager_ip)

            doc.add("Membership_State", get_membership_state())
            doc.add("Cluster_State", get_cluster_state())

            doc.push("Statistic")
            doc.add("Interval", str(get_statistic_interval()))
            doc.pop()

            doc.push("Status")
            doc.add("Interval", str(get_status_interval()))
            doc.pop()

            discover_exports(doc)

            discover_server_capabilities(doc)

            doc.add('Server_Certificate', get_server_certificate())
            doc.add('Signed_Server_Certificate', get_signed_server_certificate())

            discover_file_system_statistic(doc)

            doc.pop()  # Server

            return doc.finish()
        finally:
            lock.release()

    def discover_xen(self, doc):
        kernel_items = [("release", "Host_Kernel_Release"),
                        ("version", "Host_Kernel_Version"), ]

        version_items = [("xen_major", "Major"),
                         ("xen_minor", "Minor"),
                         ("xen_extra", "Extra"), ]

        xen_items = [("xen_caps", "Capabilities"), ]

        compiler_items = [("cc_compiler", "Compiler"),
                          ("cc_compile_by", "By"),
                          ("cc_compile_domain", "Domain"),
                          ("cc_compile_date", "Date"), ]

        xeninfo = get_hv_info()

        for (xeninfo_name, discovery_name) in kernel_items:
            doc.add(discovery_name, xeninfo[xeninfo_name])

        doc.push("VMM")
        doc.push("Version")
        for (xeninfo_name, discovery_name) in version_items:
            doc.add(discovery_name, xeninfo[xeninfo_name])
        doc.pop()

        doc.push("Compile_Information")
        for (xeninfo_name, discovery_name) in compiler_items:
            doc.add(discovery_name, xeninfo[xeninfo_name])
        doc.pop()

        for (xeninfo_name, discovery_name) in xen_items:
            doc.add(discovery_name, xeninfo[xeninfo_name])

        doc.pop()

    def discover_hardware(self):
        """
        Discover information about the hardware present on this server.
        This includes CPU, I/O Adapters, Memory, BIOS info and
        storage devices.

        Other discovery is handled by invoking a python extension that
        calls legacy C code from the OVS-agent.

        @param None

        @return    Returns the discovery data as an XML document.
                   Raises an exception on failure.
        """
        lock = filelock.FileLock('/var/run/ovs-agent/discover_hardware.lock')
        lock.acquire(wait=120)
        try:
            return self._discover_linux()
        finally:
            lock.release()

    def _discover_linux(self):

        # Get the non-storage hardware, complete with headers
        (status, response, logmsg) = OVSDiscoverExtension.discover_hardware("")
        if status != 0:
            raise Exception(logmsg)
        # Find the marker for where to insert storage data
        idx = response.find(DISCOVERY_MARKER)
        marker_begin = response.rfind("\n", 0, idx)
        if idx < 0 or marker_begin < 0 or marker_begin >= idx:
            raise Exception("Internal Error: No storage discovery marker")
        indent = idx - marker_begin - 1
        whole_marker = " " * indent + DISCOVERY_MARKER
        # Get the storage data
        scsi_response = io_host_XML(indent)
        response = response.replace(whole_marker, scsi_response)
        # Find the marker for where to insert server reserved data
        idx = response.find(SERVER_RESERVED_MARKER)
        marker_begin = response.rfind("\n", 0, idx)
        if idx < 0 or marker_begin < 0 or marker_begin >= idx:
            raise Exception("Internal Error: No server reserved marker")
        indent = idx - marker_begin - 1
        whole_marker = " " * indent + SERVER_RESERVED_MARKER
        reserve_response = self.do_server_reserve(indent)

        return response.replace(whole_marker, reserve_response)

    def discover_physical_luns(self, args=''):
        """
        Discover information about one or more LUNs visible on this server.
        If an empty argument string is specified, return information
        about all LUNs

        @param      Space-separated list of SCSI IDs to report on

        @return    If successful, returns the discovery data in XML
                   Raises an exception on failure.
        """

        lock = filelock.FileLock('/var/run/ovs-agent/discover_physical_luns.lock')
        lock.acquire(wait=120)
        try:
            if args == "":
                disk_list = ""
            else:
                disk_list = args.split()
            indent = 0
            (indent, header) = XML_header(indent)
            doc = add_all_block_data(indent, disk_list)
            (indent, trailer) = XML_trailer(indent)
            result = header + doc + trailer
            result = discover_dm_device(result)
            result = discover_nvme_device(result)
            if glob.glob('/dev/disk/by-id/md-uuid*'):
                result = discover_md_device(result)
            return result
        finally:
            lock.release()

    def discover_mounted_file_systems(self, args):
        """
        Get mounted file systems of the specified type and NFS exports managed
        by the agent.

        @return     Returns the discovery data as an XML document.
                    Raises an exception on failure
        """

        class MountInfo:
            def __init__(self, fs_file, fs_spec, fs_mntops):
                self.fs_file = fs_file
                self.fs_spec = fs_spec
                self.fs_mntops = fs_mntops

        lock = filelock.FileLock('/var/run/ovs-agent/discover_mounted_file_systems.lock')
        lock.acquire(wait=120)
        try:
            f = open("/proc/mounts")
            all_mounts = None
            mounts = {}
            doc = SimpleDiscoveryXML("Discover_Mounted_File_Systems_Result")
            fstypes = args.split()
            for fstype in fstypes:
                if fstype == 'all':
                    all_mounts = 'all'
                    break
            for line in f:
                fs_spec, fs_file, fs_vfstype, fs_mntops, fs_freq, fs_passno = line.split()
                if not mounts.has_key(fs_vfstype):
                    mounts[fs_vfstype] = list()
                mounts[fs_vfstype].append(MountInfo(fs_file, fs_spec, fs_mntops))
            # return all of the mounted files systems or just the specified ones
            if all_mounts == 'all':
                fstypes = mounts.keys()
            for fstype in fstypes:
                mount_list = mounts.get(fstype)
                if mount_list != None:
                    doc.push("Filesystem", "Type", fstype)
                    for mount_entry in mount_list:
                        #doc.push("Mount", "Dir", mount_entry.fs_file)
                        doc.push("Mount", "Dir", mount_entry.fs_file.rstrip('/'))
                        doc.add("Device", mount_entry.fs_spec)
                        doc.add("Mount_Options", mount_entry.fs_mntops)
                        doc.pop()
                    doc.pop()
            return doc.finish()
        finally:
            lock.release()

    def do_server_reserve(self, indent):
        exclusions = ''
        exclusions_list = get_lun_exclusions()
        if exclusions_list:
            exclusions = XML_tag("ServerReserved", indent, True)
            lun = 0
            for entry in exclusions_list:
                exclusions += XML_tag('Disk Num="' + ("%d" % lun) + '"', indent + 2, True)
                exclusions += XML_tag_with_value("WWID", indent + 4, entry)
                exclusions += XML_tag("Disk", indent + 2, False)
                lun += 1
            exclusions += XML_tag("ServerReserved", indent, False)
        return exclusions

    def get_system_disk_space(self):
        output = run_cmd(['df', '-B', 'M', '-P'])
        lines = output.splitlines()
        sys_disk_info = {}
        for line in lines:
            if not line.startswith('Filesystem'):
                fs_name, fs_size, fs_used, fs_avail, fs_pcnt, fs_mnt = line.split()
                if fs_mnt == '/' or fs_mnt == '/boot':
                    sys_disk_info[fs_mnt] = int(fs_avail.strip('M'))
        return sys_disk_info

def get_lun_exclusions():
    luns = []
    if os.path.exists("/etc/boot_sysdisk_wwid"):
        f = open("/etc/boot_sysdisk_wwid", 'r')
        try:
            lines = f.readlines()
        finally:
            f.close()
        for line in lines:
            if line.startswith('/ = '):
                line = line.replace('/ = ', "").strip()
                luns.append(line)
    return luns


def discover_file_system_statistic(doc):
    doc.push('File_System_Statistic')
    fs_stats_interval = get_fs_statistic_interval()
    doc.add('Interval', str(fs_stats_interval))

    fs_uuids = get_fs_stats_uuids()
    if fs_uuids:
        doc.push('File_System_Id')
        for fs_uuid in fs_uuids:
            doc.add('Id', fs_uuid)
        doc.pop() # File_system_id
    doc.pop() # File_System_Statistic


def discover_exports(doc):
    doc.push('Exports')
    for v in dump_db(LOCAL_EXPORTS_DB).values():
        doc.push('Export', 'UUID', v['export_uuid'])
        doc.add('Type', v['export_type'])
        doc.add('Client', v['client'])
        doc.add('Path', v['path'])
        doc.add('Options', v['options'])
        doc.pop()  # Export
    doc.pop()  # Exports


def discover_server_capabilities(doc):
    doc.push('Capabilities')
    for (key, value) in CAPABILITIES.items():
        doc.add(str(key), str(value))
    doc.pop()


def xml_create_text_node(doc, parent_node, name, text, **kwargs):
    new_node = doc.createElement(name)
    new_text_node = doc.createTextNode(str(text))
    new_node.appendChild(new_text_node)
    for key, value in kwargs.items():
        new_node.setAttribute(key, str(value))
    parent_node.appendChild(new_node)


def get_md_raid_info(dev):
    md_info = {}
    md_info['RaidDevices'] = []
    output = run_cmd(['mdadm', '-D', dev])
    for entry in output.splitlines():
        fields = entry.split(':')
        if 'Raid Level' in fields[0]:
            md_info['RaidLevel'] = fields[1].strip()
            continue
        if 'Total Devices' in fields[0]:
            md_info['TotalDevices'] = fields[1].strip()
            continue
        if '/dev/' in fields[0] and 'md' not in fields[0]:
            raid = fields[0].rsplit(' /', 1)[1]
            md_info['RaidDevices'].append(os.path.basename(raid))
            continue
    return md_info

        
def discover_md_device(result):
    doc = xml.dom.minidom.parseString(result)
    scsi = doc.getElementsByTagName("SCSI")[0]
    disk_path = "/dev/disk/by-id"
    dev_list = glob.glob(os.path.join(disk_path, 'md-uuid*'))
    raid_wwids = {}
    raid_devices = []

    # association raid devices name with wwid
    dm_devs = glob.glob(os.path.join(disk_path, 'dm-name*'))
    for dev_name in dm_devs:
        device = os.readlink(dev_name).replace('../../', '')
        # save device mapper wwid by key dm id, dm-1, dm-2, ..
        raid_wwids[device] = dev_name.split('name-')[1]
    nvme_devs = glob.glob(os.path.join(disk_path, 'nvme*'))
    for dev_name in nvme_devs:
        device = os.readlink(dev_name).replace('../../', '')
        # save NVMe wwid by key device name, nvme0n1, nvme0n2, ..
        raid_wwids[device] = os.path.basename(dev_name)

    rootdev = get_rootdev()
    bootdev = get_bootdev()
    swap = get_swaps()[0]
    systems = [rootdev, bootdev, swap]
    for dev_uuid in dev_list:
        # Discover MD Raid if 
        # 1. device has no partition
        # 2. device has partition, only the partitions will be
        #    discovered
        # 3. device is not a system disk
        if 'part' not in dev_uuid and glob.glob(dev_uuid + '-part*'):
            continue
        device = os.readlink(dev_uuid)
        dev_name = device.replace('../../', '/dev/')
        md_info = {}

        try:
            md_info = get_md_raid_info(dev_name)
        except Exception, err:
            logging.error('Error getting md raid info: %s', err)

        if md_info:
            if dev_name not in systems:
                disk = doc.createElement('Disk')
                disk.setAttribute('TYPE', 'LOCAL')
                disk.setAttribute('WWID', os.path.basename(dev_uuid))
                array_size = get_file_size(dev_name)
                xml_create_text_node(doc, disk, 'Size', array_size, unit='bytes')
                xml_create_text_node(doc, disk, 'DevPath', dev_uuid)
                scsi.appendChild(disk)
            else:
                logging.debug('device on the system disk: %s', dev_name)
            # build raid device wwid list from its device name
            for raid_name in md_info.get('RaidDevices'):
                if raid_wwids.has_key(raid_name):
                    raid_devices.append(raid_wwids[raid_name])

    # post processing, remove raid devices from xml output list
    if raid_devices:
        for node in doc.getElementsByTagName("Disk"):
            wwid = node.getAttribute("WWID")
            if wwid in raid_devices:
                scsi.removeChild(node)

    return doc.toxml()


def discover_nvme_device(result):
    doc = xml.dom.minidom.parseString(result)
    scsi = doc.getElementsByTagName("SCSI")[0]
    disk_path = '/dev/disk/by-id/'
    dev_list = glob.glob(os.path.join(disk_path, 'nvme*'))

    for dev_name in dev_list:
       dev_name = os.path.basename(dev_name)
       if not dev_name.startswith('nvme'):
           continue
       # Discover NVMe devices if
       # 1. device has no partition
       # 2. device has partitions, then only the partitions
       #    will be discovered not parent device
       # 3. and the device is not on the system disk
       if 'part' not in dev_name and glob.glob(os.path.join(disk_path, dev_name + '-part*')):
           continue
       device = os.readlink(os.path.join(disk_path, dev_name))
       dev = device.replace('../../', '/dev/')
       rootdev = get_rootdev()
       dev_info = {}
       if rootdev not in dev:
           try:
               dev_info = get_dev_info(dev)
           except Exception, err:
               logging.error('Error getting nvme dev info: %s', err)
       else:
           logging.debug('device on the system disk: %s', dev)

       if dev_info:
           nvme_name = disk_path + '%s' % dev_name
           dev_size = get_file_size(nvme_name)
           disk = doc.createElement('Disk')
           disk.setAttribute('TYPE', 'LOCAL')
           disk.setAttribute('WWID', dev_name)
           xml_create_text_node(doc, disk, 'Transport', 'LOCAL')
           xml_create_text_node(doc, disk, 'Vendor', dev_info.get('ID_VENDOR'))
           xml_create_text_node(doc, disk, 'Product', dev_info.get('ID_MODEL'))
           xml_create_text_node(doc, disk, 'Rev', dev_info.get('ID_REVISION'))
           xml_create_text_node(doc, disk, 'Size', dev_size, unit='bytes')
           xml_create_text_node(doc, disk, 'DevPath', nvme_name)
           scsi.appendChild(disk)
    return doc.toxml()


def discover_dm_device(result):
    dev_info = get_root_dev_info()
    if not dev_info:
        return result

    dev_id = dev_info.get('ID_SERIAL')
    if not dev_id:
        return result

    dm_name = OVM_SYS_REPO_PART_PREFIX + dev_id

    dm_dev = '/dev/mapper/%s' % dm_name
    if not os.path.exists(dm_dev):
        return result

    dm_vendor = dev_info.get('ID_VENDOR', 'Device_Mapper')
    dm_model = dev_info.get('ID_MODEL', 'Device_Mapper')
    dm_revision = dev_info.get('ID_REVISION', '1.0')

    try:
        dm_info = run_cmd(['dmsetup', 'info', '-C', '--noheadings', '-o',
                           'major,minor,uuid', dm_dev])
        dm_major, dm_minor, dm_uuid = dm_info.strip().split(':')
        dm_dev_node = '%s:%s' % (dm_major, dm_minor)
        dm_size = get_file_size(dm_dev)
    except Exception, err:
        logging.error('Error getting device %s info: %s', dm_dev, err)
        return result

    doc = xml.dom.minidom.parseString(result)
    scsi = doc.getElementsByTagName("SCSI")[0]
    disk = doc.createElement('Disk')
    disk.setAttribute('TYPE', 'LOCAL')
    disk.setAttribute('WWID', dm_uuid)
    xml_create_text_node(doc, disk, 'Transport', 'LOCAL')
    xml_create_text_node(doc, disk, 'Vendor', dm_vendor)
    xml_create_text_node(doc, disk, 'Product', dm_model)
    xml_create_text_node(doc, disk, 'Rev', dm_revision)
    xml_create_text_node(doc, disk, 'Size', dm_size, unit='bytes')
    multipath = doc.createElement('Multipath')
    multipath.setAttribute('DeviceMap', dm_dev)
    path = doc.createElement('Path')
    path.setAttribute('Dev', dm_name)
    xml_create_text_node(doc, path, 'DevPath', dm_dev)
    xml_create_text_node(doc, path, 'Device', dm_dev_node)
    xml_create_text_node(doc, path, 'Path_State', 'active')
    xml_create_text_node(doc, path, 'Check_State', 'ready')
    multipath.appendChild(path)
    disk.appendChild(multipath)
    scsi.appendChild(disk)

    return doc.toxml()


def XML_header(indent):
    header = '<?xml version="1.0" encoding="UTF-8"?>\n'
    indent += XML_INDENT

    header = header + XML_tag("Discover_Physical_LUNS_Result", indent, True)
    indent += XML_INDENT
    return (indent, header)


def XML_trailer(indent):
    indent -= XML_INDENT
    trailer = XML_tag("Discover_Physical_LUNS_Result", indent, False)
    return (indent, trailer)


def XML_tag_with_value(element, indent, value):
    template = "%-*s<%s>%s</%s>\n"
    if element == "" or indent < 0:
        return "? invalid arguments"

    return template % (indent, "", element, value, element)


def XML_tag(element, indent, start):
    template = "%-*s<%s%s>\n"
    if element == "" or indent < 0:
        return "? invalid arguments"

    if start:
        close = ""
    else:
        close = "/"
    return template % (indent, "", close, element)


def parse_oracle_release():
    result = {'OS_Name': '',
              'OS_Major_Version': '',
              'OS_Minor_Version': ''}
    try:
        release = open('/etc/oracle-release')
        content = release.read()
        release.close()
        name, version = content.split('release')
        major = version.split('.')[0].strip()
        minor = version.split('.')[1].strip()
        result['OS_Name'] = name.strip()
        result['OS_Major_Version'] = major
        result['OS_Minor_Version'] = minor
    except Exception, err:
        logging.warn('Error parsing release: %s', err)
    return result


def yum_clean():
    try:
        run_cmd(['/usr/bin/yum', 'clean', 'all'])
    except Exception, err:
        logging.warn('Cannot remove yum cache data: %s', err)
