# Copyright (C) 2006-2016 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.

import re
import os
import time
import signal
import logging
from xen.xend.xenstore.xstransact import xstransact
from xen.xend.xenstore.xsutil import GetDomainPath

from agent.lib.constants import LOCAL_SERVER_DB
from agent.api.hypervisor.base import Hypervisor
from agent.lib import filelock
from agent.lib.db import read_item
from agent.lib.settings import get_vm_lock_file, get_vm_path
from agent.lib.xenvm import (get_cfg_path,  read_cfg,
                             configure_cfg, unconfigure_cfg)
from agent.lib.xenxm import cleanup_xc_restore
from agent.target.hypervisor import (is_vm_running_locally,
                                     dump_core, get_domid,
                                     create_vm, shutdown_vm, destroy_vm,
                                     reboot_vm, pause_vm, unpause_vm,
                                     save_vm, restore_vm, migrate_vm, migrate_vm_with_storage,
                                     storage_migration_setup, storage_migration_cleanup,
                                     storage_migration_cfgfile_setup, storage_migration_cfgfile_cleanup,
                                     storage_migration_delete_vm, notify_vm_vdisk,trigger_vm, sysrq_vm, set_vm_params,
                                     get_vm, get_vms)
from agent.utils import xmlrpclib


MAX_KEYNAME_LEN = 256
MAX_KEYVALUE_LEN = 8192
MAX_ENTRY_SIZE = 2048
MAX_MESSAGE_NUMBER = 256


def get_cfg_vm_name(repo_id, vm_id):
    cfgfile = get_cfg_path(repo_id, vm_id)
    return read_cfg(cfgfile)['name']


def get_cfg_vm_uuid(repo_id, vm_id):
    cfgfile = get_cfg_path(repo_id, vm_id)
    return read_cfg(cfgfile)['uuid']


class Xen(Hypervisor):
    """
    Xen hypervisor management.
    """

    def get_services(self, version=None):
        return {
            'create_vm': self.create_vm,
            'delete_vm': self.delete_vm,
            'start_vm': self.start_vm,
            'stop_vm': self.stop_vm,
            'reboot_vm': self.reboot_vm,
            'pause_vm': self.pause_vm,
            'unpause_vm': self.unpause_vm,
            'save_vm': self.save_vm,
            'restore_vm': self.restore_vm,
            'migrate_vm': self.migrate_vm,
            'migrate_vm_with_storage': self.migrate_vm_with_storage,
            'storage_migration_setup':  self.storage_migration_setup,
            'storage_migration_cleanup': self.storage_migration_cleanup,
            'storage_migration_cfgfile_setup':  self.storage_migration_cfgfile_setup,
            'storage_migration_cfgfile_cleanup': self.storage_migration_cfgfile_cleanup,
            'storage_migration_delete_vm':self.storage_migration_delete_vm,
            'notify_vm_vdisk': self.notify_vm_vdisk,
            'trigger_vm': self.trigger_vm,
            'sysrq_vm': self.sysrq_vm,
            'dump_vm_core': self.dump_vm_core,
            'list_vm_core': self.list_vm_core,
            'delete_vm_core': self.delete_vm_core,
            'configure_vm': self.configure_vm,
            'unconfigure_vm': self.unconfigure_vm,
            'get_vm_config': self.get_vm_config,
            'set_vm_config': self.set_vm_config,
            'list_vm': self.list_vm,
            'list_vms': self.list_vms,
            'send_to_guest': self.send_to_guest,
            'configure_vm_ha': self.configure_vm_ha,
            'create_template': self.create_template,
            'delete_template': self.delete_template,
            'configure_template': self.configure_template,
            'unconfigure_template': self.unconfigure_template,
            'get_template_config': self.get_template_config,
            'set_template_config': self.set_template_config,
            'import_vm': self.import_vm,
            'import_template': self.import_template,
            'import_assembly': self.import_assembly,
            'create_assembly': self.create_assembly,
            'get_assembly_config': self.get_assembly_config,
            'set_assembly_config': self.set_assembly_config,
            'get_assembly_config_xml': self.get_assembly_config_xml,
            'set_assembly_config_xml': self.set_assembly_config_xml,
            'assembly_add_file': self.assembly_add_file,
            'assembly_del_file': self.assembly_del_file,
            'assembly_rename_file': self.assembly_rename_file,
            'pack_assembly': self.pack_assembly,
            'unpack_assembly': self.unpack_assembly,
            'deploy_assembly': self.deploy_assembly,
            'deploy_assembly_to_vm': self.deploy_assembly_to_vm,
            'delete_assembly': self.delete_assembly,
            'setup_migration_target': self.setup_migration_target,
            'cleanup_migration_target': self.cleanup_migration_target
        }

    def start_vm(self, repo_id, vm_id):

        def on_stop(sig, frame):
            self.stop_vm(repo_id, vm_id, True)
            raise Exception('Task stopped by user. Kill the VM.')

        signal.signal(signal.SIGUSR2, on_stop)

        server_roles = read_item(LOCAL_SERVER_DB, 'roles')
        if isinstance(server_roles, set) and 'xen' in server_roles:
            path = get_cfg_path(repo_id, vm_id)
            create_vm(path)
        else:
            raise RuntimeError('Server is not allowed to run VMs.')

    def stop_vm(self, repo_id, vm_id, force=False):
        if repo_id:
            domain = get_cfg_vm_name(repo_id, vm_id)
        else:
            domain = vm_id
        if not force:
            shutdown_vm(domain)
        else:
            try:
                domid = get_domid(domain)
                destroy_vm(domain)
                cleanup_xc_restore(domid)
            except RuntimeError, error:
                logging.info('Destroy VM %s failed: %s', domain, error)
                migrating_domain = 'migrating-%s' % domain
                try:
                    destroy_vm(migrating_domain)
                except RuntimeError, err:
                    logging.info('Destroy VM %s failed: %s', migrating_domain, err)

    def reboot_vm(self, repo_id, vm_id, wait=True):

        def on_stop(sig, frame):
            self.stop_vm(repo_id, vm_id, True)
            raise Exception('Task stopped by user. Kill the VM.')

        signal.signal(signal.SIGUSR2, on_stop)

        reboot_vm(get_cfg_vm_name(repo_id, vm_id), wait)

    def pause_vm(self, repo_id, vm_id):
        pause_vm(get_cfg_vm_name(repo_id, vm_id))

    def unpause_vm(self, repo_id, vm_id):
        unpause_vm(get_cfg_vm_name(repo_id, vm_id))

    def save_vm(self, repo_id, vm_id, checkpoint=False):
        state_file = os.path.join(get_vm_path(repo_id, vm_id), 'state')

        def on_stop(sig, frame):
            try:
                self.stop_vm(repo_id, vm_id, True)
            except Exception, err:
                logging.debug('Failed to stop vm %s: %s', vm_id, err)
            try:
                os.unlink(state_file)
            except Exception:
                pass
            raise Exception('Task stopped by user. Kill the VM and remove state file.')

        signal.signal(signal.SIGUSR2, on_stop)

        try:
            save_vm(get_cfg_vm_name(repo_id, vm_id), state_file, checkpoint)
        except Exception, err:
            try:
                os.unlink(state_file)
            except Exception:
                pass
            raise err

    def restore_vm(self, repo_id, vm_id, paused=False):

        def on_stop(sig, frame):
            try:
                self.stop_vm(repo_id, vm_id, True)
            except Exception, err:
                logging.debug('Failed to stop vm %s: %s', vm_id, err)
            raise Exception('Task stopped by user. Kill the VM.')

        signal.signal(signal.SIGUSR2, on_stop)

        try:
            state_file = os.path.join(get_vm_path(repo_id, vm_id), 'state')
            restore_vm(state_file, paused)
            os.unlink(state_file)
        except Exception, err:
            try:
                self.stop_vm(repo_id, vm_id, True)
            except Exception, err2:
                logging.debug('Failed to stop vm %s: %s', vm_id, err2)
            raise err

    def migrate_vm(self, repo_id, vm_id, dest, live=True, ssl=True):
        migrate_vm(get_cfg_vm_name(repo_id, vm_id), dest, live, ssl)

    def migrate_vm_with_storage(self, repo_id, vm_id, dest, vmfiles, vmcfg, live=True, ssl=True):
        migrate_vm_with_storage(get_cfg_vm_name(repo_id, vm_id), dest, vmfiles, vmcfg, live, ssl)

    def storage_migration_setup(self, vmfiles):
        storage_migration_setup(vmfiles)

    def storage_migration_cleanup(self, vmfiles, delete_vmfiles):
        storage_migration_cleanup(vmfiles, delete_vmfiles)

    def storage_migration_cfgfile_setup(self, repo_uuid, vm_uuid):
        storage_migration_cfgfile_setup(repo_uuid, vm_uuid)

    def storage_migration_cfgfile_cleanup(self, repo_uuid, vm_uuid):
        storage_migration_cfgfile_cleanup(repo_uuid, vm_uuid)

    def storage_migration_delete_vm(self, repo_uuid, vm_uuid):
        storage_migration_delete_vm(repo_uuid, vm_uuid)

    def notify_vm_vdisk(self, virtual_disk):
        notify_vm_vdisk(virtual_disk)

    def trigger_vm(self, repo_id, vm_id, name, vcpu=None):
        trigger_vm(get_cfg_vm_name(repo_id, vm_id), name, vcpu)

    def sysrq_vm(self, repo_id, vm_id, letter):
        sysrq_vm(get_cfg_vm_name(repo_id, vm_id), letter)

    def dump_vm_core(self, repo_id, vm_id, live=False, crash=False, reset=False):
        now = time.gmtime()
        core_filename = 'core-%s' % time.strftime('%Y-%m%d-%H%M.%S', now)
        core_file = os.path.join(get_vm_path(repo_id, vm_id), core_filename)
        dump_core(get_cfg_vm_name(repo_id, vm_id), core_file, live, crash, reset)
        return xmlrpclib.DateTime(now)

    def list_vm_core(self, repo_id, vm_id):
        result = []
        for filename in os.listdir(get_vm_path(repo_id, vm_id)):
            match = re.match(r'^core-(?P<date>.+)$', filename)
            if match:
                date = time.strptime(match.group('date'), '%Y-%m%d-%H%M.%S')
                result.append(xmlrpclib.DateTime(date))
        return result

    def delete_vm_core(self, repo_id, vm_id, core_date):
        date = time.strptime(core_date.value, '%Y%m%dT%H:%M:%S')
        core_filename = 'core-%s' % time.strftime('%Y-%m%d-%H%M.%S', date)
        core_file = os.path.join(get_vm_path(repo_id, vm_id), core_filename)
        os.unlink(core_file)

    def configure_vm(self, repo_id, vm_id, params):
        lock = filelock.FileLock(get_vm_lock_file(repo_id, vm_id))
        lock.acquire(wait=120)
        try:
            path = get_cfg_path(repo_id, vm_id)
            cfg = read_cfg(path)
            if is_vm_running_locally(cfg['uuid']):
                set_vm_params(cfg['name'], params)
            return configure_cfg(path, params)
        finally:
            lock.release()

    def unconfigure_vm(self, repo_id, vm_id, params):
        lock = filelock.FileLock(get_vm_lock_file(repo_id, vm_id))
        lock.acquire(wait=120)
        try:
            return unconfigure_cfg(get_cfg_path(repo_id, vm_id), params)
        finally:
            lock.release()

    def list_vm(self, repo_id, vm_id):
        vm_name = get_cfg_vm_name(repo_id, vm_id)
        try:
            return get_vm(vm_name)
        except Exception:
            vm_uuid = get_cfg_vm_uuid(repo_id, vm_id)
            if is_vm_running_locally(vm_uuid):
                return get_vm('migrating-%s' %vm_name)
            else:
                raise

    def list_vms(self):
        return get_vms()

    def send_to_guest(self, repo_id, vm_id, params):
        domid = get_domid(get_cfg_vm_name(repo_id, vm_id))
        dompath = GetDomainPath(domid)
        lock = filelock.FileLock(get_vm_lock_file(repo_id, vm_id))
        lock.acquire(wait=120)
        try:
            if not xstransact.List(dompath):
                raise RuntimeError('VMAPI is not ready.')
            last_write_str = xstransact.Read(dompath, 'control/oracle-vmapi/to-guest/last-write')
            if not last_write_str:
                last_write_str = '0'
                xstransact.Write(dompath, 'control/oracle-vmapi/to-guest/last-write', last_write_str)
            last_write = int(last_write_str)
            for (key, value) in params.items():
                entries = xstransact.List('%s/control/oracle-vmapi/to-guest' % dompath)
                if (len(entries) - 1) > MAX_MESSAGE_NUMBER:
                    raise RuntimeError('No space to write message.')
                if len(key) > MAX_KEYNAME_LEN:
                    raise RuntimeError('Message key name too long (should be below or equal to %s): %s' % (MAX_KEYNAME_LEN, key))
                if len(value) > MAX_KEYVALUE_LEN:
                    raise RuntimeError('Message key value too long (should be below or equal to %s): %s' % (MAX_KEYVALUE_LEN, value))
                to_store = {}
                last_write += 1
                iteration = 0
                while len(value) > 0:
                    to_store['control/oracle-vmapi/to-guest/%s/%s' % (last_write, iteration)] =  value[:MAX_ENTRY_SIZE]
                    value = value[MAX_ENTRY_SIZE:]
                    iteration += 1
                to_store['control/oracle-vmapi/to-guest/%s' % last_write] = key
                to_store['control/oracle-vmapi/to-guest/last-write'] = str(last_write)
                xstransact.Write(dompath, to_store)
        finally:
            lock.release()
