# Copyright (C) 2006-2016 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.

import os
import logging
from os.path import join, basename, exists, dirname
from agent.api.poolfs.base import PoolFilesystem
from agent.lib import filelock
from agent.lib.constants import POOLFS_META, POOLFS_VERSION, NFS_MOUNT
from agent.lib.discover import SimpleDiscoveryXML
from agent.lib.settings import parse_metafile
from agent.target.cluster import (POOLFS_IMG, POOLFS_IMG_SIZE, POOLFS_LABEL,
                                  POOLFS_NODE_SLOTS,
                                  mount_device, umount_device,
                                  ocfs2_file_lookup, ocfs2_file_read,
                                  ocfs2_file_write)
from agent.target.host import run_cmd


class PoolOCFS2(PoolFilesystem):
    """
    Class for managing the shared metadata in a pool using bdb on a small clustered ocfs2 fs.
    """

    def get_services(self, version=None):
        return {
            'discover_pool_filesystem': self.discover_pool_filesystem,
            'create_pool_filesystem': self.create_pool_filesystem,
            'destroy_pool_filesystem': self.destroy_pool_filesystem,
            'ocfs2_get_meta_data': self.ocfs2_get_meta_data,
        }

    def discover_pool_filesystem(self):
        """
        Discover information about the pool filesystem.
        Returns an empty document if no pool filesystem found.

        @param      None
        @return     Discovery data as an XML document.
                    Raises an exception on failure.
        """
        lock = filelock.FileLock('/var/run/ovs-agent/discover_pool_filesystem.lock')
        lock.acquire(wait=120)
        try:
            doc = SimpleDiscoveryXML("Discover_Pool_Filesystem_Result")
            add_nfs_poolfs_info(doc)
            add_lun_poolfs_info(doc)
            return doc.finish()
        finally:
            lock.release()


    def create_pool_filesystem(self, poolfs_type, poolfs_target, cluster_name,
            poolfs_uuid, poolfs_nfsbase_uuid, manager_uuid, pool_uuid,
            block_size="4K", cluster_size="4K", journal_size="4M"):
        """
        Create a pool filesystem.

        @param      poolfs_type - "nfs" or "lun"
        @param      poolfs_target -
                    physical location of the pool filesystem
        @param      cluster_name - cluster to use the pool filesystem
        @param      poolfs_uuid -
                    a unique identifier for the pool filesystem
        @param      poolfs_nfsbase_uuid -
                    a unique identifier for the base NFS filesystem
        @param      manager_uuid - a unique identifier for the manager
        @param      pool_uuid - a unique identifier for the server pool
        @param      block_size - mkfs.ocfs2 option
        @param      cluster_size - mkfs.ocfs2 option
        @param      journal_size - mkfs.ocfs2 option

        @return     None. Raises an exception on failure.
        """
        if poolfs_type == "nfs":
            nfs_create_pool_filesystem(poolfs_target, cluster_name, poolfs_uuid,
                    poolfs_nfsbase_uuid, manager_uuid, pool_uuid)
        elif poolfs_type == "lun":
            lun_create_pool_filesystem(poolfs_target, cluster_name, poolfs_uuid,
                    manager_uuid, pool_uuid)
        else:
            raise ValueError("Invalid pool filesystem type: %s" % poolfs_type)


    def destroy_pool_filesystem(self, poolfs_type, poolfs_target, poolfs_uuid, poolfs_nfsbase_uuid):
        """
        Destroy a pool filesystem.

        @param      poolfs_type - "nfs" or "lun"
        @param      poolfs_target -
                    physical location of the pool filesystem
        @param      poolfs_uuid -
                    a unique identifier for the pool filesystem
        @param      poolfs_nfsbase_uuid -
                    a unique identifier for the base NFS filesystem

        @return     None. Raises an exception on failure.
        """
        if poolfs_type == "nfs":
            nfs_destroy_pool_filesystem(poolfs_target, poolfs_nfsbase_uuid)
        elif poolfs_type == "lun":
            lun_destroy_pool_filesystem(poolfs_target)
        else:
            raise ValueError("Invalid pool filesystem type: %s" % poolfs_type)

    def ocfs2_get_meta_data(self, device, filename):
        """
        @return     Meta data as a dict.
                    Raises an exception on failure.
        """
        # flush the cache
        run_cmd(['sync'])
        if ocfs2_file_lookup(device, filename) == 0:
            contents = ocfs2_file_read(device, filename)
            return parse_metafile(contents)
        else:
            return {}


def nfs_create_pool_filesystem(poolfs_target, cluster_name, poolfs_uuid,
        poolfs_nfsbase_uuid, manager_uuid, pool_uuid):
    # mount NFS base
    mount_point = join(NFS_MOUNT, poolfs_nfsbase_uuid)
    mount_device(poolfs_target, mount_point)
    poolfs_meta = join(mount_point, POOLFS_META)
    if os.path.exists(poolfs_meta):
        umount_device(poolfs_target) 
        raise Exception("NFS mount point %s already in use "
                        "in another pool filesystem" % mount_point)
    # make pool filesystem
    poolfs_img = join(mount_point, POOLFS_IMG)
    create_poolfs_image(poolfs_img)
    make_pool_filesystem(poolfs_img, poolfs_uuid, cluster_name)
    # generate poolfs metafile contents
    contents = """OVS_POOLFS_UUID=%s
OVS_POOLFS_NFSBASE_UUID=%s
OVS_POOLFS_MGR_UUID=%s
OVS_POOLFS_VERSION=%s
OVS_POOLFS_POOL_UUID=%s
""" % (poolfs_uuid,
poolfs_nfsbase_uuid,
manager_uuid,
POOLFS_VERSION,
pool_uuid,
)
    try:
        # write poolfs metafile
        f = file(poolfs_meta, "w")
        try:
            f.write(contents)
        finally:
            f.close()
        # umount poolfs nfsbase
        umount_device(poolfs_target)
    except Exception, e:
        if exists(poolfs_meta):
            os.remove(poolfs_meta)
        raise e


def lun_create_pool_filesystem(poolfs_target, cluster_name, poolfs_uuid,
        manager_uuid, pool_uuid):
    #check if an OCFS2 cluster is already present on this disk
    try:
        cmd = ["o2cluster", "-o", "%s" % poolfs_target]
        run_cmd(cmd)
    except Exception, e:
        logging.debug("o2cluster returned %s", e)
        logging.debug("lun_create_pool_filesystem: no OCFS2 system present, proceeding...")
    else:
        raise Exception("%s already in a cluster" % poolfs_target)

    # destroy pool filesystem
    lun_destroy_pool_filesystem(poolfs_target)
    # make pool filesystem
    make_pool_filesystem(poolfs_target, poolfs_uuid, cluster_name)
    # generate poolfs metafile contents
    contents = """OVS_POOLFS_UUID=%s
OVS_POOLFS_MGR_UUID=%s
OVS_POOLFS_VERSION=%s
OVS_POOLFS_POOL_UUID=%s
OVS_POOLFS_LUN_UUID=%s
""" % (poolfs_uuid,
manager_uuid,
POOLFS_VERSION,
pool_uuid,
basename(poolfs_target),
)
    try:
        # write poolfs metafile
        ocfs2_file_write(poolfs_target, POOLFS_META, contents)
    except Exception, e:
        lun_destroy_pool_filesystem(poolfs_target)
        raise e


def nfs_destroy_pool_filesystem(poolfs_target, poolfs_nfsbase_uuid):
    mount_point = join(NFS_MOUNT, poolfs_nfsbase_uuid)
    mount_device(poolfs_target, mount_point)
    poolfs_meta = join(mount_point, POOLFS_META)
    if exists(poolfs_meta):
        os.remove(poolfs_meta)
    umount_device(poolfs_target)


def lun_destroy_pool_filesystem(poolfs_target):
    cmd = [ "dd", "if=/dev/zero", "of=%s" % poolfs_target,
            "bs=1M", "count=4" ]
    run_cmd(cmd)


def add_nfs_poolfs_info(doc):
    """
    Add XML for NFS based pool filesystem.
    """
    output = run_cmd(["cat", "/proc/mounts"])
    lines = output.splitlines()

    for line in lines:
        parts = line.split()
        if parts[1].startswith("/nfsmnt/") and parts[2].startswith("nfs"):
            poolfs_meta = join(parts[1], POOLFS_META)
            if exists(poolfs_meta):
                f = file(poolfs_meta)
                try:
                    contents = f.read()
                finally:
                    f.close()
                d = parse_metafile(contents)

                doc.push("Pool_Filesystem")
                doc.add("Pool_Filesystem_Type", "nfs")
                doc.add("Pool_Filesystem_Target", parts[0])
                doc.add("Pool_Filesystem_Uuid", d.get("OVS_POOLFS_UUID"))
                doc.add("Pool_Filesystem_Nfsbase_Uuid",
                        d.get("OVS_POOLFS_NFSBASE_UUID"))
                doc.add("Pool_Filesystem_Manager_Uuid",
                        d.get("OVS_POOLFS_MGR_UUID"))
                doc.add("Pool_Filesystem_Version",
                        d.get("OVS_POOLFS_VERSION"))
                doc.add("Pool_Filesystem_Pool_Uuid",
                        d.get("OVS_POOLFS_POOL_UUID"))
                doc.pop()


def add_lun_poolfs_info(doc):
    """
    Add XML for LUN based pool filesystem.
    """
    output = run_cmd(["multipathd", "show", "maps", "format", "%n %N"])
    lines = output.splitlines()
    # Skip output header and device paths count is 0
    for entry in lines:
        if entry.startswith("name"):
            continue
        key, value = entry.split()
        if value == '0':
            logging.debug("Device %s paths count is 0.", key)
            continue

        # Prevent mounted.ocfs2 command on unmapped devices
        output = run_cmd(["mounted.ocfs2", "-d", "/dev/mapper/" + key])
        # Skip the header of mounted.ocfs2 output
        line = output.splitlines()
        if not len(line) > 1:
            continue

        parts = line[1].split()
        if parts[-1] == POOLFS_LABEL:
            try:
                if ocfs2_file_lookup(parts[0], POOLFS_META) != 0:
                    continue
                contents = ocfs2_file_read(parts[0], POOLFS_META)
                d = parse_metafile(contents)

                doc.push("Pool_Filesystem")
                doc.add("Pool_Filesystem_Type", "lun")
                doc.add("Pool_Filesystem_Target", parts[0])
                doc.add("Pool_Filesystem_Uuid", d.get("OVS_POOLFS_UUID"))
                doc.add("Pool_Filesystem_Lun_Uuid",
                        d.get("OVS_POOLFS_LUN_UUID"))
                doc.add("Pool_Filesystem_Manager_Uuid",
                        d.get("OVS_POOLFS_MGR_UUID"))
                doc.add("Pool_Filesystem_Version",
                        d.get("OVS_POOLFS_VERSION"))
                doc.add("Pool_Filesystem_Pool_Uuid",
                        d.get("OVS_POOLFS_POOL_UUID"))
                doc.pop()
            except Exception, e:
                logging.warn('Cannot get poolfs info: %s', e)


def create_poolfs_image(poolfs_img):
    """
    Create a pool filesystem image.

    @param      poolfs_img - pool filesystem image

    @return     None. Raises an exception on failure.
    """
    if exists(poolfs_img):
        os.remove(poolfs_img)

    stat = os.statvfs(dirname(poolfs_img))
    free_space = stat.f_bsize * stat.f_bfree / 1024 / 1024
    if free_space < POOLFS_IMG_SIZE + 1:
        raise Exception("Not enough space for pool filesystem.")

    cmd = [ "dd", "if=/dev/zero", "of=%s" % poolfs_img,
            "bs=1M", "seek=%s" % POOLFS_IMG_SIZE, "count=0" ]
    run_cmd(cmd)


def make_pool_filesystem(device, poolfs_uuid, cluster_name,
        block_size="4K", cluster_size="4K", journal_size="4M"):
    """
    Make an ocfs2 filesystem for global heartbeat.

    @param      device - target device
    @param      poolfs_uuid - pool filesystem uuid
    @param      cluster_name - cluster name
    @param      block_size - block size
    @param      cluster_size - cluster size
    @param      journal size - journal size

    @return     None. Raises an exception on failure.
    """
    cmd = [ "mkfs.ocfs2",
            "-b", block_size,
            "-C", cluster_size,
            "-J", "size=%s" % journal_size,
            "-N", POOLFS_NODE_SLOTS,
            "-L", POOLFS_LABEL,
            "-U", poolfs_uuid,
            "--global-heartbeat",
            "--cluster-stack=o2cb",
            "--cluster-name=%s" % cluster_name,
            "--force",
            device ]
    run_cmd(cmd)
