# Copyright (C) 2006-2016 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.

from agent.api.base import Agent

import logging
import socket
import time

class RemoteManagement(Agent):
    """
    Class for remote server management.
    """

    def sys_shutdown(self):
        """
        Shutdown the server.

        @raise IModuleException: on failure

        """
        raise NotImplementedError()


    def sys_reboot(self):
        """
        Reboot the server.

        @raise IModuleException: on failure

        """
        raise NotImplementedError()


    def remote_power_off(self, controller_type, tgt_host, bmc_username="",
                         bmc_password=""):
        """
        Power off a server.

        @param controller_type: the mechanism to use for executing the
        power off request: C{'IPMI'} for IPMI
        @type  controller_type: C{str}

        @param tgt_host: for IPMI, an IP address where to send the power
        off request
        @type  tgt_host: C{str}

        @param bmc_username: a username for controlling the target host
        @type  bmc_username: C{str}

        @param bmc_password: a password for controlling the target host
        @type  bmc_password: C{str}

        @raise IModuleException: on failure

        """
        logging.info("remote_power_off (IPMI): %s", tgt_host)
        if controller_type == "IPMI":
            self.ipmi_power_off(host=tgt_host, username=bmc_username,
                                password=bmc_password)
        else:
            raise Exception("Controller type not supported: %s" %
                            controller_type)


    def remote_power_on(self, controller_type, tgt, arg1="", arg2=""):
        """
        Power on a server.

        @param controller_type: the mechanism to use for executing the
        power on request: C{'IPMI'} for IPMI, C{'WOL'}
        for Wake-on-LAN (WOL)
        @type  controller_type: C{str}

        @param tgt: for IPMI, an IP address where to send the power
        on request; for WOL, a MAC address for the system to power on
        @type  tgt: C{str}

        @param arg1: for IPMI, a username for controlling the target host;
        for WOL, the network device to use for sending the WOL request
        @type  arg1: C{str}

        @param arg2: for IPMI, a password for controlling the target host;
        for WOL, this parameter is not used
        @type  arg2: C{str}

        @raise IModuleException: on failure

        """
        if controller_type == "WOL":
            logging.info("remote_power_on (%s): net_device=%s %s",
                         controller_type, arg1, tgt)
            self._send_wol(tgt, net_device=arg1)
        elif controller_type == "IPMI":
            logging.info("remote_power_on (%s): %s", controller_type, tgt)
            self.ipmi_power_on(host=tgt, username=arg1, password=arg2)
        else:
            raise Exception("Controller type not supported: %s" %
                            controller_type)


    def ipmi_power_on(self, host, username=None, password=None):
        raise NotImplementedError()


    def ipmi_power_off(self, host, username=None, password=None):
        raise NotImplementedError()


    def _send_wol(self, tgt_mac, net_device=None):
        """
        Attempt to wake the host with the magic packet.
        http://en.wikipedia.org/wiki/Wake-on-LAN#Magic_packet

        The target machine *MUST* have the network card in etherwake
        mode. This is typically achieved with "ethtool -s ethX wol g"
        on the target host before shutdown. Also be aware that if
        standby power to the machine is lost, then this setting will
        usually be lost and the machine will not be able to be revived
        with this method. A few motherboards do allow the wol settings
        to be preserved after a powerloss. Basically this is a rather
        hit or miss affair.

        @param tgt_mac: a MAC address for the system to power on (MAC
        format: aa:bb:cc:dd:ee:ff)
        @type  tgt_mac: C{str}

        @param net_device: the network device to use for sending the
        WOL request
        @type  net_device: C{str}

        """
        mac = tgt_mac.replace(':', '')
        if len(mac) != 12:
            raise ValueError('Invalid MAC address')

        data = '\xff' * 6 + mac.decode('hex') * 16

        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        try:
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)

            if net_device:
                self._bind_sock_to_device(sock, net_device)

            # send the packet multiple times to ensure delivery
            for i in xrange(3):
                sock.sendto(data, ('<broadcast>', 9))
                time.sleep(1)

        finally:
            sock.close()


    def _bind_sock_to_device(self, sock, net_device):
        raise NotImplementedError()
