# Copyright (C) 2006-2016 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.

import IN
import socket
import time
import logging

from agent.api.remote.base import RemoteManagement
from agent.lib.daemon import delay_cmd
from agent.target.host import run_cmd


class LinuxRemoteManagement(RemoteManagement):
    """
    Class for remote server management.
    """

    def get_services(self, version=None):
        return {
            'sys_shutdown': self.sys_shutdown,
            'sys_reboot': self.sys_reboot,
            'remote_power_off': self.remote_power_off,
            'remote_power_on': self.remote_power_on,
        }

    def sys_shutdown(self):
        """
        Shutdown the server.

        @raise IModuleException: on failure

        """
        delay_cmd(5, ["shutdown", "-P", "-h", "now", "shutdown requested by OVM"])

    def sys_reboot(self):
        """
        Reboot the server.

        @raise IModuleException: on failure

        """
        delay_cmd(5, ["shutdown", "-r", "now", "reboot requested by OVM"])

    def ipmi_power_on(self, host, username=None, password=None):
        #
        # It transpires that Dell BMC's are somewhat unreliable
        # (actually it may not just be exclusive to them but the
        # result is that we need to send this command a number of
        # times to make sure it was received)
        # leave in a small interval for the card to wake up and
        # process the command
        #
        retry_ipmi_power_cmd(host, username, password, 'on')

    def ipmi_power_off(self, host, username=None, password=None):
        retry_ipmi_power_cmd(host, username, password, 'off')

    def _bind_sock_to_device(self, sock, net_device):
        sock.setsockopt(socket.SOL_SOCKET, IN.SO_BINDTODEVICE, net_device)


def _ipmitool(host, interface, command, username=None, password=None):
    cmd = ["ipmitool", "-I", interface, "-H", host]
    if username:
        cmd.extend(["-U", username])
    if password:
        cmd.extend(["-P", password])
    cmd.extend(command)
    return run_cmd(cmd, (0, 1))


def ipmi_power_status(host, username, password, interface, state):
    if state == 'on':
        expected_status = 'Chassis Power is on'
    elif state == 'off':
        expected_status = 'Chassis Power is off'
    else:
        raise Exception('Unknown power state: %s' % state)
    try:
        status = _ipmitool(host, interface, ["power", "status"],
                           username, password)
        if status.strip() == expected_status:
            return
        else:
            raise Exception('Unable to verify Chassis Power state is: %s '
                            'for host %s using %s.' % (state, host, interface))
    except Exception as e:
        raise Exception('Failed to get power status of %s by %s: %s' %
                        (host, interface, e))


def retry_ipmi_power_cmd(host, username, password, state, interval=5, count=5):
    if state == 'on':
        cmd = ["power", 'on']
    elif state == 'off':
        cmd = ["power", 'off']
    else:
        raise Exception('Unknown power state: %s' % state)

    for i in range(count):
        interface = 'lanplus'
        try:
            _ipmitool(host, interface, cmd, username, password)
            # When the power is already in the correct state
            # some vendors return a bad status. Use the power status
            # command to verify the power on/off state
            ipmi_power_status(host, username, password, interface, state)
            return
        except Exception as e:
            logging.info('Unable to change power to %s of host %s using %s '
                         'retrying with lan. %s', state, host, interface, e)
        interface = 'lan'
        try:
            _ipmitool(host, interface, cmd, username, password)
            ipmi_power_status(host, username, password, interface, state)
            return
        except Exception as e:
            logging.info('Unable to change power to %s of host %s '
                         'using %s. %s', state, host, interface, e)
        time.sleep(interval)

    raise Exception('Cannot change power state to %s of host %s '
                    'using lan/lanplus.' % (state, host))
