# Copyright (C) 2006-2016 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.

import logging
import re
import socket
import agent.lib.constants
from agent.api.base import Agent
from agent.lib import filelock
from agent.lib.constants import (LOCAL_SERVER_DB, MEMBERSHIP_STATE_OWNED,
                                 MEMBERSHIP_STATE_POOLED,
                                 MEMBERSHIP_STATE_UNOWNED)
from agent.lib.daemon import restart_xmlrpc_server, reload_notification_server
from agent.lib.db import (read_item, write_item, update_db, truncate_db,
                          get_cluster_db_home, dump_db, delete_item)
from agent.lib.discover import SimpleDiscoveryXML
from agent.lib.settings import (get_agt_url, get_membership_state,
                                get_owning_manager_uuid, get_server_pool_uuid,
                                is_clustered, write_manager_certificate,
                                write_signed_server_certificate,
                                remove_manager_certificate,
                                remove_signed_server_certificate,
                                get_manager_certificate,
                                get_pool_member_ip_list)
from agent.target.host import get_server_unique_id, reset_console_access


def validate_roles(roles):
    role_set = set([x for x in roles.split(",") if x.strip()])
    if not role_set <= set(["master", "utility", "xen"]):
        raise Exception("Invalid roles: %s" % role_set)
    return role_set


class ServerPool(Agent):
    """
    Class for OVS server pool management.
    """

    def get_services(self, version=None):
        return {
            'discover_server_pool': self.discover_server_pool,
            'take_ownership': self.take_ownership,
            'release_ownership': self.release_ownership,
            'create_server_pool': self.create_server_pool,
            'join_server_pool': self.join_server_pool,
            'destroy_server_pool': self.destroy_server_pool,
            'leave_server_pool': self.leave_server_pool,
            'set_server_pool_alias': self.set_server_pool_alias,
            'update_server_roles': self.update_server_roles,
            'set_pool_member_ip_list': self.set_pool_member_ip_list,
        }

    def discover_server_pool(self):
        """
        Discover information about the server pool that contains this server.
        Returns an empty document if the server is unassociated.

        @return     If successful, returns the discovery data as an XML document
                    Raises an exception on failure.
        """
        lock = filelock.FileLock('/var/run/ovs-agent/discover_server_pool.lock')
        lock.acquire(wait=120)
        try:
            doc = SimpleDiscoveryXML("Discover_Server_Pool_Result")
            uuid = get_server_pool_uuid()
            if uuid:
                doc.push("Server_Pool")
                doc.add("Unique_Id", uuid)
                add_poolfs_info(doc)
                if is_clustered():
                    add_clustered_info(doc)
                else:
                    add_pooled_info(doc)
                doc.pop() # Server_Pool
            return doc.finish()
        finally:
            lock.release()

    def take_ownership(self, config):
        """
        Mark this server as being managed by a OVM Manager.

        @param config - dict contains manager info
        @return     None.  Raises an exception on failure.
        """
        return take_ownership(config)

    def release_ownership(self, manager_uuid):
        """
        Release this server from being managed by an OVM Manager.

        @param      manager_uuid - UUID of the manager releasing ownership

        @return     None.  Raises an exception on failure.
        """
        return release_ownership(manager_uuid)

    def create_server_pool(self, pool_alias, pool_uuid, pool_virtual_ip,
                           node_number, server_hostname, server_ip, roles):
        """
        Create a server pool and add this server as its initial pool master.

        @param      pool_alias - human readable name for the pool
        @param      pool_uuid - globally unique id for the pool
        @param      pool_virtual_ip - server pool virtual IP
        @param      node_number -
                    node number in server pool and ocfs2 cluster
        @param      server_hostname -
                    hostname manager uses to talk to this server
        @param      server_ip -
                    ip that the manager uses to talk to this server
        @param      roles - text string with list of roles
        @return     None. Raises exception on failure
        """
        pool_add_server(pool_alias, pool_uuid, pool_virtual_ip,
            node_number, server_hostname, server_ip, roles, initial=True)

    def join_server_pool(self, pool_alias, pool_uuid, pool_virtual_ip,
                         node_number, server_hostname, server_ip, roles):
        """
        Add a server to an existing server pool.

        @param      pool_alias - human readable name for the pool
        @param      pool_uuid - globally unique id for the pool
        @param      pool_virtual_ip - server pool virtual IP
        @param      node_number -
                    node number in server pool and ocfs2 cluster
        @param      server_hostname -
                    hostname manager uses to talk to this server
        @param      server_ip -
                    ip that the manager uses to talk to this server
        @param      roles - text string with list of roles

        @return     None. Raises an exception on failure.
        """
        pool_add_server(pool_alias, pool_uuid, pool_virtual_ip,
            node_number, server_hostname, server_ip, roles, initial=False)

    def destroy_server_pool(self, pool_uuid):
        """
        Unjoin the last server (the master) from the server pool leaving
        no record of the pool.

        @param      pool_uuid -
                    globally unique ID of pool this server is in
        @return     None.  Raises an exception on failure.
        """
        pool_remove_server(pool_uuid, final=True)

    def leave_server_pool(self, pool_uuid):
        """
        Unjoin the server from the server pool.

        @param      pool_uuid -
                    globally unique ID of pool this server is in
        @return     None.  Raises an exception on failure
        """
        pool_remove_server(pool_uuid, final=False)

    def set_server_pool_alias(self, pool_alias):
        """
        Set a new human readable name for the server pool.

        @param      pool_alias - human readable name for the pool

        @return     None.  Raises an exception on failure.
        """
        if not pool_alias:
            raise ValueError("Null or missing parameter")
        state = get_membership_state()
        if state in (MEMBERSHIP_STATE_UNOWNED, MEMBERSHIP_STATE_OWNED):
            raise Exception("Server is not yet in a server pool")
        write_item(LOCAL_SERVER_DB, "pool_alias", pool_alias)
        if get_master_flag() and is_clustered():
            write_item("server_pool", "pool_alias", pool_alias,
                       get_cluster_db_home())


    def update_server_roles(self, roles):
        """
        Update the roles for the server.

        @param      roles - role names, separated by commas
        @return     None.  Raises an exception on failure.
        """
        role_set = validate_roles(roles)
        write_item(LOCAL_SERVER_DB, "roles", role_set)

    def set_pool_member_ip_list(self, ip_list):
        """
        Set server pool members.

        @param      ip_list - ip list of pool members
        @return     None. Raises an exception on failure.
        """
        state = get_membership_state()
        if state in (MEMBERSHIP_STATE_UNOWNED, MEMBERSHIP_STATE_OWNED):
            raise Exception("Cannot set pool member ip list "
                            "in membership state: %s" % state)
        write_item(LOCAL_SERVER_DB, "pool_member_ip_list", ip_list)
        if is_clustered():
            write_item("server_pool", "pool_member_ip_list", ip_list,
                       get_cluster_db_home())
            try:
                db_home = get_cluster_db_home()
                servers = dump_db("server_pool_servers", db_home)
                for (host, info) in servers.items():
                    ip_addr = info['registered_ip']
                    if ip_addr not in ip_list:
                        logging.debug("set_pool_member_ip_list: " 
                                      "IP %s not in ip_list, removing it", ip_addr)
                        delete_item("server_pool_servers", host, db_home)
                        
            except Exception, e:
                logging.debug("set_pool_member_ip_list:"
                              "failed to check server_pool_servers db, error %s", e)


def take_ownership(config):
    manager_uuid = config.get('manager_uuid')
    if not manager_uuid:
        raise Exception('Manager UUID is required.')
    current_manager_uuid = get_owning_manager_uuid()
    if current_manager_uuid and current_manager_uuid != manager_uuid:
        raise Exception('Server is already owned by manager: %s' % current_manager_uuid)
    manager_event_url = config.get('manager_event_url')
    if not manager_event_url:
        raise Exception('Manager event url is required.')
    manager_statistic_url = config.get('manager_statistic_url')
    if not manager_statistic_url:
        raise Exception('Manager statistic url is required.')
    manager_certificate = config.get('manager_certificate')
    if not manager_certificate:
        raise Exception('Manager certificate is required.')
    signed_server_certificate = config.get('signed_server_certificate')
    if not signed_server_certificate:
        raise Exception('Signed server certificate is required.')
    manager_ip = agent.lib.constants.MANAGER_SRC_IP
    logging.debug('Manager IP: %s', agent.lib.constants.MANAGER_SRC_IP)
    info = {
        'manager_uuid': manager_uuid,
        'manager_ip': manager_ip,
        'manager_event_url': re.sub(r'0\.0\.0\.0', manager_ip, manager_event_url),
        'manager_statistic_url': re.sub(r'0\.0\.0\.0', manager_ip, manager_statistic_url),
    }
    write_signed_server_certificate(signed_server_certificate)
    update_db(LOCAL_SERVER_DB, info)
    reload_notification_server()
    reset_console_access()
    old_manager_certificate = get_manager_certificate()
    if manager_certificate.strip() != old_manager_certificate.strip():
        write_manager_certificate(manager_certificate)
        restart_xmlrpc_server()


def release_ownership(manager_uuid):
    state = get_membership_state()
    if state != MEMBERSHIP_STATE_OWNED:
        raise Exception('Cannot release server in membership state: %s' % state)
    current_manager_uuid = get_owning_manager_uuid()
    if not current_manager_uuid:
        raise Exception('Server is not owned.')
    if current_manager_uuid != manager_uuid:
        raise Exception('Server is owned by manager: %s' % current_manager_uuid)
    remove_manager_certificate()
    remove_signed_server_certificate()
    truncate_db(LOCAL_SERVER_DB)
    reload_notification_server()
    reset_console_access()
    restart_xmlrpc_server()


def pool_add_server(pool_alias, pool_uuid, pool_virtual_ip,
        node_number, server_hostname, server_ip, roles, initial=False):
    """
    Add this server as a pool member, possibly as the initial master

    @param      pool_alias - human readable name for the pool
    @param      pool_uuid - globally unique id for the pool
    @param      pool_virtual_ip - server pool virtual IP
    @param      node_number -
                node number in server pool and ocfs2 cluster
    @param      server_hostname -
                hostname manager uses to talk to this server
    @param      server_ip -
                ip that the manager uses to talk to this server
    @param      roles - text string with list of roles
    @param      initial -
                True if this is the initial pool member (and master)

    @return     None.  Raises an exception on failure.
    """
    role_set = validate_roles(roles)

    if not get_owning_manager_uuid():
        raise Exception("Server cannot be pooled: not owned by a manager")

    # Enforce the manager is using the server's actual (kernel) hostname
    # This will matter for cluster setup (and maybe in other places??)
    try:
        actual_hostname = socket.gethostname()
    except Exception, e:
        raise Exception("Unable to obtain hostname: %s" % e)

    if not actual_hostname:
        raise Exception("Server cannot be pooled: null hostname")

    if server_hostname != actual_hostname:
        raise Exception("Server's actual hostname is: %s" % actual_hostname)

    ip = socket.gethostbyname(server_hostname)
    if server_ip != ip:
        raise Exception("Invalid hostname resolution: %s -> %s"
                                        % (server_hostname, ip))
    uuid = get_server_pool_uuid()
    if uuid:
        raise Exception("Server already a member of pool: %s" % uuid)

    server_id = get_server_unique_id()
    if not server_id:
        raise Exception("Cannot get Server unique id.")

    d = {
        "pool_alias": pool_alias,
        "pool_uuid": pool_uuid,
        "roles": role_set,
        "registered_hostname": server_hostname,
        "registered_ip": server_ip,
        "node_number": node_number,
        "is_master": False}

    try:
        update_db(LOCAL_SERVER_DB, d)
    except Exception, e:
        raise Exception("Unable to update database: %s" % e)


def pool_remove_server(pool_uuid, final=False):
    """
    Unjoin the server from the server pool leaving no record that the
    server was in a pool

    @param      pool_uuid -
                globally unique ID of the pool this server is in
    @param      final -
                True if this is the final server (i.e. final master)

    @return     None.  Raises an exception on failure.
    """
    # It's up to the manager to know there are no other server in the pool
    if get_membership_state() != MEMBERSHIP_STATE_POOLED:
        raise Exception("Server must be in Pooled state")

    if get_server_pool_uuid() != pool_uuid:
        raise Exception("Server not in specified pool")

    d = {
        "pool_alias": None,
        "pool_uuid": None,
        "registered_hostname": None,
        "registered_ip": None,
        "node_number": None,
        "is_master": None,
        "pool_master_ip": None,
        "pool_member_ip_list": None}

    try:
        update_db(LOCAL_SERVER_DB, d)
    except Exception, e:
        raise Exception("Unable to update database: %s" % e)


def get_master_flag():
    return read_item(LOCAL_SERVER_DB, "is_master")


def get_current_master_url():
    """
    Returns the URL used for making XMLRPC calls to the current pool
    master. Before the server is clustered, this comes from the local
    server database. Once we are clustered, return the cluster wide
    value that tracks pool master failover.

    @return     Current master's URL
    """
    state = get_membership_state()
    if state in (MEMBERSHIP_STATE_UNOWNED, MEMBERSHIP_STATE_OWNED):
        return ""

    if is_clustered():
        pool_master_ip = read_item("server_pool", "pool_master_ip",
                                    get_cluster_db_home())
    else:
        return ""

    if pool_master_ip:
        return get_agt_url(pool_master_ip)
    else:
        return ""


def add_clustered_info(doc):
    """
    Add XML for server pool members and virtual IP if clustered
    """
    d = dump_db("server_pool", get_cluster_db_home())
    pool_alias = d.get("pool_alias")
    member_list = d.get("pool_member_ip_list")

    if pool_alias:
        doc.add("Pool_Alias", pool_alias)

    doc.push("Member_List")

    for ip in member_list:
        doc.push("Member")
        doc.add("Registered_IP", ip)
        doc.pop() # Member

    doc.pop() # Member_List

def add_pooled_info(doc):
    """
    Add XML for server pool members that we know about when not yet clustered
    """
    d = dump_db(LOCAL_SERVER_DB)
    pool_alias = d.get("pool_alias")
    member_list = d.get("pool_member_ip_list")

    if pool_alias:
        doc.add("Pool_Alias", pool_alias)

    doc.push("Member_List")

    if member_list:
        for srv in member_list:
            doc.push("Member")
            doc.add("Registered_IP", srv)
            doc.pop() # Member

    doc.pop() # Member_List

def add_poolfs_info(doc):
    """Add XML for the pool filesystem if available."""
    d = dump_db(LOCAL_SERVER_DB)
    poolfs_type = d.get("poolfs_type")
    poolfs_target = d.get("poolfs_target")
    poolfs_uuid = d.get("poolfs_uuid")
    poolfs_nfsbase_uuid = d.get("poolfs_nfsbase_uuid")

    if poolfs_type:
        doc.push("Pool_Filesystem")
        doc.add("Pool_Filesystem_Type", poolfs_type)
        doc.add("Pool_Filesystem_Target", poolfs_target)
        doc.add("Pool_Filesystem_Uuid", poolfs_uuid)
        doc.add("Pool_Filesystem_Nfsbase_Uuid", poolfs_nfsbase_uuid)
        doc.pop() # Pool_Filesystem
