# Copyright (C) 2006-2016 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.

import socket
import logging
from multiprocessing import Pool
from time import sleep
from random import shuffle
from agent.utils import xmlrpclib
from agent.lib.constants import SLEEP_ON_ERR
from agent.lib.config import get_config
from agent.lib.daemon import init_log
from agent.lib.db import dump_db, read_item, get_cluster_db_home
from agent.lib.settings import get_agt_url, is_master, is_clustered
from agent.target.hypervisor import is_vm_running, vm_exists

LAST_VMS = set()


def list_server_vms(host):
    try:
        sp = xmlrpclib.ServerProxy(get_agt_url(host))
        vms = sp.list_vms()
        dom0_uuid = "00000000-0000-0000-0000-000000000000"
        return [vm["uuid"] for vm in vms if vm["uuid"] != dom0_uuid]
    except Exception, e:
        return 'Failed to get VM list on %s: %s' % (host, e)


def get_live_vms():
    hosts = [host['registered_ip'] for host in
             dump_db("server_pool_servers", get_cluster_db_home()).values()]
    pool = Pool(processes=len(hosts))
    results = pool.map(list_server_vms, hosts)
    pool.close()
    pool.join()
    live_vms = []
    for result in results:
        if isinstance(result, list):
            live_vms.extend(result)
        else:
            logging.error(result)
    return live_vms


def get_monitored_vms():
    return dump_db("monitored_vms", get_cluster_db_home()).keys()


def ha_start_vm(uuid):
    cluster_db_home = get_cluster_db_home()
    vm = read_item("monitored_vms", uuid, cluster_db_home)
    repo_id = vm['repo_id']
    vm_id = vm['vm_id']
    if not vm_exists(repo_id, vm_id):
        return "VM %s does not exist" % vm_id
    hosts = [host['registered_ip'] for host in
             dump_db("server_pool_servers", cluster_db_home).values()]
    shuffle(hosts)
    msg = ""
    for host in hosts:
        try:
            if is_vm_running(repo_id, vm_id):
                return "VM %s already started" % vm_id
            sp = xmlrpclib.ServerProxy(get_agt_url(host))
            sp.start_vm(repo_id, vm_id)
            return "VM %s started on %s" % (vm_id, host)
        except Exception, e:
            msg += "Failed to start VM %s on %s: %s " % (vm_id, host, e)
    return "Cannot start VM %s: all attempts failed on %s %s" % (vm_id,
                                                                 hosts, msg)


def apply_ha_policies(monitored_vms, live_vms):
    global LAST_VMS
    vms = set(monitored_vms) - set(live_vms)
    if vms:
        if vms != LAST_VMS:
            logging.debug("Monitored VMs: %s", monitored_vms)
            logging.debug("Live VMs: %s", live_vms)
        LAST_VMS = vms
        pool = Pool(len(vms))
        results = pool.map(ha_start_vm, vms)
        pool.close()
        pool.join()
        for result in results:
            if result:
                logging.debug(result)


def serve_forever():
    logging.info("HA SERVER STARTED")
    config = get_config()
    ha_interval = int(config.get("ha", "interval"))
    ha_timeout = int(config.get("ha", "timeout"))
    # set socket default timeout for all connections
    # which prevents xmlrpc from hanging infinitely
    socket.setdefaulttimeout(ha_timeout)
    while True:
        try:
            sleep(ha_interval)
            if is_clustered() and is_master():
                live_vms = get_live_vms()
                monitored_vms = get_monitored_vms()
                apply_ha_policies(monitored_vms, live_vms)

        except Exception, e:
            logging.exception("Error in HA process: %s", e)
            sleep(SLEEP_ON_ERR)


if __name__ == '__main__':
    init_log()
    serve_forever()
