# Copyright (C) 2006-2016 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.

import inspect
import re
import struct
import logging
import logging.handlers
import pickle
import SocketServer
import select
import sys

from agent.lib.config import get_config
from agent.lib.daemon import ignore_eintr


def mask_log(obj):
    if isinstance(obj, str):
        return re.sub(r'://(\S+?):(\S*?)@(\S+?)', r'://\1:******@\3', obj)
    elif isinstance(obj, list):
        return [mask_log(x) for x in obj]
    elif isinstance(obj, tuple):
        return tuple([mask_log(x) for x in obj])
    elif isinstance(obj, set):
        return set([mask_log(x) for x in obj])
    elif isinstance(obj, frozenset):
        return frozenset([mask_log(x) for x in obj])
    elif isinstance(obj, dict):
        return dict([(mask_log(k), mask_log(v)) for k, v in obj.items()])
    else:
        return obj


class LogRecordStreamHandler(SocketServer.StreamRequestHandler):
    """Handler for a streaming logging request.

    This basically logs the record using whatever logging policy is
    configured locally.
    """

    def handle(self):
        """
        Handle multiple requests - each expected to be a 4-byte length,
        followed by the LogRecord in pickle format. Logs the record
        according to whatever policy is configured locally.
        """
        while True:
            chunk = self.connection.recv(4)
            if len(chunk) < 4:
                break
            slen = struct.unpack(">L", chunk)[0]
            chunk = self.connection.recv(slen)
            while len(chunk) < slen:
                chunk = chunk + self.connection.recv(slen - len(chunk))

            try:
                obj = self.unPickle(chunk)
                record = logging.makeLogRecord(obj)
                self.handleLogRecord(record)
            except Exception, e:
                #
                # If there is any failure while handling the request,
                # we log the error and raw data of the request.
                #
                msg = "Failed to handle log request: %s\n\n%s\n\n"
                args = (str(e), re.sub(r"[^ -~\n\t]", ".", chunk))
                frame = inspect.getframeinfo(inspect.currentframe())
                record = logging.LogRecord(None, logging.ERROR,
                                           frame.filename, frame.lineno,
                                           msg, args, sys.exc_info(),
                                           frame.function)
                self.handleLogRecord(record)

    def unPickle(self, data):
        return pickle.loads(data)
                                                               
    def handleLogRecord(self, record):
        # if a name is specified, we use the named logger rather than the one
        # implied by the record.
        if self.server.logname is not None:
            name = self.server.logname
        else:
            name = record.name
        logger = logging.getLogger(name)
        # N.B. EVERY record gets logged. This is because Logger.handle
        # is normally called AFTER logger-level filtering. If you want
        # to do filtering, do it at the client end to save wasting
        # cycles and network bandwidth!
        record.msg = mask_log(record.msg)
        record.args = mask_log(record.args)
        logger.handle(record)


class OVSLogServer(SocketServer.ThreadingTCPServer):

    allow_reuse_address = 1

    def __init__(self, host='127.0.0.1',
                 port=logging.handlers.DEFAULT_TCP_LOGGING_PORT,
                 handler=LogRecordStreamHandler):
        SocketServer.ThreadingTCPServer.__init__(self, (host, port), handler)
        self.abort = 0
        self.timeout = 1
        self.logname = None

    def serve_until_stopped(self):
        abort = 0
        while not abort:
            rd, wr, ex = select.select([self.socket.fileno()], [], [],
                                       self.timeout)
            if rd:
                self.handle_request()
            abort = self.abort


def server_init_log(logfile, loglevel):
    max_bytes = 5 * 1024 * 1024
    backup_count = 5
    date_format = '%Y-%m-%d %H:%M:%S'
    file_format = ('[%(asctime)s %(process)d] %(levelname)s '
                   '(%(module)s:%(lineno)d) %(message)s')
    logger = logging.getLogger()
    logger.propagate = 0
    handler = logging.handlers.RotatingFileHandler(filename=logfile,
                                                   maxBytes=max_bytes,
                                                   backupCount=backup_count)
    handler.setFormatter(logging.Formatter(file_format, date_format))
    logger.addHandler(handler)
    logger.setLevel(loglevel)


def serve_forever():
    config = get_config()
    logfile = config.get('logger', 'log-file')
    loglevel = logging.getLevelName(config.get('logger', 'log-level'))
    server_init_log(logfile, loglevel)
    logserver = OVSLogServer()
    ignore_eintr(OVSLogServer.serve_until_stopped, logserver)


if __name__ == '__main__':
    #
    # The log daemon needs the location of Storage Connect modules
    # to handle log requests from Storage Connect code, because these
    # requests can contain references to Storage Connect modules (for
    # example Storage Connect exceptions).
    # 
    sys.path.append('/opt/storage-connect')
    serve_forever()
