# Copyright (C) 2006-2016 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.

import binascii
import logging
import os
import re
import glob
import platform
import shutil
import tarfile
import tempfile
import subprocess
import xml.dom

from xml.dom.minidom import parse, Document

from agent.lib.files import ovs_download_file, rmtree, unpack_file, get_uuid
from agent.lib.settings import (get_assembly_path, get_iso_path,
                                get_virtualdisk_path)
from agent.lib.template import (create_template, delete_template,
                                find_file, make_filename)
from agent.lib.xenvm import create_vm, delete_vm, get_vm_config
from agent.target.host import (disk_cfg_get_device, get_file_size,
                               guess_mimetype, run_cmd, set_nice)


CHECKSUM_RE = r'^(RSA-)?(?P<algorithm>SHA(1|256))\((?P<href>.+)\)= (?P<checksum>[A-Fa-f0-9]+)$'

VSSD_PROPERTIES = [
    'AutomaticRecoveryAction',
    'AutomaticShutdownAction',
    'AutomaticStartupAction',
    'AutomaticStartupActionDelay',
    'AutomaticStartupActionSequenceNumber',
    'ConfigurationDataRoot',
    'ConfigurationFile',
    'ConfigurationID',
    'CreationTime',
    'LogDataRoot',
    'RecoveryFile',
    'SnapshotDataRoot',
    'SuspendDataRoot',
    'SwapFileDataRoot',
    'VirtualSystemIdentifier',
    'VirtualSystemType',
    'Caption',
    'ChangeableType',
    'ConfigurationName',
    'Description',
    'ElementName',
    'InstanceID',
    ]

RASD_PROPERTIES = [
    'Address',
    'AddressOnParent',
    'AllocationUnits',
    'AutomaticAllocation',
    'AutomaticDeallocation',
    'Caption',
    'ChangeableType',
    'ConfigurationName',
    'ConsumerVisibility',
    'Description',
    'ElementName',
    'InstanceID',
    'Limit',
    'MappingBehavior',
    'OtherResourceType',
    'Parent',
    'PoolID',
    'Reservation',
    'ResourceSubType',
    'ResourceType',
    'VirtualQuantity',
    'VirtualQuantityUnits',
    'Weight',
    'Connection',
    'HostResource',
    ]

SASD_PROPERTIES = RASD_PROPERTIES + [
    'Access',
    'HostExtentName',
    'HostExtentNameFormat',
    'HostExtentNameNamespace',
    'HostExtentStartingAddress',
    'HostResourceBlockSize',
    'OtherHostExtentNameFormat',
    'OtherHostExtentNameNamespace',
    'VirtualResourceBlockSize',
    ]

EPASD_PROPERTIES = RASD_PROPERTIES + [
    'AllowedPriorities',
    'AllowedToReceiveMACAddresses',
    'AllowedToReceiveVLANs',
    'AllowedToTransmitMACAddresses',
    'AllowedToTransmitVLANs',
    'DefaultPortVID',
    'DefaultPriority',
    'DesiredVLANEndpointMode',
    'GroupID',
    'ManagerID',
    'NetworkPortProfileID',
    'NetworkPortProfileIDType',
    'OtherEndpointMode',
    'OtherNetworkPortProfileIDTypeInfo',
    'PortCorrelationID',
    'PortVID',
    'Promiscuous',
    'ReceiveBandwidthLimit',
    'ReceiveBandwidthReservation',
    'SourceMACFilteringEnabled',
    'VSITypeID',
    'VSITypeIDVersion',
    ]


def import_assembly(repo_id, assembly_id, url, option=None):
    set_nice()
    assembly_path = get_assembly_path(repo_id, assembly_id)
    unpacked_path = os.path.join(assembly_path, 'unpacked')
    imports_path = os.path.join(assembly_path, 'imports')
    if not os.path.exists(assembly_path):
        os.makedirs(assembly_path)

    try:
        if not os.path.exists(unpacked_path):
            os.makedirs(unpacked_path)
        if not os.path.exists(imports_path):
            os.makedirs(imports_path)
        basename = make_filename(url)
        if basename[-4:] not in ['.ovf', '.ova', '.rpm']:
            raise RuntimeError('.ovf, .ova or .rpm file is needed.')
        if basename.endswith('.rpm'):
            rpm_filename = os.path.join(assembly_path, basename)
            rpm_filename = ovs_download_file(url, rpm_filename, option, 'assembly')
            tmp_path = os.path.join(assembly_path, 'tmp')
            if not os.path.exists(tmp_path):
                os.makedirs(tmp_path)
            unpack_file(rpm_filename, tmp_path, 'assembly')
            ovf_files = find_file('*.ovf', tmp_path)
            if ovf_files:
                move_files(tmp_path, unpacked_path)
            else:
                ova_files = find_file('*.ova', tmp_path)
                if not ova_files:
                    raise RuntimeError('Cannot find OVF or OVA file.')
                ova_filename = ova_files[0]
                check_ova(ova_filename)
                unpack_file(ova_filename, unpacked_path, 'assembly')
            rmtree(tmp_path)
        elif basename.endswith('.ova'):
            ova_filename = os.path.join(assembly_path, 'package.ova')
            ova_filename = ovs_download_file(url, ova_filename, option, 'assembly')
            check_ova(ova_filename)
            unpack_file(ova_filename, unpacked_path, 'assembly')
        else:
            ovf_filename = os.path.join(unpacked_path, basename)
            mf_url = re.sub(r'\.ovf$', '.mf', url)
            cert_url = re.sub(r'\.ovf$', '.cert', url)
            mf_filename = re.sub(r'\.ovf$', '.mf', ovf_filename)
            cert_filename = re.sub(r'\.ovf$', '.cert', ovf_filename)
            ovs_download_file(url, ovf_filename, option)
            try:
                ovs_download_file(mf_url, mf_filename, option)
            except Exception, err:
                logging.warn('Error geting %s: %s', mf_url, err)
            try:
                ovs_download_file(cert_url, cert_filename, option)
            except Exception, err:
                logging.warn('Error geting %s: %s', cert_url, err)

        ovf_filename = find_ovf(unpacked_path)
        ovf_cfg = parseOvfFile(ovf_filename)
        obj_total = len(ovf_cfg['File'])
        obj_current = 1
        for file_info in ovf_cfg['File']:

            def download_file(file_href):
                dst_filename = os.path.join(unpacked_path, file_href)
                base_url = os.path.dirname(url)
                file_href = os.path.join(base_url, file_href)
                if not os.path.exists(dst_filename):
                    dst_dir = os.path.dirname(dst_filename)
                    if not os.path.exists(dst_dir):
                        os.makedirs(dst_dir)
                    ovs_download_file(file_href, dst_filename, option, 'assembly_disk', obj_current, obj_total)

            file_href = file_info['href']
            if '://' in file_href:
                # Do not download file/http/https reference files.
                continue
            if file_info['size'] and file_info['chunkSize']:
                chunk_number, remainder = divmod(int(file_info['size']), int(file_info['chunkSize']))
                if remainder:
                    chunk_number += 1
                for chunk in range(chunk_number):
                    download_file('%s.%09d' % (file_href, chunk))
            else:
                download_file(file_href)
            obj_current += 1

        validate_ovf(ovf_filename)
    except Exception:
        rmtree(assembly_path)
        raise


def create_assembly(repo_id, assembly_id, vms):
    if not vms:
        raise RuntimeError('Empty VM list.')
    set_nice()
    assembly_path = get_assembly_path(repo_id, assembly_id)
    unpacked_path = os.path.join(assembly_path, 'unpacked')
    imports_path = os.path.join(assembly_path, 'imports')
    if not os.path.exists(assembly_path):
        os.makedirs(assembly_path)
    try:
        if not os.path.exists(unpacked_path):
            os.makedirs(unpacked_path)
        if not os.path.exists(imports_path):
            os.makedirs(imports_path)
        ovf_filename = os.path.join(unpacked_path, 'package.ovf')
        cfg = {
            'File': [],
            'Disk': {'Info': 'Virtual disks used in the package.',
                     'Disk': []},
            'Network': {'Info': 'Logical networks used in the package.',
                        'Network': []},
            'VirtualSystem': [],
            'VirtualSystemCollection': [],
            'VirtualHardware': [],
        }
        vsc_id = ''
        if len(vms) > 1:
            vsc_id = get_uuid()
            vsc = {'id': vsc_id, 'Info': 'A collection of virtual machines.'}
            cfg['VirtualSystemCollection'].append(vsc)
        for vm_repo_id, vm_id in vms:
            vm_cfg = get_vm_config(vm_repo_id, vm_id)
            vs = {
                'id': vm_cfg['uuid'],
                'Name': vm_cfg['name'],
                'Info': 'A virtual machine.',
                'VirtualSystemCollection': vsc_id,
                'Disk': [],
                'CD': [],
                'Ethernet': [],
            }
            ovm_domain_type = vm_cfg.get('OVM_domain_type')
            guest_os_type = vm_cfg.get('guest_os_type', 'default')
            builder = vm_cfg.get('builder')
            if ovm_domain_type == 'ldoms_pvm':
                vs['VirtualSystemType'] = 'DMTF:Oracle:Sparc:OracleVM_LDoms:Solaris'
            elif ovm_domain_type == 'xen_pvm':
                vs['VirtualSystemType'] = 'DMTF:Oracle:X86:OracleVM_Xen:PVM_Linux'
            elif ovm_domain_type == 'xen_hvm_pv':
                if guest_os_type == 'windows':
                    vs['VirtualSystemType'] = 'DMTF:Oracle:X86:OracleVM_Xen:PVHVM_Windows'
                elif guest_os_type == 'solaris':
                    vs['VirtualSystemType'] = 'DMTF:Oracle:X86:OracleVM_Xen:Solaris'
                else:
                    vs['VirtualSystemType'] = 'DMTF:Oracle:X86:OracleVM_Xen:PVHVM_Linux'
            elif ovm_domain_type == 'xen_hvm':
                if guest_os_type == 'windows':
                    vs['VirtualSystemType'] = 'DMTF:Oracle:X86:OracleVM_Xen:Windows'
                elif guest_os_type == 'solaris':
                    vs['VirtualSystemType'] = 'DMTF:Oracle:X86:OracleVM_Xen:Solaris'
                else:
                    vs['VirtualSystemType'] = 'DMTF:Oracle:X86:OracleVM_Xen:HVM_Linux'
            else:
                if builder == 'hvm':
                    if guest_os_type == 'windows':
                        vs['VirtualSystemType'] = 'DMTF:Oracle:X86:OracleVM_Xen:Windows'
                    elif guest_os_type == 'solaris':
                        vs['VirtualSystemType'] = 'DMTF:Oracle:X86:OracleVM_Xen:Solaris'
                    else:
                        vs['VirtualSystemType'] = 'DMTF:Oracle:X86:OracleVM_Xen:HVM_Linux'
                else:
                    vs['VirtualSystemType'] = 'DMTF:Oracle:X86:OracleVM_Xen:PVM_Linux'
            vs['Memory'] = int(vm_cfg['memory'])
            vs['Processor'] = int(vm_cfg['vcpus'])
            cfg['VirtualSystem'].append(vs)
            for disk in vm_cfg['disk']:
                backend, frontend, mode = disk.split(',')
                if frontend.endswith(':cdrom'):
                    # ignore attached media and only add empty cdrom
                    vs_disk_info = {'HostResource': ''}
                    vs['CD'].append(vs_disk_info)
                    continue

                src, guid, file_id = disk_cfg_get_device(backend)
                if not src or not file_id:
                    continue

                filename = file_id + '.vmdk'
                dst = os.path.join(unpacked_path, filename)
                if not os.path.exists(dst):
                    run_cmd(['vbox-img', 'convert', '--srcfilename', src,
                             '--srcformat', 'raw', '--dstfilename', dst,
                             '--dstformat', 'vmdk', '--variant', 'stream'])
                filesize = str(get_file_size(dst))
                file_info = {
                    'id': guid,
                    'href': filename,
                    'size': filesize,
                }
                cfg['File'].append(file_info)
                diskid = get_uuid()
                disk_info = {
                    'diskId': diskid,
                    'fileRef': guid,
                    'capacity': filesize,
                    'format': 'http://www.vmware.com/interfaces/specifications/vmdk.html#streamOptimized',
                }
                cfg['Disk']['Disk'].append(disk_info)
                vs_disk_info = {'HostResource': 'ovf:/disk/%s' % diskid}
                vs['Disk'].append(vs_disk_info)
            for vif in vm_cfg['vif']:
                # FIXME: how to map EPASD to OVM network?
                vs['Ethernet'].append({})
        createOvfFile(ovf_filename, cfg)
    except Exception:
        rmtree(assembly_path)
        raise


def deploy_assembly(repo_id, assembly_id, to_deploy=None, target_repo_id=None, option=None, to_template=False):
    set_nice()
    assembly_path = get_assembly_path(repo_id, assembly_id)
    unpacked_path = os.path.join(assembly_path, 'unpacked')
    imports_path = os.path.join(assembly_path, 'imports')
    ovf_filename = find_ovf(unpacked_path)
    ovf_cfg = parseOvfFile(ovf_filename)

    if to_deploy is None:
        to_deploy = {}
        for vs_info in ovf_cfg['VirtualSystem']:
            to_deploy[vs_info['id']] = get_uuid()

    if target_repo_id is None:
        target_repo_id = repo_id

    if not os.path.exists(imports_path):
        os.makedirs(imports_path)

    obj_total = len(ovf_cfg['File'])
    obj_current = 1
    for file_info in ovf_cfg['File']:

        def download_file(file_href):
            dst_filename = os.path.join(imports_path, os.path.basename(file_href))
            if not os.path.exists(dst_filename):
                dst_dir = os.path.dirname(dst_filename)
                if not os.path.exists(dst_dir):
                    os.makedirs(dst_dir)
                ovs_download_file(file_href, dst_filename, option, 'assembly_disk', obj_current, obj_total)

        file_href = file_info['href']
        if '://' in file_href:
            # Only download file/http/https reference files.
            if file_info['size'] and file_info['chunkSize']:
                chunk_number, remainder = divmod(int(file_info['size']), int(file_info['chunkSize']))
                if remainder:
                    chunk_number += 1
                for chunk in range(chunk_number):
                    download_file('%s.%09d' % (file_href, chunk))
            else:
                download_file(file_href)
            obj_current += 1

    virtualdisk_to_deploy = {}
    iso_to_deploy = {}
    for vs_id, vm_id in to_deploy.items():
        vs_info = None
        for vs in ovf_cfg['VirtualSystem']:
            if vs['id'] == vs_id:
                vs_info = vs
                break
        if not vs_info:
            raise RuntimeError('Cannot find virtual system with id: %s' % vs_id)
        for disk in vs_info['Disk']:
            (resource_type, resource_id) = parseHostResource(disk['HostResource'])
            if resource_type in ['file', 'disk']:
                virtualdisk_to_deploy[resource_id] = ('%s.img' % get_uuid(), resource_type)
        for disk in vs_info['CD']:
            (resource_type, resource_id) = parseHostResource(disk['HostResource'])
            if resource_type in ['file', 'disk']:
                iso_to_deploy[resource_id] = ('%s.iso' % get_uuid(), resource_type)

    virtualdisk_deployed = {}
    iso_deployed = {}
    vm_deployed = {}
    try:
        virtualdisk_deployed = deploy_disk_images(ovf_filename, target_repo_id,
                                                  virtualdisk_to_deploy,
                                                  get_virtualdisk_path)

        iso_deployed = deploy_disk_images(ovf_filename, target_repo_id,
                                          iso_to_deploy, get_iso_path)

        vm_deployed = deploy_xen_assembly(ovf_filename, virtualdisk_deployed,
                                          iso_deployed, to_deploy,
                                          target_repo_id, to_template)

        # revert map: VM, virtual disk and ISO ID as key and origination as value
        result = {
            'VirtualDisk': dict((v, k) for k, v in virtualdisk_deployed.iteritems()),
            'ISO': dict((v, k) for k, v in iso_deployed.iteritems()),
        }
        if to_template:
            result['Template'] = dict((v, k) for k, v in vm_deployed.iteritems())
        else:
            result['VM'] = dict((v, k) for k, v in vm_deployed.iteritems())
        return result
    except Exception, e:
        for virtualdisk_id in virtualdisk_deployed.values():
            virtualdisk_path = get_virtualdisk_path(target_repo_id, virtualdisk_id)
            try:
                os.unlink(virtualdisk_path)
            except Exception:
                pass
        for iso_id in iso_deployed.values():
            iso_path = get_iso_path(target_repo_id, iso_id)
            try:
                os.unlink(iso_path)
            except Exception:
                pass
        for vm_id in vm_deployed.values():
            try:
                if to_template:
                    delete_template(target_repo_id, vm_id)
                else:
                    delete_vm(target_repo_id, vm_id)
            except Exception:
                pass
        raise e


def deploy_disk_images(ovf_filename, repo_id, dskimg_to_deploy, get_path):
    dskimg_deployed = {}
    try:
        for (resource_id, (dskimg_id, resource_type)) in dskimg_to_deploy.items():
            dskimg_path = get_path(repo_id, dskimg_id)
            if resource_type == 'file':
                deployFile(ovf_filename, resource_id, dskimg_path)
            elif resource_type == 'disk':
                deployDisk(ovf_filename, resource_id, dskimg_path)
            dskimg_deployed[resource_id] = dskimg_id
    except Exception as e:
        for (resource_id, (dskimg_id, resource_type)) in dskimg_to_deploy.items():
            dskimg_path = get_path(repo_id, dskimg_id)
            try:
                os.unlink(dskimg_path)
            except Exception:
                pass
        raise e
    return dskimg_deployed


def deploy_xen_assembly(ovf_filename, virtualdisk_deployed, iso_deployed,
                        to_deploy, target_repo_id, to_template):
    vm_deployed = {}
    try:
        for vs_id, vm_id in to_deploy.items():
            vm_cfg = getDomainConfig(ovf_filename, vs_id)
            vm_cfg['name'] = vm_id
            vm_cfg['uuid'] = vm_id
            new_disk = []
            for disk in vm_cfg['disk']:
                backend, frontend, mode = disk.split(',')
                resource_type, resource_id = parseHostResource(backend)
                if frontend.endswith(':cdrom'):
                    if resource_id:
                        iso_id = iso_deployed[resource_id]
                        iso_path = get_iso_path(target_repo_id, iso_id)
                        new_disk.append('file:%s,%s,%s' % (iso_path, frontend, mode))
                    else:
                        # empty cdrom: ',hdc:cdrom,r'
                        new_disk.append(disk)
                else:
                    if not resource_id:
                        raise RuntimeError('No backend is found for %s' % frontend)
                    virtualdisk_id = virtualdisk_deployed[resource_id]
                    virtualdisk_path = get_virtualdisk_path(target_repo_id, virtualdisk_id)
                    new_disk.append('file:%s,%s,%s' % (virtualdisk_path, frontend, mode))
            vm_cfg['disk'] = new_disk
            if to_template:
                create_template(target_repo_id, vm_id, vm_cfg)
            else:
                create_vm(target_repo_id, vm_id, vm_cfg)
            vm_deployed[vs_id] = vm_id
    except Exception:
        for vm_id in vm_deployed.values():
            try:
                if to_template:
                    delete_template(target_repo_id, vm_id)
                else:
                    delete_vm(target_repo_id, vm_id)
            except Exception:
                pass
        raise
    return vm_deployed


def get_assembly_config(repo_id, assembly_id):
    assembly_path = get_assembly_path(repo_id, assembly_id)
    unpacked_path = os.path.join(assembly_path, 'unpacked')
    ovf_filename = find_ovf(unpacked_path)
    cfg = parseOvfFile(ovf_filename)
    # FIXME: backward compatibility: only return the Disk list.
    cfg['Disk'] = cfg['Disk'].get('Disk', [])
    return cfg


def get_assembly_config_xml(repo_id, assembly_id):
    assembly_path = get_assembly_path(repo_id, assembly_id)
    unpacked_path = os.path.join(assembly_path, 'unpacked')
    ovf_filename = find_ovf(unpacked_path)
    ovf_file = open(ovf_filename)
    ovf_cfg = ovf_file.read()
    ovf_file.close()
    return ovf_cfg


def set_assembly_config(repo_id, assembly_id, cfg):
    assembly_path = get_assembly_path(repo_id, assembly_id)
    unpacked_path = os.path.join(assembly_path, 'unpacked')
    ovf_filename = find_ovf(unpacked_path)
    # FIXME: backward compatibility: only receive the Disk list. Use hard coded Info.
    cfg['Disk'] = {
        'Info': 'Virtual disks used in the package.',
        'Disk': cfg['Disk'],
    }
    createOvfFile(ovf_filename, cfg)


def set_assembly_config_xml(repo_id, assembly_id, cfg):
    assembly_path = get_assembly_path(repo_id, assembly_id)
    unpacked_path = os.path.join(assembly_path, 'unpacked')
    ovf_filename = find_ovf(unpacked_path)
    ovf_file = open(ovf_filename, 'w')
    ovf_file.write(cfg)
    ovf_file.close()


def assembly_add_file(repo_id, assembly_id, url, filename, option=None):
    set_nice()
    assembly_path = get_assembly_path(repo_id, assembly_id)
    unpacked_path = os.path.join(assembly_path, 'unpacked')
    filepath = os.path.join(unpacked_path, filename)
    dirpath = os.path.dirname(filepath)
    if not os.path.exists(dirpath):
        os.makedirs(dirpath)
    ovs_download_file(url, filepath, option, 'assembly')


def assembly_del_file(repo_id, assembly_id, filename):
    assembly_path = get_assembly_path(repo_id, assembly_id)
    unpacked_path = os.path.join(assembly_path, 'unpacked')
    filepath = os.path.join(unpacked_path, filename)
    if os.path.exists(filepath):
        os.unlink(filepath)


def assembly_rename_file(repo_id, assembly_id, filename, new_filename):
    assembly_path = get_assembly_path(repo_id, assembly_id)
    unpacked_path = os.path.join(assembly_path, 'unpacked')
    filepath = os.path.join(unpacked_path, filename)
    new_filepath = os.path.join(unpacked_path, new_filename)
    new_dirpath = os.path.dirname(new_filepath)
    if not os.path.exists(new_dirpath):
        os.makedirs(new_dirpath)
    os.rename(filepath, new_filepath)


def pack_assembly(repo_id, assembly_id):
    assembly_path = get_assembly_path(repo_id, assembly_id)
    unpacked_path = os.path.join(assembly_path, 'unpacked')
    ovf_filename = find_ovf(unpacked_path)
    ova_filename = os.path.join(assembly_path, 'package.ova')
    mf_filename = re.sub(r'\.ovf$', '.mf', ovf_filename)

    create_manifest(ovf_filename)

    ova = tarfile.open(ova_filename, 'w')
    ova.add(ovf_filename, os.path.basename(ovf_filename))
    ova.add(mf_filename, os.path.basename(mf_filename))
    ovf = parseXmlFile(ovf_filename)
    files = getFile(ovf)
    for file_info in files:
        if '://' in file_info['href']:
            # Do not pack file/http/https referenced files.
            continue
        if file_info['size'] and file_info['chunkSize']:
            chunk_number, remainder = divmod(int(file_info['size']), int(file_info['chunkSize']))
            if remainder:
                chunk_number += 1
            for chunk in range(chunk_number):
                chunk_href = '%s.%09d' % (file_info['href'], chunk)
                filepath = os.path.join(unpacked_path, chunk_href)
                ova.add(filepath, file_info['href'])
        else:
            filepath = os.path.join(unpacked_path, file_info['href'])
            ova.add(filepath, file_info['href'])
    ova.close()


def unpack_assembly(repo_id, assembly_id):
    assembly_path = get_assembly_path(repo_id, assembly_id)
    unpacked_path = os.path.join(assembly_path, 'unpacked')
    ova_filename = find_ova(assembly_path)
    if not os.path.exists(unpacked_path):
        os.makedirs(unpacked_path)
    run_cmd(['gtar', 'xf', ova_filename, '-C', unpacked_path])


def delete_assembly(repo_id, assembly_id):
    assembly_path = get_assembly_path(repo_id, assembly_id)
    rmtree(assembly_path, raise_exception=True)


def create_manifest(ovf_filename, algorithm='SHA1'):
    mf_filename = re.sub(r'\.ovf$', '.mf', ovf_filename)
    unpacked_path = os.path.dirname(ovf_filename)
    mf_file = open(mf_filename, 'w')
    ovf = parseXmlFile(ovf_filename)
    files = getFile(ovf)
    for file_info in files:
        if '://' in file_info['href']:
            # Do not checksum file/http/https referenced files.
            continue
        if file_info['size'] and file_info['chunkSize']:
            chunk_number, remainder = divmod(int(file_info['size']), int(file_info['chunkSize']))
            if remainder:
                chunk_number += 1
            for chunk in range(chunk_number):
                chunk_href = '%s.%09d' % (file_info['href'], chunk)
                filepath = os.path.join(unpacked_path, chunk_href)
                checksum = get_checksum(filepath, algorithm)
                mf_file.write('%s(%s)= %s\n' % (algorithm, file_info['href'], checksum))
        else:
            filepath = os.path.join(unpacked_path, file_info['href'])
            checksum = get_checksum(filepath, algorithm)
            mf_file.write('%s(%s)= %s\n' % (algorithm, file_info['href'], checksum))
    mf_file.close()


def extract_certificate(filename):
    cmd = ['openssl', 'x509', '-inform', 'pem', '-in', filename, '-pubkey', '-noout']
    ret = run_cmd(cmd)
    tmpfd, pubkey = tempfile.mkstemp()
    os.write(tmpfd, ret)
    os.close(tmpfd)
    certificate = open(filename)
    firstline = certificate.readline()
    certificate.close()
    match = re.match(CHECKSUM_RE, firstline.strip())
    if not match:
        raise RuntimeError('Malformed certificate first line: %s' % firstline)
    algorithm = match.group('algorithm')
    digest = match.group('checksum')
    binary_digest = binascii.unhexlify(digest)
    tmpfd, signature = tempfile.mkstemp()
    os.write(tmpfd, binary_digest)
    os.close(tmpfd)
    return pubkey, algorithm, signature


def parse_manifest(filename):
    result = []
    manifest = open(filename)
    for line in manifest:
        match = re.match(CHECKSUM_RE, line.strip())
        if match:
            result.append([match.group('href'), match.group('algorithm'),
                           match.group('checksum')])
    manifest.close()
    return result


def get_checksum(filename, algorithm):
    output = run_cmd(['%ssum' % algorithm.lower(), '-b', filename])
    checksum = output.split()[0]
    return checksum


def validate_ovf(ovf_filename):
    manifest = re.sub(r'\.ovf$', '.mf', ovf_filename)
    certificate = re.sub(r'\.ovf$', '.cert', ovf_filename)
    if os.path.exists(manifest):
        parsed_manifest = parse_manifest(manifest)
        for filename, algorithm, checksum in parsed_manifest:
            filepath = os.path.join(os.path.dirname(ovf_filename), filename)
            current_checksum = get_checksum(filepath, algorithm)
            if current_checksum.lower() != checksum.lower():
                raise RuntimeError('Check file %s digest failed.' % filename)
        if os.path.exists(certificate):
            pubkey, algorithm, signature = extract_certificate(certificate)
            try:
                cmd = ['openssl', 'dgst', '-%s' % algorithm.lower(), '-verify',
                       pubkey, '-signature', signature, manifest]
                run_cmd(cmd)
            finally:
                os.unlink(pubkey)
                os.unlink(signature)


def check_ova(filename):
    if not tarfile.is_tarfile(filename):
        raise RuntimeError('Invalid OVA package: not a tarfile.')


def move_files(src, dst):
    for (path, dirs, files) in os.walk(src):
        for filename in files:
            src_path = os.path.join(path, filename)
            dst_path = os.path.join(dst, filename)
            shutil.move(src_path, dst_path)


def find_ovf(unpacked_path):
    ovf_filenames = find_file('*.ovf', unpacked_path)
    if not ovf_filenames:
        raise RuntimeError('Cannot find OVF descriptor file.')
    if len(ovf_filenames) > 1:
        raise RuntimeError('Multiple OVF descriptor files found.')
    return ovf_filenames[0]


def find_ova(assembly_path):
    ova_filenames = find_file('*.ova', assembly_path)
    if not ova_filenames:
        raise RuntimeError('No OVA package to unpack.')
    if len(ova_filenames) > 1:
        raise RuntimeError('Multiple OVA packages found.')
    return ova_filenames[0]


def pipeline(cmds, outputfile=None, success_return_codes=(0,)):
    if not cmds:
        raise RuntimeError('Empty command list.')
    output = None
    if outputfile:
        output = open(outputfile, 'w')
    try:
        proc = None
        for cmd in cmds:
            if proc:
                stdin = proc.stdout
            else:
                stdin = None
            if output and cmd == cmds[-1]:
                stdout = output
            else:
                stdout = subprocess.PIPE
            if not isinstance(cmd, list):
                raise RuntimeError('Only list is accepted as command args.')
            cmd = [str(x) for x in cmd]
            new_proc = subprocess.Popen(cmd, stdin=stdin, stdout=stdout,
                                        stderr=subprocess.PIPE, close_fds=True)
            if proc:
                proc.stdout.close()
            proc = new_proc
        (stdoutdata, stderrdata) = proc.communicate()
        if proc.returncode not in success_return_codes:
            raise RuntimeError('Command failed: cmds (%s) output file (%s) '
                      'success_return_codes %s) rc (%s) stdout (%s) stderr '
                      '(%s)' % (cmds, outputfile, success_return_codes,
                      proc.returncode, stdoutdata, stderrdata))
        return stdoutdata
    finally:
        if output:
            output.close()


def convertFile(src_filename, src_format, dst_filename, dst_format,
                src_compression=None, src_chunk_size=None, dst_size=None,
                dst_populated_size=None, dst_variant=None):
    # FIXME: add dst_size, dst_populated_size and parent support
    cmds = []
    if src_chunk_size:
        src_filename_list = sorted(glob.glob('%s.*' % src_filename))
        if not src_filename_list:
            raise RuntimeError('Could not find source files')
    else:
        if not os.path.isfile(src_filename):
            raise RuntimeError('Source file not exist or not regular file.')
        src_filename_list = [src_filename]
    cmds.append(['cat'] + src_filename_list)
    if src_compression:
        compressed_filename = src_filename_list[0]
        mimetype = guess_mimetype(compressed_filename)
        if mimetype != 'application/x-gzip':
            raise RuntimeError('Compressed source file (%s) mime type is not '
                               'gzip: %s' % (compressed_filename, mimetype))
        cmds.append(['gzip', '-dc'])
    if src_format == 'vmdk':
        cmd = ['vbox-img', 'convert', '--stdin', '--srcformat', src_format,
               '--dstformat', dst_format]
        if dst_variant:
            cmd += ['--variant', dst_variant]
        if dst_format == 'raw' and platform.system() != 'SunOS':
            cmd += ['--stdout']
            cmds.append(cmd)
            cmd = ['cp', '--sparse=always', '/dev/stdin', dst_filename]
            cmds.append(cmd)
        else:
            cmd += ['--dstfilename', dst_filename]
            cmds.append(cmd)
        pipeline(cmds)
    elif src_format in ['vdi', 'vhd']:
        (_, tmp_filename) = tempfile.mkstemp(dir=os.path.dirname(dst_filename))
        pipeline(cmds, tmp_filename, (0, 2))
        cmd = ['vbox-img', 'convert', '--srcfilename', tmp_filename,
               '--srcformat', src_format, '--dstfilename', dst_filename,
               '--dstformat', dst_format]
        if dst_variant:
            cmd += ['--variant', dst_variant]
        try:
            run_cmd(cmd)
        finally:
            os.unlink(tmp_filename)
    elif src_format == 'raw':
        pipeline(cmds, dst_filename)
    else:
        raise RuntimeError('Unsupported source file format: %s' % src_format)


def createFile(filename, fmt, size, populated_size=None, sparse=True):
    # FIXME: add format, populated_size and parent support
    dirpath = os.path.dirname(filename)
    if not os.path.exists(dirpath):
        os.makedirs(dirpath)
    fd = None
    try:
        try:
            fd = os.open(filename, os.O_WRONLY | os.O_CREAT | os.O_DSYNC)
            if sparse:
                os.ftruncate(fd, size)
            else:
                bufsize = 1024 * 1024
                for _ in range(0, size / bufsize):
                    os.write(fd, '\x00' * bufsize)
                os.write(fd, '\x00' * size % bufsize)
        except OSError, err:
            raise RuntimeError('Error creating file %s: %s' % (filename, err))
    finally:
        if fd is not None:
            os.close(fd)


def blkname(i):
    ascii_lowercase = 'abcdefghijklmnopqrstuvwxyz'
    if i < 26:
        return ascii_lowercase[i]
    return blkname(i/26 - 1) + ascii_lowercase[i%26]


def getFile(ovf):
    result = []
    for node in ovf.documentElement.childNodes:
        if node.localName != 'References':
            continue
        for fnode in node.getElementsByTagNameNS('*', 'File'):
            file_info = {
                'id': fnode.getAttribute('ovf:id'),
                'href': fnode.getAttribute('ovf:href'),
                'size': fnode.getAttribute('ovf:size'),
                'chunkSize': fnode.getAttribute('ovf:chunkSize'),
                'compression': fnode.getAttribute('ovf:compression'),
            }
            result.append(file_info)
    return result


def getDisk(ovf):
    result = {}
    for node in ovf.documentElement.childNodes:
        if node.localName != 'DiskSection':
            continue
        result['Info'] = getInfo(node)
        result['Disk'] = []
        for disk in ovf.getElementsByTagNameNS('*', 'Disk'):
            disk_info = {
                'diskId': disk.getAttribute('ovf:diskId'),
                'fileRef': disk.getAttribute('ovf:fileRef'),
                'capacity': disk.getAttribute('ovf:capacity'),
                'capacityAllocationUnits': disk.getAttribute('ovf:capacityAllocationUnits'),
                'format': disk.getAttribute('ovf:format'),
                'populatedSize': disk.getAttribute('ovf:populatedSize'),
                'parentRef': disk.getAttribute('ovf:parentRef'),
            }
            result['Disk'].append(disk_info)
    return result


def getNetwork(ovf):
    result = {}
    for node in ovf.documentElement.childNodes:
        if node.localName != 'NetworkSection':
            continue
        result['Info'] = getInfo(node)
        result['Network'] = []
        for network in node.getElementsByTagNameNS('*', 'Network'):
            network_info = {'name': network.getAttribute('ovf:name')}
            for child in network.childNodes:
                if child.localName == 'Description':
                    network_info['Description'] = getNodeValue(child)
            result['Network'].append(network_info)
    return result


def getVirtualSystemCollection(ovf):
    result = []
    for node in ovf.documentElement.childNodes:
        if node.localName != 'VirtualSystemCollection':
            continue
        vsc_info = {
            'id': node.getAttribute('ovf:id'),
            'Info': getInfo(node),
        }
        for child in node.childNodes:
            if child.localName == 'Name':
                vsc_info['Name'] = getNodeValue(child)
        result.append(vsc_info)
    return result


def getAllocationUnits(units):
    '''DSP0004 v2.5.0 outlines the Programmatic Unit forms for OVF. This
    pertains specifically to rasd:AllocationUnits, which accepts both the
    current and deprecated forms. New implementations should not use Unit
    Qualifiers as this form is deprecated.
      - PUnit form, as in 'byte * 2^20'
      - PUnit form w/ Units Qualifier(deprecated), as in 'MegaBytes'
    '''
    units = units.lower().strip()
    if units == 'kb':
        return 2 ** 10
    elif units == 'mb':
        return 2 ** 20
    elif units == 'gb':
        return 2 ** 30
    elif units == 'tb':
        return 2 ** 40

    if units.startswith('byte') or units.startswith('bit'):
        # FIXME: support hertz * 10^6 ?
        units = units.replace('^', '**')
        units = units.replace('byte', '2 ** 0')
        units = units.replace('bit', '2 ** -3')
        factor = int(eval(units))
        return factor

    if units.startswith('deca'):
        factor = 10 ** 1
    elif units.startswith('hecto'):
        factor = 10 ** 2
    elif units.startswith('kilo'):
        # FIXME: usually mean 2 ** 10 instead of 10 ** 3
        factor = 2 ** 10
    elif units.startswith('mega'):
        # FIXME: usually mean 2 ** 20 instead of 10 ** 6
        factor = 2 ** 20
    elif units.startswith('giga'):
        # FIXME: usually mean 2 ** 30 instead of 10 ** 9
        factor = 2 ** 30
    elif units.startswith('tera'):
        # FIXME: usually mean 2 ** 40 instead of 10 ** 12
        factor = 2 ** 40
    elif units.startswith('kibi'):
        factor = 2 ** 10
    elif units.startswith('mebi'):
        factor = 2 ** 20
    elif units.startswith('gibi'):
        factor = 2 ** 30
    elif units.startswith('tebi'):
        factor = 2 ** 40
    else:
        raise ValueError('Incompatible PUnit quantifier: %s' % units)

    if units.endswith('bytes'):
        factor *= 1
    elif units.endswith('bits'):
        factor *= 2 ** -3
    else:
        raise ValueError('Incompatible PUnit quantifier: %s' % units)

    return factor


def getVirtualSystemMemory(items):
    max_memory = 0
    memory = 0
    for item in items:
        if item.get('ResourceType') == '4':
            quantity = int(item['VirtualQuantity'])
            units = getAllocationUnits(item['AllocationUnits']) / 1024.0 / 1024.0
            if item.get('bound') == 'max':
                max_memory = int(quantity * units)
            else:
                memory = int(quantity * units)
    if not memory:
        if max_memory:
            memory = max_memory
        else:
            memory = 512
    return memory, max_memory


def getVirtualSystemProcessor(items):
    processor = 0
    max_processor = 0
    for item in items:
        if item.get('ResourceType') == '3':
            virtual_processor = int(item['VirtualQuantity'])
            if item.get('bound') == 'max':
                max_processor = virtual_processor
            else:
                processor = virtual_processor
    if not processor:
        if max_processor:
            processor = max_processor
        else:
            processor = 1
    return processor, max_processor


def getVirtualSystemBusType(items, instance_id):
    bus = 'ide'
    if instance_id:
        for item in items:
            if (item['InstanceID'] == instance_id and
                item['ResourceType'] in ['6', '7', '8', '9', '20']):
                bus = 'scsi'
    return bus


def getVirtualSystemDisk(items):
    result = []
    for item in items:
        disk_info = {}
        if not item.get('ResourceType') in ['17', '19']:
            continue
        disk_info['HostResource'] = item.get('HostResource', '')
        disk_info['bus'] = getVirtualSystemBusType(items, item.get('Parent'))
        result.append(disk_info)
    return result


def getVirtualSystemCD(items):
    result = []
    for item in items:
        cd_info = {}
        if not item.get('ResourceType') in ['15', '16']:
            continue
        cd_info['HostResource'] = item.get('HostResource', '')
        cd_info['bus'] = getVirtualSystemBusType(items, item.get('Parent'))
        result.append(cd_info)
    return result


def getVirtualSystemEthernet(items):
    result = []
    for item in items:
        if item['ResourceType'] == '10':
            # FIXME: how to map EPASD to OVM network?
            result.append({})
    return result


def getVirtualSystemType(system):
    vstype = ''
    if system:
        vstype = system[0].get('VirtualSystemType', '')
    return vstype


def getVirtualSystemInfo(node):
    vs_info = {
        'id': node.getAttribute('ovf:id'),
        'Info': getInfo(node),
    }
    for child in node.childNodes:
        if child.localName == 'Name':
            vs_info['Name'] = getNodeValue(child)
        if child.localName == 'VirtualHardwareSection':
            vhd = getVirtualHardwareInfo(child)
            systems = vhd['System']
            items = vhd['Item']
            storage_items = vhd['StorageItem']
            ethernet_port_items = vhd['EthernetPortItem']
            vs_info['VirtualSystemType'] = getVirtualSystemType(systems)
            vs_info['Memory'], maxMemory = getVirtualSystemMemory(items)
            if maxMemory:
                vs_info['MaxMemory'] = maxMemory
            vs_info['Processor'], maxProcessor = getVirtualSystemProcessor(items)
            if maxProcessor:
                vs_info['MaxProcessor'] = maxProcessor
            vs_info['Disk'] = getVirtualSystemDisk(items + storage_items)
            vs_info['CD'] = getVirtualSystemCD(items + storage_items)
            vs_info['Ethernet'] = getVirtualSystemEthernet(items + ethernet_port_items)
    return vs_info


def getVirtualSystem(ovf):
    result = []
    for node in ovf.documentElement.childNodes:
        if node.localName == 'VirtualSystem':
            vs_info = getVirtualSystemInfo(node)
            vs_info['VirtualSystemCollection'] = ''
            result.append(vs_info)
        if node.localName == 'VirtualSystemCollection':
            for vs in node.childNodes:
                if vs.localName == 'VirtualSystem':
                    vs_info = getVirtualSystemInfo(vs)
                    vs_info['VirtualSystemCollection'] = node.getAttribute('ovf:id')
                    result.append(vs_info)
    return result


def parseSettingData(node, properties):
    sd = {}
    for subnode in node.childNodes:
        for prop in properties:
            if subnode.localName == prop:
                sd[prop] = getNodeValue(subnode)
    return sd


def getVirtualHardwareInfo(node):
    vhd_info = {
        'VirtualSystem': '',
        'id': node.getAttribute('ovf:id'),
        'transport': node.getAttribute('ovf:transport'),
        'Info': getInfo(node),
        'System': [],
        'Item': [],
        'StorageItem': [],
        'EthernetPortItem': [],
    }
    for child in node.childNodes:
        if child.localName == 'System':
            sd = parseSettingData(child, VSSD_PROPERTIES)
            vhd_info['System'].append(sd)
        if child.localName == 'Item':
            sd = parseSettingData(child, RASD_PROPERTIES)
            if child.getAttribute('ovf:bound'):
                sd['bound'] = child.getAttribute('ovf:bound')
            vhd_info['Item'].append(sd)
        if child.localName == 'StorageItem':
            sd = parseSettingData(child, SASD_PROPERTIES)
            vhd_info['StorageItem'].append(sd)
        if child.localName == 'EthernetPortItem':
            sd = parseSettingData(child, EPASD_PROPERTIES)
            vhd_info['EthernetPortItem'].append(sd)
    return vhd_info


def getVirtualHardware(ovf):
    result = []
    for vs in ovf.getElementsByTagNameNS('*', 'VirtualSystem'):
        for node in vs.childNodes:
            if node.localName == 'VirtualHardwareSection':
                vhd_info = getVirtualHardwareInfo(node)
                vhd_info['VirtualSystem'] = node.getAttribute('ovf:id')
                result.append(vhd_info)
    return result


def getProductInfo(node):
    product_info = {
        'VirtualSystem': '',
        'VirtualSystemCollection': '',
        'class': node.getAttribute('ovf:class'),
        'instance': node.getAttribute('ovf:instance'),
        'Info': getInfo(node),
        'Property': [],
    }
    for child in node.childNodes:
        if child.localName == 'Product':
            product_info['Product'] = getNodeValue(child)
        if child.localName == 'vendor':
            product_info['Vendor'] = getNodeValue(child)
        if child.localName == 'version':
            product_info['Version'] = getNodeValue(child)
        if child.localName == 'FullVersion':
            product_info['FullVersion'] = getNodeValue(child)
        if child.localName == 'ProductUrl':
            product_info['ProductUrl'] = getNodeValue(child)
        if child.localName == 'VendorUrl':
            product_info['VendorUrl'] = getNodeValue(child)
        if child.localName == 'AppUrl':
            product_info['AppUrl'] = getNodeValue(child)
        # FIXME: add <Icon> support.
        # FIXME: <Category> has design problem. Don't support it.
        if child.localName == 'Property':
            prop = {
                'key': child.getAttribute('ovf:key'),
                'type': child.getAttribute('ovf:type'),
                'qualifiers': child.getAttribute('ovf:qualifiers'),
                'userConfigurable': child.getAttribute('ovf:userConfigurable'),
                'value': child.getAttribute('ovf:value'),
                'password': child.getAttribute('ovf:password'),
            }
            for subnode in child.childNodes:
                if subnode.localName == 'Label':
                    prop['Label'] = getNodeValue(subnode)
                if subnode.localName == 'Description':
                    prop['Description'] = getNodeValue(subnode)
                # FIXME: add <Value> support.
            product_info['Property'].append(prop)
    return product_info


def getProduct(ovf):
    result = []
    for vsc in ovf.getElementsByTagNameNS('*', 'VirtualSystemCollection'):
        for node in vsc.childNodes:
            if node.localName == 'ProductSection':
                product_info = getProductInfo(node)
                product_info['VirtualSystemCollection'] = vsc.getAttribute('ovf:id')
                result.append(product_info)
    for vs in ovf.getElementsByTagNameNS('*', 'VirtualSystem'):
        for node in vs.childNodes:
            if node.localName == 'ProductSection':
                product_info = getProductInfo(node)
                product_info['VirtualSystem'] = vs.getAttribute('ovf:id')
                result.append(product_info)
    return result


def parseXmlFile(ovf_filename):
    ovf = parse(ovf_filename)
    ovf.normalize()
    return ovf


def parseOvfFile(ovf_filename):
    cfg = {}
    ovf = parseXmlFile(ovf_filename)
    cfg['File'] = getFile(ovf)
    cfg['Disk'] = getDisk(ovf)
    cfg['Network'] = getNetwork(ovf)
    cfg['VirtualSystemCollection'] = getVirtualSystemCollection(ovf)
    cfg['VirtualSystem'] = getVirtualSystem(ovf)
    cfg['Product'] = getProduct(ovf)
    return cfg


def getNodeValue(node):
    for child in node.childNodes:
        if child.nodeType == xml.dom.Node.TEXT_NODE:
            return child.nodeValue.strip()


def getInfo(node):
    nodes = node.getElementsByTagNameNS('*', 'Info')
    if nodes:
        return getNodeValue(nodes[0])
    else:
        return ''


def getNodeWithId(ovf, tag, ident):
    for node in ovf.getElementsByTagNameNS('*', tag):
        if node.getAttribute('ovf:id') == ident:
            return node


def createTextNode(ovf, node, name, text):
    new_node = ovf.createElement(name)
    new_text_node = ovf.createTextNode(text)
    new_node.appendChild(new_text_node)
    node.appendChild(new_node)


def createEnvelope(ovf, version='1.1.0', lang='en-US'):
    node = ovf.createElement('Envelope')
    node.setAttribute('xmlns:xsi', 'http://www.w3.org/2001/XMLSchema-instance')
    node.setAttribute('xmlns:ovf', 'http://schemas.dmtf.org/ovf/envelope/1')
    node.setAttribute('xmlns:ovfstr', 'http://schema.dmtf.org/ovf/strings/1')
    node.setAttribute('xmlns:vssd', 'http://schemas.dmtf.org/wbem/wscim/1/'
                                    'cim-schema/2/CIM_VirtualSystemSettingData')
    node.setAttribute('xmlns:rasd', 'http://schemas.dmtf.org/wbem/wscim/1/'
                                    'cim-schema/2/CIM_ResourceAllocationSettingData')
    node.setAttribute('ovf:version', version)
    node.setAttribute('xml:lang', lang)
    node.setAttribute('xmlns:ovm', 'http://schemas.oracle.com/ovf/1')
    node.setAttribute('xmlns', 'http://schemas.dmtf.org/ovf/envelope/1')
    ovf.appendChild(node)


def createReferences(ovf, files):
    '''DSP8023_1.1.0:
       - ovf:id and ovf:href are required for <File>.
    '''
    node = ovf.createElement('References')
    for file_info in files:
        file_node = ovf.createElement('File')
        file_node.setAttribute('ovf:id', file_info['id'])
        file_node.setAttribute('ovf:href', file_info['href'])
        if file_info.get('size'):
            file_node.setAttribute('ovf:size', file_info.get('size'))
        if file_info.get('compression'):
            file_node.setAttribute('ovf:compression', file_info.get('compression'))
        if file_info.get('chunkSize'):
            file_node.setAttribute('ovf:chunkSize', file_info.get('chunkSize'))
        node.appendChild(file_node)
    ovf.documentElement.appendChild(node)


def createDiskSection(ovf, disks):
    '''DSP8023_1.1.0:
       - <Info> is required for <DiskSection>.
       - ovf:diskId and ovf:capacity are required for <Disk>.
    '''
    if not disks:
        return
    node = ovf.createElement('DiskSection')
    createTextNode(ovf, node, 'Info', disks['Info'])
    for disk in disks['Disk']:
        disk_node = ovf.createElement('Disk')
        disk_node.setAttribute('ovf:diskId', disk['diskId'])
        disk_node.setAttribute('ovf:capacity', disk['capacity'])
        if disk.get('fileRef'):
            disk_node.setAttribute('ovf:fileRef', disk.get('fileRef'))
        if disk.get('capacityAllocationUnits'):
            disk_node.setAttribute('ovf:capacityAllocationUnits', disk.get('capacityAllocationUnits'))
        if disk.get('format'):
            disk_node.setAttribute('ovf:format', disk.get('format'))
        if disk.get('populatedSize'):
            disk_node.setAttribute('ovf:populatedSize', disk.get('populatedSize'))
        if disk.get('parentRef'):
            disk_node.setAttribute('ovf:parentRef', disk.get('parentRef'))
        node.appendChild(disk_node)
    ovf.documentElement.appendChild(node)


def createNetworkSection(ovf, networks):
    '''DSP8023_1.1.0:
       - <Info> is required for <NetworkSection>.
       - ovf:name is required for <Network>.
       - <Description> is optional for <Network>.
    '''
    if not networks:
        return
    node = ovf.createElement('NetworkSection')
    createTextNode(ovf, node, 'Info', networks['Info'])
    for network in networks['Network']:
        network_node = ovf.createElement('Network')
        network_node.setAttribute('ovf:name', network['name'])
        if network.get('Description'):
            createTextNode(ovf, network_node, 'Description', network.get('Description'))
        node.appendChild(network_node)
    ovf.documentElement.appendChild(node)


def createVirtualSystemCollection(ovf, vscs):
    '''DSP8023_1.1.0:
       - ovf:id is required for <VirtualSystemCollection>.
       - <Info> is required for <VirtualSystemCollection>.
       - <Name> is optional for <VirtualSystemCollection>.
    '''
    for vsc in vscs:
        vsc_node = ovf.createElement('VirtualSystemCollection')
        vsc_node.setAttribute('ovf:id', vsc['id'])
        createTextNode(ovf, vsc_node, 'Info', vsc['Info'])
        if vsc.get('Name'):
            createTextNode(ovf, vsc_node, 'Name', vsc.get('Name'))
        ovf.documentElement.appendChild(vsc_node)


def createVirtualSystem(ovf, vss):
    '''DSP8023_1.1.0:
       - ovf:id is required for <VirtualSystem>.
       - <Info> is required for <VirtualSystem>.
       - <Name> is optional for <VirtualSystem>.
    '''
    for vs in vss:
        vs_node = ovf.createElement('VirtualSystem')
        vs_node.setAttribute('ovf:id', vs['id'])
        createTextNode(ovf, vs_node, 'Info', vs['Info'])
        if vs.get('Name'):
            createTextNode(ovf, vs_node, 'Name', vs.get('Name'))
        vsc = vs['VirtualSystemCollection']
        if vsc:
            parent = getNodeWithId(ovf, 'VirtualSystemCollection', vsc)
        else:
            parent = ovf.documentElement
        parent.appendChild(vs_node)


def createVirtualHardwareSection(ovf, vhds):
    '''DSP8023_1.1.0:
       - ovf:id and ovf:transport are optional for <VirtualHardwareSection>.
       - <Info> is required for <VirtualHardwareSection>.
    '''
    for vhd in vhds:
        vhd_node = ovf.createElement('VirtualHardwareSection')
        createTextNode(ovf, vhd_node, 'Info', vhd['Info'])
        if vhd.get('id'):
            vhd_node.setAttribute('ovf:id', vhd.get('id'))
        if vhd.get('transport'):
            vhd_node.setAttribute('ovf:transport', vhd.get('transport'))
        for system in vhd['System']:
            subnode = ovf.createElement('System')
            for key, value in system.items():
                createTextNode(ovf, subnode, 'vssd:%s' % key, value)
            vhd_node.appendChild(subnode)
        for item in vhd['Item']:
            subnode = ovf.createElement('Item')
            for key, value in item.items():
                createTextNode(ovf, subnode, 'rasd:%s' % key, value)
            vhd_node.appendChild(subnode)
        vs = vhd['VirtualSystem']
        parent = getNodeWithId(ovf, 'VirtualSystem', vs)
        parent.appendChild(vhd_node)


def createProductSection(ovf, products):
    '''DSP8023_1.1.0:
       - ovf:class and ovf:instance are optional for <ProductSection>.
       - <Info> is required for <ProductSection>.
    '''
    for product in products:
        product_node = ovf.createElement('ProductSection')
        if product.get('class'):
            product_node.setAttribute('ovf:class', product.get('class'))
        if product.get('instance'):
            product_node.setAttribute('ovf:instance', product.get('instance'))
        createTextNode(ovf, product_node, 'Info', product['Info'])

        if product.get('Product'):
            createTextNode(ovf, product_node, 'Product', product.get('Product'))
        if product.get('Vendor'):
            createTextNode(ovf, product_node, 'Vendor', product.get('Vendor'))
        if product.get('Version'):
            createTextNode(ovf, product_node, 'Version', product.get('Version'))
        if product.get('FullVersion'):
            createTextNode(ovf, product_node, 'FullVersion', product.get('FullVersion'))
        if product.get('ProductUrl'):
            createTextNode(ovf, product_node, 'ProductUrl', product.get('ProductUrl'))
        if product.get('VendorUrl'):
            createTextNode(ovf, product_node, 'VendorUrl', product.get('VendorUrl'))
        if product.get('AppUrl'):
            createTextNode(ovf, product_node, 'AppUrl', product.get('AppUrl'))

        # FIXME: add <Icon> support.
        # FIXME: <Category> has design problem. Don't support it.
        for prop in product['Property']:
            subnode = ovf.createElement('Property')
            subnode.setAttribute('ovf:key', prop['key'])
            subnode.setAttribute('ovf:type', prop['type'])
            if 'qualifiers' in prop:
                subnode.setAttribute('ovf:qualifiers', prop['qualifiers'])
            if 'userConfigurable' in prop:
                subnode.setAttribute('ovf:userConfigurable', prop['userConfigurable'])
            if 'value' in prop:
                subnode.setAttribute('ovf:value', prop['value'])
            if 'password' in prop:
                subnode.setAttribute('ovf:password', prop['password'])
            if 'Description' in prop:
                createTextNode(ovf, subnode, 'Description', prop['Description'])
            if 'Label' in prop:
                createTextNode(ovf, subnode, 'Label', prop['Label'])
            # FIXME: to support <Value>.
            product_node.appendChild(subnode)

        vs = product['VirtualSystem']
        vsc = product['VirtualSystemCollection']
        if vs:
            parent = getNodeWithId(ovf, 'VirtualSystem', vs)
        elif vsc:
            parent = getNodeWithId(ovf, 'VirtualSystemCollection', vsc)
        else:
            raise RuntimeError('ProductSection is not associated with '
                               'VirtualSystem or VirtualSystemCollection.')
        parent.appendChild(product_node)


def generateVirtualHardware(cfg):
    vhds = []
    for vs in cfg['VirtualSystem']:
        instanceid = 0
        system = {
            'ElementName': 'system',
            'InstanceID': str(instanceid),
            'VirtualSystemType': vs['VirtualSystemType'],
        }
        instanceid += 1
        items = []
        processor = {
            'Description': 'Number of Virtual CPUs',
            'ElementName': 'processor',
            'InstanceID': str(instanceid),
            'ResourceType': '3',
            'VirtualQuantity': str(vs['Processor']),
        }
        items.append(processor)
        instanceid += 1
        memory = {
            'AllocationUnits': 'byte * 2^20',
            'Description': 'Memory Size',
            'ElementName': 'memory',
            'InstanceID': str(instanceid),
            'ResourceType': '4',
            'VirtualQuantity': str(vs['Memory']),
        }
        items.append(memory)
        instanceid += 1
        controller = {
            'Description': 'SCSI Controller',
            'ElementName': 'scsiController0',
            'InstanceID': str(instanceid),
            'ResourceType': '6',
        }
        items.append(controller)
        scsi_controller_id = str(instanceid)
        instanceid += 1
        address_on_parent = 0
        for index, disk in enumerate(vs['Disk']):
            if not disk.get('HostResource'):
                continue
            item = {
                'Description': 'Disk %s' % index,
                'ElementName': 'disk%s' % index,
                'HostResource': disk['HostResource'],
                'InstanceID': str(instanceid),
                'Parent': scsi_controller_id,
                'AddressOnParent': str(address_on_parent),
                'ResourceType': '17',
            }
            items.append(item)
            instanceid += 1
            address_on_parent += 1
        for index, cd in enumerate(vs['CD']):
            item = {
                'Description': 'CD %s' % index,
                'ElementName': 'cd%s' % index,
                'InstanceID': str(instanceid),
                'Parent': scsi_controller_id,
                'AddressOnParent': str(address_on_parent),
                'AutomaticAllocation': 'true',
                'ResourceType': '15',
            }
            if cd.get('HostResource'):
                item['HostResource'] = cd.get('HostResource')
            items.append(item)
            instanceid += 1
            address_on_parent += 1
        for index, ethernet in enumerate(vs['Ethernet']):
            # FIXME: how to map EPASD to OVM network?
            item = {
                'Description': 'Ethernet controller %s' % index,
                'ElementName': 'ethernet%s' % index,
                'InstanceID': str(instanceid),
                'ResourceType': '10',
            }
            items.append(item)
            instanceid += 1
        vhd = {
            'Info': 'Virtual hardware requirements for a virtual machine.',
            'VirtualSystem': vs['id'],
            'System': [system],
            'Item': items,
        }
        vhds.append(vhd)
    return vhds


def createOvfFile(ovf_filename, cfg):
    ovf = Document()
    createEnvelope(ovf)
    createReferences(ovf, cfg['File'])
    createDiskSection(ovf, cfg.get('Disk', []))
    createNetworkSection(ovf, cfg.get('Network', []))
    createVirtualSystemCollection(ovf, cfg.get('VirtualSystemCollection', []))
    createVirtualSystem(ovf, cfg['VirtualSystem'])
    vhds = generateVirtualHardware(cfg)
    createVirtualHardwareSection(ovf, vhds)
    createProductSection(ovf, cfg.get('Product', []))
    ovf_file = open(ovf_filename, 'w')
    ovf.writexml(ovf_file)
    ovf_file.close()


def isPvGuest(vstype):
    return (vstype.startswith('DMTF:Oracle:X86:OracleVM_Xen:PVM') or
            vstype in ['DMTF:Oracle:OracleVM:PVM', 'xenpv'])


def isHvmPvGuest(vstype):
    return vstype.startswith('DMTF:Oracle:X86:OracleVM_Xen:PVHVM')


def isLDomsGuest(vstype):
    return vstype.startswith('DMTF:Oracle:Sparc:OracleVM_LDoms')


def getDomainConfig(ovf_filename, vs_id):
    result = {}
    ovf_cfg = parseOvfFile(ovf_filename)
    vs_info = None
    for vs in ovf_cfg['VirtualSystem']:
        if vs['id'] == vs_id:
            vs_info = vs
            break
    if not vs_info:
        raise RuntimeError('Cannot find virtual system with id: %s' % vs_id)

    result['name'] = str(vs_info['id'])
    result['memory'] = int(vs_info['Memory'])
    if vs_info.get('MaxMemory'):
        result['maxmem'] = vs_info['MaxMemory']
    if vs_info['Processor']:
        result['vcpus'] = vs_info['Processor']
    if vs_info.get('MaxProcessor'):
        result['maxvcpus'] = vs_info['MaxProcessor']

    # FIXME: we convert all disks to 'xvd'.
    result['disk'] = []
    xen_disk_count = 0
    for disk_info in vs_info['Disk']:
        if disk_info['HostResource']:
            backend = disk_info['HostResource']
        else:
            raise RuntimeError('HostResource is required for disk: %s' % disk_info)
        frontend = 'xvd%s' % blkname(xen_disk_count)
        xen_disk_count += 1
        mode = 'w'
        result['disk'].append(str('%s,%s,%s' % (backend, frontend, mode)))

    for disk_info in vs_info['CD']:
        if disk_info['HostResource']:
            backend = disk_info['HostResource']
        elif not isPvGuest(vs_info['VirtualSystemType']):
            backend = ''
        else:
            raise RuntimeError('HostResource is required for disk: %s' % disk_info)
        frontend = 'xvd%s' % blkname(xen_disk_count)
        xen_disk_count += 1
        frontend += ':cdrom'
        mode = 'r'
        result['disk'].append(str('%s,%s,%s' % (backend, frontend, mode)))

    result['vif'] = []
    for vif in vs_info['Ethernet']:
        result['vif'].append('')

    if isLDomsGuest(vs_info['VirtualSystemType']):
        result['OVM_domain_type'] = 'ldoms_pvm'
    elif isPvGuest(vs_info['VirtualSystemType']):
        result['builder'] = 'linux'
        result['bootloader'] = '/usr/bin/pygrub'
        result['vfb'] = ['type=vnc']
    elif isHvmPvGuest(vs_info['VirtualSystemType']):
        result['OVM_domain_type'] = 'xen_hvm_pv'
        result['builder'] = 'hvm'
        result['serial'] = 'pty'
        result['vnc'] = 1
    else:
        result['builder'] = 'hvm'
        result['serial'] = 'pty'
        result['vnc'] = 1

    return result


def parseHostResource(resource):
    if resource is None:
        return (None, None)
    m = re.match(r'^(ovf:)?/(file|disk)/(.+)$', resource)
    if m:
        return (m.group(2), m.group(3))
    else:
        return (None, None)


def deployFile(ovf_filename, file_id, dst_filename):
    ovf = parseXmlFile(ovf_filename)
    files = getFile(ovf)
    file_info = None
    for item in files:
        if item['id'] == file_id:
            file_info = item
            break
    if not file_info:
        raise RuntimeError('Cannot find file with id: %s' % file_id)

    unpacked_path = os.path.dirname(ovf_filename)
    imports_path = os.path.join(unpacked_path, '../imports')
    if '://' in file_info['href']:
        src_filename = os.path.join(imports_path, os.path.basename(file_info['href']))
    else:
        src_filename = os.path.join(unpacked_path, file_info['href'])

    shutil.copyfile(src_filename, dst_filename)
    convertFile(src_filename, 'raw', dst_filename, 'raw', file_info['compression'],
                file_info['chunkSize'], 0, 0)


def deployDisk(ovf_filename, disk_id, dst_filename):
    ovf = parseXmlFile(ovf_filename)
    files = getFile(ovf)
    disks = getDisk(ovf)
    disk_info = None
    for disk in disks['Disk']:
        if disk['diskId'] == disk_id:
            disk_info = disk
            break
    if not disk_info:
        raise RuntimeError('Cannot find disk with id: %s' % disk_id)

    dst_size = int(disk_info['capacity'])
    if disk_info['capacityAllocationUnits']:
        dst_size *= getAllocationUnits(disk_info['capacityAllocationUnits'])

    # FIXME: how to get disk format? No definiton on OVF spec. Always use 'raw' for now.
    dst_format = 'raw'

    if not disk_info['fileRef']:
        createFile(dst_filename, dst_format, dst_size, disk_info['populatedSize'])
        return

    file_info = None
    for item in files:
        if item['id'] == disk_info['fileRef']:
            file_info = item
            break
    if not file_info:
        raise RuntimeError('Disk %s refrenced file %s is not exist' % (disk_id, disk_info['fileRef']))

    unpacked_path = os.path.dirname(ovf_filename)
    imports_path = os.path.join(unpacked_path, '../imports')
    if '://' in file_info['href']:
        src_filename = os.path.join(imports_path, os.path.basename(file_info['href']))
    else:
        src_filename = os.path.join(unpacked_path, file_info['href'])

    if file_info['href'].lower().endswith('vmdk'):
        src_format = 'vmdk'
    elif file_info['href'].lower().endswith('vdi'):
        src_format = 'vdi'
    elif file_info['href'].lower().endswith('vhd'):
        src_format = 'vhd'
    else:
        src_format = 'raw'
    convertFile(src_filename, src_format, dst_filename, dst_format, file_info['compression'],
                file_info['chunkSize'], dst_size, disk_info['populatedSize'])

