# Copyright (C) 2006-2016 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.

import errno
import glob
import logging
import logging.handlers
import os
import select
import signal
import sys
import time

from distutils.sysconfig import get_python_lib

from agent.lib.config import get_config
from agent.target.host import run_cmd


def init_log():
    config = get_config()
    logger = logging.getLogger()
    logger.propagate = 0
    handler = logging.handlers.SocketHandler('127.0.0.1',
                  logging.handlers.DEFAULT_TCP_LOGGING_PORT)
    for hdlr in logger.handlers:
        logger.removeHandler(hdlr)
    logger.addHandler(handler)
    loglevel = logging.getLevelName(config.get('logger', 'log-level'))
    logger.setLevel(loglevel)


def ignore_eintr(function, *arguments):
    """
    Execute a function and ignore any EINTR error.

    If the function is interrupted by a EINTR error, then it is re-executed.

    """
    while True:
        try:
            return function(*arguments)
        except (OSError, IOError), err:
            if err != errno.EINTR:
                raise
        except select.error, (err, msg):
            if err != errno.EINTR:
                raise


def write_pidfile(pidfile):
    flags = (os.O_CREAT | os.O_EXCL | os.O_WRONLY)
    mode = (((os.R_OK | os.W_OK) << 6) | ((os.R_OK) << 3) | ((os.R_OK)))
    fd = os.open(pidfile, flags, mode)
    f = os.fdopen(fd, 'w')
    f.write('%d\n' % os.getpid())
    f.close()


def delete_pidfile(pidfile):
    try:
        os.unlink(pidfile)
    except OSError, err:
        if err.errno == errno.ENOENT:
            pass
        else:
            raise


def get_processes_children(pid_list):
    if not pid_list:
        return []
    try:
        cmd = ['pgrep', '-P', ','.join([str(pid) for pid in pid_list])]
        return run_cmd(cmd).splitlines()
    except Exception:
        return []


def get_process(pidfile):
    f = open(pidfile)
    try:
        return f.read().strip()
    finally:
        f.close()


def get_agent_process():
    result = []
    pidfile_list = glob.glob('/var/run/ovs-agent/*.pid')
    for pidfile in pidfile_list:
        pid = get_process(pidfile)
        result.append((pid, pidfile))
    return result


def stop_process(pid, childrenonly=False):
    pid = str(pid)
    children = set(get_processes_children([pid]))
    new_children = set(get_processes_children(children))
    all_children = children.union(new_children)
    while children != all_children:
        children = all_children
        new_children = set(get_processes_children(children))
        all_children = children.union(new_children)
    if childrenonly:
        all_pids = children
    else:
        all_pids = children.union(set([pid]))
    for pid in all_pids:
        for _ in range(3):
            try:
                os.kill(int(pid), signal.SIGKILL)
            except OSError:
                pass


def stop_agent_process():
    for pid, pidfile in get_agent_process():
        stop_process(pid)
        try:
            delete_pidfile(pidfile)
        except Exception:
            pass


def reload_server(name, sig=signal.SIGHUP):
    pidfile = ('/var/run/ovs-agent/%s.pid' % name)
    try:
        if os.path.exists(pidfile):
            pid = get_process(pidfile)
            os.kill(int(pid), sig)
    except Exception, err:
        logging.warn('Cannot reload server %s: %s', name, err)


def reload_notification_server():
    reload_server('notificationserver')


def reload_xmlrpc_server():
    reload_server('xmlrpc')


def reload_stats_server():
    reload_server('stats')


def reload_status_server():
    reload_server('status')


def reload_fsstats_server():
    reload_server('fsstats')


def reload_apparent_size_server():
    reload_server('apparentsize')


def restart_xmlrpc_server():
    reload_server('xmlrpc', signal.SIGUSR1)


def clean_up_lock():
    for lockfile in glob.glob('/var/run/ovs-agent/*.lock'):
        os.unlink(lockfile)


def get_daemon_cmd(name):
    cmd = os.path.join(get_python_lib(1), 'agent/daemon/%s.py' % name)
    if not os.path.exists(cmd):
        raise Exception('Unknow daemon: %s' % name)
    return cmd


def close_fds():
    try:
        maxfd = os.sysconf("SC_OPEN_MAX")
    except:
        maxfd = 1024
    os.closerange(3, maxfd) 


def run_service(name):
    pid = os.fork()
    if pid:
        return

    pid = os.fork()
    if pid:
        os._exit(0)

    os.chdir('/')
    os.setsid()
    os.umask(0077)

    os.close(0)
    os.close(1)
    os.close(2)
    os.open(os.devnull, os.O_RDWR)
    os.dup2(0, 1)
    os.dup2(0, 2)

    close_fds()
    cmd = get_daemon_cmd(name)
    pidfile = '/var/run/ovs-agent/%s.pid' % name
    interpreter = sys.executable or '/usr/bin/python'
    try:
        write_pidfile(pidfile)
        os.execv(interpreter, (interpreter, cmd))
    except Exception, err:
        init_log()
        logging.exception('Error starting %s server: %s', name, err)
        os._exit(1)


def delay_cmd(delay, *args, **kwargs):
    pid = os.fork()
    if pid:
        return

    pid = os.fork()
    if pid:
        os._exit(0)

    os.chdir('/')
    os.setsid()
    os.umask(0077)

    os.close(0)
    os.close(1)
    os.close(2)
    os.open(os.devnull, os.O_RDWR)
    os.dup2(0, 1)
    os.dup2(0, 2)

    close_fds()
    time.sleep(delay)
    try:
        run_cmd(*args, **kwargs)
    except Exception, err:
        init_log()
        logging.exception('Error running command with delay %s: %s', delay, err)
        os._exit(1)
