# Copyright (C) 2006-2016 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.

import os
import re
from os.path import join, isdir, exists
import errno

from agent.target.host import run_cmd

# dlm & ha
DLM = "/dlm"
DLM_DOMAIN = "/dlm/ovm"
DLM_LOCK_MASTER = "/dlm/ovm/master"


def dlm_ready():
    """
    Indicate if the DLM is ready of not.

    With dlmfs, the DLM is ready once the DLM filesystem is mounted
    under /dlm.

    @return: C{True} if the DLM is ready, C{False} otherwise.
    @rtype: C{bool}

    """
    return os.path.ismount(DLM)


def dlm_join_domain(domain="ovm"):
    _dir = join(DLM, domain)
    if not isdir(_dir):
        os.mkdir(_dir)
    # else: already joined


def dlm_leave_domain(domain="ovm", force=True):
    _dir = join(DLM, domain)
    if force:
        cmd = ["rm", "-fr", _dir]
    else:
        cmd = ["rmdir", _dir]
    run_cmd(cmd)


def dlm_acquire_lock(lock):
    dlm_join_domain()
    # limit dlm lock name to 31 chars
    lock_path = join(DLM_DOMAIN, re.sub(r"[^a-zA-Z0-9]+", "", lock)[-31:])
    try:
        fd = os.open(lock_path, os.O_CREAT | os.O_EXCL | os.O_NONBLOCK)
        os.close(fd)
    except OSError, e:
        if e.errno == errno.ETXTBSY:
            try:
                os.remove(lock_path)
            except:
                pass
        raise e

    try:
        # create the EX lock
        fd = os.open(lock_path, os.O_RDWR | os.O_NONBLOCK)
        # close it and hold it
        os.close(fd)
    except Exception, e:
        # to get rid of the side effect of touch
        try:
            os.remove(lock_path)
        except:
            pass

        raise e


def dlm_release_lock(lock):
    # limit dlm lock name to 31 chars
    lock_path = join(DLM_DOMAIN, re.sub(r"[^a-zA-Z0-9]+", "", lock)[-31:])
    if exists(lock_path):
        os.remove(lock_path)


def acquire_master_dlm_lock():

    dlm_join_domain()

    # close() does not downconvert the lock level nor does it drop the lock. The
    # holder still owns the lock at that level. All close() does is allows any
    # downconvert request to succeed.
    # However, a downconvert request is only generated for queued requests. And
    # O_NONBLOCK is specifically a noqueue dlm request.

    # 1) O_CREAT | O_NONBLOCK will create a lock file if it does not exist, whether
    #    we are the lock holder or not.
    # 2) if we hold O_RDWR lock, and we close but not delete it, we still hold it.
    #    afterward, O_RDWR will succeed, but O_RDWR | O_NONBLOCK will not.
    # 3) but if we donnot hold the lock, O_RDWR will hang there waiting,
    #    which is not desirable -- any uninterruptable hang is undesirable.
    # 4) if noboday else hold the lock either, but the lock file exists as side effect
    #    of 1), with O_NONBLOCK, it may result in ETXTBUSY

    # a) we need O_NONBLOCK to avoid 3)
    # b) we need to delete it ourself to avoid 2) (for e.g. agtctl restart), but when?
    #   *) if not succeed in 1), remove the lock file (side product), avoid 4)
    #   *) if everything is good, we drop it and we clean out.
    #   *) if stopped with agtctl (killed), agtctl should remove the lock.
    #   *) if crashed, agtctl stop --force should clean out the env.

    try:
        if not exists(DLM_LOCK_MASTER):
            fd = os.open(DLM_LOCK_MASTER, os.O_CREAT | os.O_NONBLOCK)
            os.close(fd)

        master_lock = os.open(DLM_LOCK_MASTER, os.O_RDWR | os.O_NONBLOCK)

        # close it and hold it
        os.close(master_lock)
        return True

    except OSError, e:
        if e.errno == errno.ETXTBSY:
            # if we donnot remove the lock file (the side effect of touch),
            # we may never succeed to hold the exclusive lock, i.e. 4)
            if exists(DLM_LOCK_MASTER):
                os.remove(DLM_LOCK_MASTER)
        else:
            raise e


def release_master_dlm_lock():
    if exists(DLM_LOCK_MASTER):
        os.remove(DLM_LOCK_MASTER)
