# Copyright (C) 2006-2016 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.

import glob
import logging
import os
import re
import tarfile
import shutil
import subprocess
import threading
import time
import zipfile
import uuid

from urlgrabber import urlgrab
from urlgrabber.progress import BaseMeter

from agent.lib.config import get_config
from agent.lib.constants import (LOCAL_EXPORTS_DB, LOCAL_REPO_DB,
                                 LOCAL_TRANSACTION_DB)
from agent.lib.db import delete_item, dump_db, truncate_db, write_item
from agent.lib.notification import async_proc_prog
from agent.lib.settings import get_repo_path
from agent.target.host import (get_fs_type, guess_mimetype, run_cmd,
                               mount, umount, do_export_nfs)


def run_cmd_with_progress(cmd, directory, obj, obj_current, obj_total,
                          update_period, filesize, proc_stdout=None):
    if not os.path.exists(directory):
        os.makedirs(directory)
    if proc_stdout is None:
        proc = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, close_fds=True)
    else:
        proc = subprocess.Popen(cmd, cwd=directory, stdin=proc_stdout, stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE, close_fds=True)

    def progress_reporter():
        try:
            logging.debug('progress reporter for (%s) started', cmd)
            directory_size_orig = get_directory_size(directory)
            time.sleep(update_period)
            proc.poll()
            while proc.returncode is None:
                directory_size_new = get_directory_size(directory)
                unpacked_size = directory_size_new - directory_size_orig
                status = {'OPERATION': 'unpack',
                          'CURRENT_BYTES': long(unpacked_size),
                          'TOTAL_BYTES': long(filesize),
                          'OBJECT': obj,
                          'CURRENT_OBJECT_COUNT': obj_current,
                          'TOTAL_OBJECT_COUNT': obj_total}
                async_proc_prog(status)
                time.sleep(update_period)
                proc.poll()
            logging.debug('progress reporter for (%s) finished', cmd)
        except Exception, err:
            logging.debug('progress reporter for (%s) error: %s', cmd, err)

    threading.Thread(target=progress_reporter).start()
    proc.wait()
    if proc.returncode != 0:
        stdoutdata = ''
        stderrdata = ''
        try:
            stdoutdata = proc.stdout.read()
        except Exception:
            pass
        try:
            stderrdata = proc.stderr.read()
        except Exception:
            pass
        raise RuntimeError('Command: (%s) failed: returncode: %s, stdout: %s, stderr: %s' %
                           (cmd, proc.returncode, stdoutdata, stderrdata))


def init_nfs_exports():
    for v in dump_db(LOCAL_EXPORTS_DB).values():
        try:
            if v['export_type'] == 'nfs':
                do_export_nfs(v['client'], v['path'], v['options'])
        except Exception, e:
            logging.exception('Error exporting %s: %s', v, e)


def mount_repositories():
    for repo in dump_db(LOCAL_REPO_DB).values():
        try:
            mount_repository_fs(repo['fs_location'], repo['mount_point'])
        except Exception, err:
            logging.warn('Error mount repository %s to %s: %s',
                         repo['fs_location'], repo['mount_point'], err)


def mount_repository_fs(fs_location, mount_point):
    repo_mount_point = get_repo_path(mount_point)
    filesystem = get_fs_type(fs_location)
    mount(fs_location, repo_mount_point, filesystem)


def unmount_repository_fs(mount_point):
    repo_mount_point = get_repo_path(mount_point)
    if not os.path.ismount(repo_mount_point):
        raise Exception('Repository file system not mounted')
    umount(repo_mount_point)


def get_directory_size(directory):
    du_re = re.compile(r'^(?P<size>\d+)\s.*$')
    output = run_cmd(['du', '-s', '-b', directory])
    match = du_re.match(output)
    if not match:
        raise Exception('Error getting directory size.')
    return int(match.group('size'))


def remove_files(pathname, force=False, recursive=False):
    """
    Remove directory entries.

    @param pathname: The directory entries to remove. This parameter
    will expanded using C{glob} and all corresponding entries will be
    removed.
    @type  pathname: C{str}

    @param force: forcibly remove the entry, and ignore non-existent
    file.
    @type  force: C{bool}

    @param recursive: remove directories and their contents recursively.
    @param recursive: C{bool}

    """
    cmd = ['rm']
    if force:
        cmd.append('-f')
    if recursive:
        cmd.append('-r')

    entries = glob.glob(pathname)
    if entries:
        cmd += entries
        run_cmd(cmd)


def rename_file(src, dst):
    if os.path.exists(dst):
        os.remove(dst)
    if os.path.exists(src):
        shutil.move(src, dst)


def rmtree(path, ignore_errors=False, onerror=None, raise_exception=False):
    # FIXME: a workaroundfor NFS delete file issue: orabug 12668898
    try:
        return shutil.rmtree(path, ignore_errors, onerror)
    except Exception, err:
        logging.warn('Error remove tree %s: %s', path, err)
    time.sleep(3)
    try:
        return shutil.rmtree(path, ignore_errors, onerror)
    except Exception, err2:
        logging.warn('Second try failed too: %s', err2)
        if raise_exception:
            raise


def file_read(path):
    f = open(path, "r")
    try:
        return f.read()
    finally:
        f.close()


def get_fs_free_space(directory):
    fsstat = os.statvfs(directory)
    return fsstat.f_bavail * fsstat.f_frsize


def unzip_file(filename, directory, obj, obj_current, obj_total, update_period):
    filesize_re = re.compile(r'^.*\b(?P<size>\d+)\s+\d+\s+files?$', re.DOTALL)
    zipinfo = run_cmd(['unzip', '-l', filename])
    match = filesize_re.match(zipinfo)
    if not match:
        raise RuntimeError('Unrecognized zip file: %s' % filename)
    filesize = int(match.group('size'))
    disk_free = get_fs_free_space(directory)
    if filesize > disk_free:
        raise RuntimeError('Not enough space to unpack %s to directory %s: '
                           'disk free: %s, need space: %s'
                           % (filename, directory, disk_free, filesize))
    cmd = ['unzip', '-o', filename, '-d', directory]
    run_cmd_with_progress(cmd, directory, obj, obj_current, obj_total,
                          update_period, filesize)


def unpack_rpm(filename, directory, obj, obj_current, obj_total, update_period):

    try:
        import rpm
    except ImportError:
        logging.debug("rpm module is not available")
        rpm = None
    if rpm:
        ts = rpm.TransactionSet()
        fd = os.open(filename, os.O_RDONLY)
        hdr = ts.hdrFromFdno(fd)
        filesize = hdr[rpm.RPMTAG_SIZE]
        os.close(fd)
    else:
        filesize = 0
    disk_free = get_fs_free_space(directory)
    if filesize > disk_free:
        raise RuntimeError('Not enough space to unpack %s to directory %s: '
                           'disk free: %s, need space: %s'
                           % (filename, directory, disk_free, filesize))
    cmd = ['rpm2cpio', filename]
    proc = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, close_fds=True)
    # hook up proc.stdout as stdin to cpio command
    cmd = ['cpio', '-iumd']
    run_cmd_with_progress(cmd, directory, obj, obj_current, obj_total,
                          update_period, filesize, proc.stdout)
    proc.stdout.close() # close the output so proc does not wait


def untar_file(filename, directory, obj, obj_current, obj_total, update_period):
    ftype = guess_mimetype(filename)
    if ftype == 'application/x-tar':
        cmd = ['gtar', 'xf', filename, '--no-same-owner', '-C', directory]
    elif ftype == 'application/x-gzip':
        cmd = ['gtar', 'xzf', filename, '--no-same-owner', '-C', directory]
    elif ftype == 'application/x-bzip2':
        cmd = ['gtar', 'xjf', filename, '--no-same-owner', '-C', directory]
    else:
        logging.debug('unsupported tarball format: %s', ftype)
        return
    run_cmd_with_progress(cmd, directory, obj, obj_current, obj_total,
                          update_period, 0)


def unpack_file(filename, directory, obj='file', obj_current=1, obj_total=1,
                update_period=20):
    if tarfile.is_tarfile(filename):
        untar_file(filename, directory, obj, obj_current, obj_total,
                   update_period)
    elif zipfile.is_zipfile(filename):
        unzip_file(filename, directory, obj, obj_current, obj_total,
                   update_period)
    elif filename.endswith('.rpm'):
        unpack_rpm(filename, directory, obj, obj_current, obj_total,
                   update_period)


def ovs_download_file(url, filename, option=None, obj='file', obj_current=1,
                      obj_total=1, update_period=20):
    meter = OVSMeter(update_period, obj, obj_current, obj_total)
    opts = {'progress_obj': meter,
            'copy_local': 1,
            'ssl_verify_peer': False,
            'ssl_verify_host': False}
    if option is not None:
        opts.update(option)
    try:
        # OVM Manager passes in an empty http proxy ({'proxies': {'http': ''}})
        # when no proxy is set in Manager. We should override the empty proxy
        # if we have a pre-configured one in agent.ini.
        # Also OVM Manager only passes down 'http' proxy which really means
        # using this proxy for all. We should respect this assumption.
        config = get_config()
        proxy = config.get('network', 'proxy')
        if proxy:
            if 'proxies' not in opts:
                opts['proxies'] = {'http': proxy, 'https': proxy, 'ftp': proxy}
            else:
                http_proxy = opts['proxies'].get('http')
                if http_proxy:
                    opts['proxies']['https'] = http_proxy
                    opts['proxies']['ftp'] = http_proxy
                else:
                    opts['proxies'] = {'http': proxy, 'https': proxy, 'ftp': proxy}
    except Exception, err:
        # Do nothing when no proxy is configured.
        pass
    tmp_name = '%s.%s' % (filename, get_uuid())
    txn = begin_transaction(tmp_name)
    try:
        #
        # Make sure we don't pass unicode strings, otherwise urlgrab will
        # eventually raise an exception when it uses pycurl/setopt.
        #
        filepath = urlgrab(str(url), str(tmp_name), **opts)
        shutil.move(filepath, filename)
        return filename
    except Exception, err:
        try:
            os.unlink(tmp_name)
        except:
            pass
        raise err
    except KeyboardInterrupt:
        # When device is full, urlgrab sends a KeyboardInterrupt,
        # not an IOError exception.
        try:
            os.unlink(tmp_name)
        except:
            pass
        raise Exception('KeyboardInterrupt: Error copying file %s. Check space on device.' % tmp_name)
    finally:
        end_transaction(txn)


class OVSMeter(BaseMeter):
    def __init__(self, update_period, obj, obj_current, obj_total):
        BaseMeter.__init__(self)
        self.update_period = update_period
        self.obj = obj
        self.obj_current = obj_current
        self.obj_total = obj_total

    def send_notification(self, amount_read, now=None):
        if self.size is not None:
            status = {'OPERATION': 'download',
                      'CURRENT_BYTES': long(self.last_amount_read),
                      'TOTAL_BYTES': long(self.size),
                      'OBJECT': self.obj,
                      'CURRENT_OBJECT_COUNT': self.obj_current,
                      'TOTAL_OBJECT_COUNT': self.obj_total}
            async_proc_prog(status)

    def _do_update(self, amount_read, now=None):
        self.send_notification(amount_read, now)

    def _do_end(self, amount_read, now=None):
        self.send_notification(amount_read, now)


def get_uuid():
    return uuid.uuid4().hex


def begin_transaction(name):
    txn = get_uuid()
    write_item(LOCAL_TRANSACTION_DB, txn, name)
    return txn


def end_transaction(txn):
    delete_item(LOCAL_TRANSACTION_DB, txn)


def cleanup_transactions():
    for txn, filepath in dump_db(LOCAL_TRANSACTION_DB).items():
        logging.info("Cleaning up tmp file from transaction:%s file=%s", txn, filepath)
        remove_files(filepath, True, True)
    truncate_db(LOCAL_TRANSACTION_DB)
