# Copyright (C) 2006-2016 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.

import logging
import os
import re
import time
import xen.lowlevel.xc
import xen.xend.XendClient
from xen.xend.xenstore.xstransact import xstransact
from agent.target.host import run_cmd, get_server_unique_id, set_nice
from agent.target.hypervisor import get_vms
from agent.lib.stats import (StatsCollectorBase,
                             FSStatsCollectorBase,
                             get_fs_stats_uuids,
                             ApparentSizeCollectorBase)
from agent.lib.notification import send_statistic


MAX_LONG = 2 ** 63 - 1


# FIXME: Some statistics are unsigned long while our interface to Manager is
# long.  Use this algorithm to convert unsigned long to long. This conversion
# will not be needed if we change the interface to unsigned long.
def convert_to_long(number):
    return long(number) % (MAX_LONG)


def read_file(*args):
    tmpfile = open(os.path.join(*args))
    try:
        return tmpfile.read().strip()
    finally:
        tmpfile.close()


def read_xenstore(path):
    try:
        return xstransact.Read(path).strip()
    except Exception, err:
        logging.debug('Error reading xenstore (%s): %s', path, err)


def get_vm_stats():
    try:
        return get_vms()
    except Exception, err:
        logging.debug('Error getting VM stats: %s', err)


def get_disk_uuid(dev):
    try:
        info = run_cmd(['dmsetup', 'info', '--columns', '--noheadings',
                        '--options=UUID', '/dev/%s' % dev])
        if not info.startswith('part'):
            return info.rstrip('\n').replace('mpath-', '')
    except Exception:
        pass


def get_disk_stat():
    disk_read = {}
    disk_write = {}
    try:
        f = open('/proc/diskstats')
        try:
            for line in f:
                stat = line.split()
                if not stat[2].startswith('dm-'):
                    continue
                disk_uuid = get_disk_uuid(stat[2])
                if not disk_uuid:
                    continue
                read = long(stat[5]) << 9
                write = long(stat[9]) << 9
                disk_read[disk_uuid] = convert_to_long(read)
                disk_write[disk_uuid] = convert_to_long(write)
        finally:
            f.close()
    except Exception, err:
        logging.debug('Error getting Server disk read/write: %s', err)
    return disk_read, disk_write


def get_port_stat():
    port_rx = {}
    port_tx = {}
    try:
        basedir = '/sys/class/net'
        for entry in os.listdir(basedir):
            statdir = os.path.join(basedir, entry, 'statistics')
            if not os.path.exists(statdir):
                continue
            if entry.startswith('vif') or entry == 'lo':
                continue
            rx = read_file(statdir, 'rx_bytes')
            tx = read_file(statdir, 'tx_bytes')
            port_rx[entry] = convert_to_long(rx)
            port_tx[entry] = convert_to_long(tx)
    except Exception, err:
        logging.debug('Error getting Server port RX/TX: %s', err)
    return port_rx, port_tx


def get_per_cpu_idle_time(max_cpus):
    # cpu idletime from xc.getcpuinfo() is in nanosecond. Convert it into second.
    per_cpu_idle_time = []
    try:
        xc = xen.lowlevel.xc.xc()
        cpu_info = xc.getcpuinfo(max_cpus)
        per_cpu_idle_time = [float(x['idletime']) / (10 ** 9) for x in cpu_info]
    except Exception, err:
        logging.debug('Error getting Server per physical CPU idle time: %s', err)
    return per_cpu_idle_time


def get_vm_disk_stat(domid):
    disk_read = {}
    disk_write = {}
    try:
        basedir = '/sys/bus/xen-backend/devices'
        for entry in os.listdir(basedir):
            if not re.match(r'^vbd-%s-\d+$' % domid, entry):
                continue
            nodename = read_file(basedir, entry, 'nodename')
            backend = read_xenstore('/local/domain/0/%s/params' % nodename)
            if not backend:
                continue
            # remove leading 'aio:', 'ioemu:', etc.
            backend = backend.split(':', 1)[-1]
            backend = os.path.basename(backend)
            if not backend:
                continue
            statdir = os.path.join(basedir, entry, 'statistics')
            if not os.path.exists(statdir):
                continue
            read = read_file(statdir, 'rd_sect')
            write = read_file(statdir, 'wr_sect')
            disk_read[backend] = convert_to_long(read)
            disk_write[backend] = convert_to_long(write)
    except Exception, err:
        logging.debug('Error getting VM %s disk read/write: %s', domid, err)
    return disk_read, disk_write


def get_vm_port_stat(domid):
    port_rx = {}
    port_tx = {}
    try:
        basedir = '/sys/class/net'
        for entry in os.listdir(basedir):
            match = re.match(r'^vif%s\.(?P<index>\d+)$' % domid, entry)
            if not match:
                continue
            index = match.group('index')
            mac = read_xenstore('/local/domain/0/backend/vif/%s/%s/mac' % (domid, index))
            if not mac:
                continue
            mac = mac.lower().strip()
            statdir = os.path.join(basedir, entry, 'statistics')
            if not os.path.exists(statdir):
                continue
            # backend TX == VM frontend RX
            # backend RX == VM frontend TX
            rx = read_file(statdir, 'tx_bytes')
            tx = read_file(statdir, 'rx_bytes')
            port_rx[mac] = convert_to_long(rx)
            port_tx[mac] = convert_to_long(tx)
    except Exception, err:
        logging.debug('Error getting VM %s port RX/TX: %s', domid, err)
    return port_rx, port_tx


def tmem_parse(line):
    parsed = {}
    for item in line[2:].split(','):
        key, value = item.split(':')
        if key in ['U0', 'U1']:
            parsed[key] = int(value, 16)
        elif key == 'PT':
            parsed[key] = value
        else:
            parsed[key] = int(value)
    return parsed


def tmem_list():
    xc = xen.lowlevel.xc.xc()
    raw = xc.tmem_control(-1, 4, -1, 32768, 1, 0, '')
    if raw == -1:
        return {}
    parsed = {'global': {},
              'time_stats': {},
              'clients': [],
              'pools': [],
              'shared_pools': []}
    for line in raw.splitlines():
        if line.startswith('G='):
            parsed['global'] = tmem_parse(line)
        if line.startswith('T='):
            parsed['time_stats'] = tmem_parse(line)
        if line.startswith('C='):
            parsed['clients'].append(tmem_parse(line))
        if line.startswith('P='):
            parsed['pools'].append(tmem_parse(line))
        if line.startswith('S='):
            parsed['shared_pools'].append(tmem_parse(line))
    result = {}
    for pool in parsed['pools']:
        client_id = int(pool['CI'])
        if client_id not in result:
            result[client_id] = [0, 0]
        pgp_count = pool['Pc']
        if pool['PT'] == 'EP':
            result[client_id][0] += pgp_count
        if pool['PT'] == 'PP':
            result[client_id][1] += pgp_count
    return result


def tmem_query_freeable_mb():
    xc = xen.lowlevel.xc.xc()
    freeable_mb = xc.tmem_control(-1, 8, -1, 0, 0, 0, '')
    if freeable_mb == -1:
        return 0
    else:
        return freeable_mb


class StatsCollector(StatsCollectorBase):
    def __init__(self):
        StatsCollectorBase.__init__(self)
        self._previous_timestamp = 0.0
        self._previous_cpu_time = {}
        self._previous_per_cpu_idle_time = []

    def _run(self):
        current_timestamp = time.time()
        time_delta = current_timestamp - self._previous_timestamp
        if time_delta <= 0:
            self._previous_timestamp = current_timestamp
            return

        vm_stats = get_vm_stats()
        if not vm_stats:
            return
        nodeinfo = dict(xen.xend.XendClient.server.xend.node.info())
        tmeminfo = tmem_list()
        nr_cpus = nodeinfo['nr_cpus']

        disk_read, disk_write = get_disk_stat()
        port_rx, port_tx = get_port_stat()

        ptmem = sum([x for _, x in tmeminfo.values()])
        etmem = tmem_query_freeable_mb()

        dom0_cpu_time = 0.0
        current_cpu_time = {}
        for vm in vm_stats:
            try:
                domid = int(vm['domid'])
                name = vm['name']
                online_vcpu_count = int(vm['online_vcpus'])
                target_mem = vm['memory']

                xc = xen.lowlevel.xc.xc()
                try:
                    xcdominfo = xc.domain_getinfo(domid, 1)
                    current_mem = xcdominfo[0]['mem_kb'] / 1024
                except Exception:
                    current_mem = target_mem

                if domid in tmeminfo:
                    vmetmem = tmeminfo[domid][0]
                    vmptmem = tmeminfo[domid][1]
                else:
                    vmetmem = 0
                    vmptmem = 0

                vm_disk_read, vm_disk_write = get_vm_disk_stat(domid)
                vm_port_rx, vm_port_tx = get_vm_port_stat(domid)

                cpu_util = 0.0
                current_vm_cpu_time = float(vm['cpu_time'])
                previous_vm_cpu_time = self._previous_cpu_time.get(vm['uuid'])
                if previous_vm_cpu_time:
                    cpu_time_delta = abs(current_vm_cpu_time - previous_vm_cpu_time)
                    cpu_time_delta_ceiling = time_delta * online_vcpu_count
                    if cpu_time_delta > cpu_time_delta_ceiling:
                        cpu_time_delta = cpu_time_delta_ceiling
                    cpu_util = cpu_time_delta / cpu_time_delta_ceiling

                current_cpu_time[vm['uuid']] = current_vm_cpu_time

                # Bypass Domain-0: dom0 statistics will be reported as 'SERVER' type.
                if domid == 0:
                    dom0_cpu_time = current_vm_cpu_time
                    continue

                vm_stat = {
                    'timestamp': long(time.time() * 1000),
                    '@type': 'Vm',
                    'type': 'VM',
                    'id': vm['uuid'],
                    'cpuTime': current_vm_cpu_time,
                    'cpuUtil': cpu_util,
                    'currentMemory': long(current_mem),
                    'diskRead': vm_disk_read,
                    'diskWrite': vm_disk_write,
                    'domainId': domid,
                    'ephemeralTranscendentMemory': long(vmetmem),
                    'maxMemory': long(vm['maxmem']),
                    'networkReceive': vm_port_rx,
                    'networkSend': vm_port_tx,
                    'onlineVcpus': online_vcpu_count,
                    'persistentTranscendentMemory': long(vmptmem),
                    'targetMemory': long(target_mem),
                }

                send_statistic(vm_stat)
            except Exception, err:
                logging.warn('Error getting VM statistics: %s', err)
                continue

        per_cpu_util = []
        current_per_cpu_idle_time = get_per_cpu_idle_time(nr_cpus)
        if len(current_per_cpu_idle_time) == len(self._previous_per_cpu_idle_time):
            for index, current_cpu_idle_time in enumerate(current_per_cpu_idle_time):
                previous_cpu_idle_time = self._previous_per_cpu_idle_time[index]
                cpu_idle_time_delta = abs(current_cpu_idle_time - previous_cpu_idle_time)
                if cpu_idle_time_delta > time_delta:
                    cpu_idle_time_delta = time_delta
                cpu_idle = cpu_idle_time_delta / time_delta
                per_cpu_util.append(1 - cpu_idle)

        if per_cpu_util:
            average_cpu_util = sum(per_cpu_util) / float(len(per_cpu_util))
        else:
            average_cpu_util = 0.0

        total_memory = long(nodeinfo['total_memory'])
        free_memory = long(nodeinfo['free_memory'])
        used_memory = total_memory - free_memory

        meminfo = dict([item.strip() for item in line.split(':', 1)]
                       for line in read_file('/proc/meminfo').splitlines())
        free_swap = long(meminfo['SwapFree'].split()[0])

        server_stat = {
            'timestamp': long(time.time() * 1000),
            '@type': 'Server',
            'type': 'SERVER',
            'id': get_server_unique_id(),
            'cpuTime': dom0_cpu_time,
            'cpuUtil': average_cpu_util,
            'perCpuUtil': per_cpu_util,
            'diskRead': disk_read,
            'diskWrite': disk_write,
            'ephemeralTranscendentMemory': long(etmem),
            'freeMemory': free_memory,
            'networkReceive': port_rx,
            'networkSend': port_tx,
            'persistentTranscendentMemory': long(ptmem),
            'usedMemory': used_memory,
            'freeSwap': free_swap,
        }

        send_statistic(server_stat)

        self._previous_timestamp = current_timestamp
        self._previous_cpu_time = current_cpu_time
        self._previous_per_cpu_idle_time = current_per_cpu_idle_time


def parse_df_line_output(line, fs_type):
    fs_stats = {}
    fs_name, fs_size, fs_used, fs_avail, fs_pcnt, fs_mnt = line.split()
    fs_total_size = long(fs_size.strip('M'))
    fs_used_size = long(fs_used.strip('M'))
    fs_stats['free'] = fs_total_size - fs_used_size
    fs_stats['total'] = fs_total_size
    fs_stats['util'] = float(fs_used_size) / fs_total_size
    if fs_type == 'nfs':
        return fs_mnt, fs_stats
    else:
        return fs_name, fs_stats, fs_mnt


def get_repos_total_files_size(fs_mnt):
    total_files_size = 0
    set_nice()
    # limit du command on OVM repositories mount directories only
    if fs_mnt.startswith('/OVS'):
        try:
            # .snapshot directory from netapps nfs exported filesystem causes warnings and miscalculation
            # .zfs directory from ZFSSA nfs exported filesystem causes miscalculation
            # lost+found directory also can cause miscalculation
            # Exclude these directories to avoid miscalculation from du command
            # Example of du command:
            # du -sh -B M --apparent-size \
            #   --exclude /OVS/Repositories/0004fb00000300008740d79e5d5e806c/.snapshot \
            #   --exclude /OVS/Repositories/0004fb00000300008740d79e5d5e806c/.zfs \
            #   --exclude /OVS/Repositories/0004fb00000300008740d79e5d5e806c/lost+found \
            #   /OVS/Repositories/0004fb00000300008740d79e5d5e806c
            output = run_cmd(['du', '-sh', '-B', 'M', '--apparent-size', 
                '--exclude', '%s/.snapshot' % fs_mnt, 
                '--exclude', '%s/.zfs' % fs_mnt, 
                '--exclude', '%s/.lost+found' % fs_mnt, 
                fs_mnt])
            total_files_size = long(output.split()[0].strip('M'))
        except Exception, err:
            logging.debug('Error getting apparent size : %s', err)

    return total_files_size


def get_file_system_statistic(apparent_size):
    df_local = {}
    df_nfs = {}

    try:
        # local file system list
        output = run_cmd(['df', '-B', 'M', '-l', '-P'])
        lines = output.splitlines()
        for line in lines:
            fs_stats = {}
            if not line.startswith('/dev'):
                continue
            fs_name, fs_stats, fs_mnt = parse_df_line_output(line, '')
            # collect only if apparent_size is set True
            if apparent_size == True:
                fs_stats['filessize'] = get_repos_total_files_size(fs_mnt)
            df_local[fs_name] = fs_stats
    except Exception, err:
        logging.debug('Error getting local file system : %s', err)

    try:
        # network file system list
        output = run_cmd(['df', '-B', 'M', '-t', 'nfs', '-P'])
        lines = output.splitlines()
        for line in lines:
            fs_stats = {}
            if line.startswith('Filesystem'):
                continue
            fs_mnt, fs_stats = parse_df_line_output(line, 'nfs')
            # collect only if apparent_size is set True
            if apparent_size == True:
                fs_stats['filessize'] = get_repos_total_files_size(fs_mnt)
            df_nfs[fs_mnt] = fs_stats
    except Exception:
        pass # to avoid df -t nfs command error when no nfs filesystem

    return df_local, df_nfs


def process_file_system_stats(apparent_size):
    fs_stats = {}
    fs_uuids = get_fs_stats_uuids()
    if not fs_uuids:
        return fs_stats

    # apparent-size is collected if it set to True, default is False
    df_local, df_nfs = get_file_system_statistic(apparent_size)
    try:
        for fs_uuid in fs_uuids:
            found_uuid = False
            for key, value in df_local.items():
                try:
                    # get UUID of the device
                    cmd = ['blkid', '-s', 'UUID', '-o', 'value', key]
                    output = run_cmd(cmd)
                    dev_uuid = re.sub('-', '', output.rstrip())
                    if dev_uuid == fs_uuid:   # match uuid found
                        fs_stats[fs_uuid] = value
                        found_uuid = True
                        break
                except Exception, err:
                    logging.debug('Error getting device UUID : %s', err)

            if not found_uuid:
                for key, value in df_nfs.items():
                    # get UUID from the nfs mount
                    filename = key + '/.generic_fs_stamp'
                    if not os.path.exists(filename):
                        continue
                    try:
                        fd = open(filename, 'r')
                        try:
                            line = fd.readlines()
                        finally:
                            fd.close()

                        nfs_uuid = line[0]
                        if nfs_uuid == fs_uuid: # match uuid found
                            fs_stats[fs_uuid] = value
                            break
                    except Exception, err:
                        logging.debug('Error getting nfs UUID : %s', err)
        return fs_stats

    except Exception, err:
        logging.debug('Error getting file system statistic: %s', err)


class FSStatsCollector(FSStatsCollectorBase):
    def _run(self):
        fs_stats = {}
        fs_total = {}
        fs_util = {}
        fs_free = {}
        fs_stats = process_file_system_stats(False)

        # Send file system statistic
        if len(fs_stats):
            for key, value in fs_stats.items():
                fs_free[key] = value['free']
                fs_total[key] = value['total']
                fs_util[key] = value['util']

            fs_stat = {
                'timestamp': long(time.time() * 1000),
                '@type': 'FileSystem',
                'type': 'FILESYSTEM',
                'id': get_server_unique_id(),
                'fsSpaceFree': fs_free,
                'fsSpaceTotal': fs_total,
                'fsSpaceUtilization': fs_util,
            }

            send_statistic(fs_stat)


class ApparentSizeCollector(ApparentSizeCollectorBase):
    def _run(self):
        fs_stats = {}
        fs_filessize = {}
        fs_stats = process_file_system_stats(True)

        # Send apparent size statistic
        if len(fs_stats):
            for key, value in fs_stats.items():
                fs_filessize[key] = value['filessize']

            fs_stat = {
                'timestamp': long(time.time() * 1000),
                '@type': 'FileSystem',
                'type': 'FILESYSTEM',
                'id': get_server_unique_id(),
                'fsTotalFilesSize': fs_filessize,
            }

            send_statistic(fs_stat)

