# Copyright (C) 2006-2016 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.

import logging
import time

from agent.lib.constants import LOCAL_SERVER_DB
from agent.lib.daemon import (reload_stats_server, reload_status_server, 
                              reload_fsstats_server, reload_apparent_size_server)
from agent.lib.db import write_item, read_item, delete_item
from agent.lib.notification import notify_manager
from agent.target.host import get_boot_time


# all intervals are in seconds
STAT_INTERVAL_DEFAULT = 20
STATUS_INTERVAL_DEFAULT = 20
FS_STAT_INTERVAL_DEFAULT = 600
APPARENT_SIZE_INTERVAL_DEFAULT = 900


def set_statistic_interval(interval):
    stats_interval = get_statistic_interval()
    write_item(LOCAL_SERVER_DB, "stat_interval", interval)
    if stats_interval != interval:
        reload_stats_server()


def get_statistic_interval():
    return read_item(LOCAL_SERVER_DB, "stat_interval") or STAT_INTERVAL_DEFAULT


def set_status_interval(interval):
    stats_interval = get_status_interval()
    write_item(LOCAL_SERVER_DB, "status_interval", interval)
    if stats_interval != interval:
        reload_status_server()


def get_status_interval():
    return read_item(LOCAL_SERVER_DB, "status_interval") or STATUS_INTERVAL_DEFAULT


def set_fs_stats_interval(interval):
    fs_stats_interval = get_fs_statistic_interval()
    write_item(LOCAL_SERVER_DB, "fs_stat_interval", interval)
    if fs_stats_interval != interval:
        reload_fsstats_server()


def set_apparent_size_interval(interval):
    apparent_size_interval = get_apparent_size_interval()
    write_item(LOCAL_SERVER_DB, "apparent_size_interval", interval)
    if apparent_size_interval != interval:
        reload_apparent_size_server()


def get_fs_statistic_interval():
    return read_item(LOCAL_SERVER_DB, "fs_stat_interval") or FS_STAT_INTERVAL_DEFAULT


def get_apparent_size_interval():
    return read_item(LOCAL_SERVER_DB, "apparent_size_interval") or APPARENT_SIZE_INTERVAL_DEFAULT


def set_fs_stats_uuids(fs_uuid_list):
    write_item(LOCAL_SERVER_DB, "fs_stat_uuid_list", fs_uuid_list)


def get_fs_stats_uuids():
    return read_item(LOCAL_SERVER_DB, "fs_stat_uuid_list")


def delete_fs_stats_uuids():
    delete_item(LOCAL_SERVER_DB, "fs_stat_uuid_list")


def get_last_boot_time():
    # both are seconds since the epoch in UTC.
    return {'local_time': long(time.time()),
            'last_boot_time': long(get_boot_time()[0])}


class CollectorBase(object):
    def __init__(self):
        self._stop = False
        self._sleep_interval = 999999
        self.update_interval()

    def run(self):
        while not self._stop:
            try:
                self._run()
            except Exception, err:
                logging.exception('Error collecting stats: %s', err)
            time.sleep(self._sleep_interval)

    def _run(self):
        # to be overridden by subclasses
        pass

    def stop(self):
        self._stop = True

    def update_interval(self):
        # to be overridden by subclasses
        pass


class StatsCollectorBase(CollectorBase):
    def update_interval(self):
        self._sleep_interval = get_statistic_interval()


class FSStatsCollectorBase(CollectorBase):
    def update_interval(self):
        self._sleep_interval = get_fs_statistic_interval()


class ApparentSizeCollectorBase(CollectorBase):
    def update_interval(self):
        self._sleep_interval = get_apparent_size_interval()


class StatusCollector(CollectorBase):
    def _run(self):
        boot_time_sec, boot_tim_us = get_boot_time()
        data = {
            'timestamp': long(time.time() * 1000),
            'lastBootTime': long(boot_time_sec * 1000 + boot_tim_us / 1000),
        }
        notify_manager('{SERVER} {STATUS}', data)

    def update_interval(self):
        self._sleep_interval = get_status_interval()
