# Copyright (C) 2006-2016 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.

import logging
import os
import time
import errno
import subprocess
import select
import signal
import shutil

from agent.lib import filelock
from agent.lib.settings import get_vm_lock_file, get_vm_path
import agent.extension.osm

MIGRATE_STOP = False

def migrate(domid, domain, dest, vmfiles, vmcfg, live, ssl):
    
    # vmcfg for future use with xl
    
    # set up the sync pipes to the vm migration tool
    # vmfiles = [{'src_file_path': '/OVS/Repositories/0004fb0000030000d7bfedbc3d486708/VirtualDisks/0004fb000012000064d4fe66098b72b3.img',
    #             'dst_file_path': '/OVS/Repositories/0004fb0000030000d7bfedbc3d486708/VirtualDisks/0004fb000012000064d4fe66098b72b3.img',}]"
    
    if not vmfiles:
        raise Exception('Storage Migrate for domain %s does not include any storage files.' % domid)

    logging.debug("Storage migration begin domain %s" % domid)
    
    global MIGRATE_STOP
    
    MIGRATE_FOREVER = -1

    vmtoolup = False
    read_path = "/var/run/VMsyncA%s" % domid
    write_path = "/var/run/VMsyncB%s" % domid
    sync_read_fd = None
    sync_write_fd = None
    connected = False
    host = False
    retry = 0
    
    mg_list = []
    
    def on_stop(sig, frame):
        MIGRATE_STOP = True
    
    signal.signal(signal.SIGUSR2, on_stop)
    
    
    def abort_cleanup():
        logging.debug("Starting abort cleanup")
        
        # If the VM has not already resumed on the other side deleting the img files
        # through the abort migrate will prevent the vm from starting 
        
        # flags = 0 means default abort immediately 
        abort_migrates(mg_list, 0)
        close_migrates(mg_list, 0)
        free_files(mg_list)
        
        if host:
            try:
                libosm.free_host() 
            except Exception, err:
                logging.warning('Free host ignoring error: %s.' % err)
                pass 
        if connected:
            try:
                libosm.disconnect()
            except Exception, err:
                logging.warning('Disconnect ignoring error: %s.' % err)
                pass
        if sync_read_fd:
            os.close(sync_read_fd)
            os.unlink(read_path)
        if sync_write_fd:
            os.close(sync_write_fd)
            os.unlink(write_path)
        if vmtoolup:
            try:
                proc.kill()
            except:
                pass
            
        logging.debug("abort cleanup done")
        # End of abort_cleanup()
    
        
    if os.path.exists(read_path):
        os.unlink(read_path)
    if os.path.exists(write_path):
        os.unlink(write_path)
        
    try:
        os.mkfifo(read_path)
        os.mkfifo(write_path)    
        sync_read_fd = os.open(read_path, os.O_RDWR | os.O_NONBLOCK)
        sync_write_fd = os.open(write_path, os.O_RDWR | os.O_NONBLOCK)
        libosm = agent.extension.osm.osm()
        libosm.connect()
        connected = True
        libosm.alloc_host(dest)
        host = True
    
        mg_list, files = create_migrate_files_list(libosm, vmfiles)
        
        # files = "src:dst,src:dst,..."
        
        logging.debug('Migrating files domid %s.' % domid)
        
        # start storage migration
        
        for mg_file in mg_list:
            # time_start and time_stop are for debug and will be removed
            mg_file['osm_handle'].migrate_file(MIGRATE_FOREVER)
            mg_file['time_start'] = (time.time() * 1000)
            mg_file['time_stop'] = 0
            mg_file['action_state'] = 'MIGRATED'
                   
        # start vm migration
        
        cmd = ['xm', 'migrate']
        if live:
            cmd += ['--live']
        if ssl:
            cmd += ['--ssl']
        cmd += ['--vmfiles=%s' % files]
        cmd += [domain, dest]
        
        proc = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, close_fds=True)
        vmtoolup = True
    except Exception, err:
        logging.error('Storage Migration error: %s.' % err)
        abort_cleanup()
        raise err
        
    for stage in ['StorageReady', 'GuestReady', 'GuestSuspend', 'CompleteMigrate']:   
        while True:
            #If we are completing migration we don't allow a stop.
            if MIGRATE_STOP:
                if stage == 'CompleteMigrate':
                    logging.debug('Storage Migrate domain %s cannot be stopped at this time.' % domid)
                    MIGRATE_STOP = False
                else:
                    abort_cleanup()
                    raise Exception('Storage Migrate domain %s stopped by user.' % domid)
    
            # the migration tool should not stop until we tell it to.
            if proc.poll() is not None:
                logging.error("XM migrate tool exited domain %s" % domid)
                abort_cleanup()
                raise RuntimeError('Command: %s failed: %d'
                               % (cmd, proc.returncode))
                
            if stage == 'StorageReady':
                # check the storage migrate status
                try:
                    (mg_list, done) = migrate_converged(mg_list)
                    #logging.debug("Storage migrate status domain %s: %s" % (domid, mg_list))
                    if done:
                        logging.debug("Storage migrate storage converged domain %s" % domid)
                        break
                    time.sleep(1.0)
                except Exception, err:
                    logging.error('Storage Migration error: %s.' % err)
                    abort_cleanup()
                    raise err 
                 
            if stage == 'GuestReady':
                rd, wr, ex = select.select([sync_read_fd], [], [sync_read_fd], 30)
                if ex:
                    abort_cleanup()
                    raise RuntimeError("VM %s select error" % domid)  
                if rd:
                    resp = read_exact(sync_read_fd, 2, "Sync pipe read error") 
                    if resp == 'GC':
                        logging.debug("Storage migrate VM converged domain %s" % domid)
                        # Tell the guest it may finish the migrate and suspend when ready
                        try:
                            write_exact(sync_write_fd, 'FM', "Sync pipe write error") 
                            break    
                        except:
                            abort_cleanup()
                            raise RuntimeError("Could not write to pipe for domain %s" % domid)
                              
                    else:
                        abort_cleanup()
                        raise RuntimeError("Unexpected return %s from migrate tool domain %s" % (resp, domid))
                        
            if stage == 'GuestSuspend':
                # Has the VM suspended ?
                rd, wr, ex = select.select([sync_read_fd], [], [sync_read_fd], 30)
                if ex:
                    abort_cleanup()
                    raise RuntimeError("VM %s select error" % domid)  
                if rd:
                    resp = read_exact(sync_read_fd, 2, "Sync pipe read error") 
                    if resp == 'GS':  
                        logging.debug("Storage migrate VM suspended domain %s" % domid)
                        try:
                            # flags = 0 means default OSML_MIG_COMP_ZERO_CHANGES 
                            # time start and time stop are for debug and will be removed
                            start = time.time()
                            complete_sync_all_migrates(libosm, 0)
                            stop = time.time() 
                            logging.debug('*** Complete all storage migrates for domain %s took: %.2f seconds' % (domid, (stop - start)))
                            # for test purposes only
                            #reflink_for_compare_checksums(mg_list, dest)
                            break
                        except Exception, err:
                            abort_cleanup()
                            raise RuntimeError("VM %s:%s" % (domid,err))
                    else:
                        abort_cleanup()
                        raise RuntimeError("Unexpected return %s from migrate tool domain %s" % (resp, domid))
                else:
                    retry = retry + 1
                    # If we are at this stage and the guest has converged it should be
                    # trying to suspend.
                    if retry > 20:
                        abort_cleanup()
                        raise RuntimeError("VM %s is not suspending" % domid)
                
            if stage == 'CompleteMigrate':
                try:
                    # Tell xen to resume the guest
                    write_exact(sync_write_fd, 'RG', "Sync pipe write error")
                    close_migrates(mg_list, 0) 
                    break
                except Exception, err:
                    abort_cleanup()
                    raise RuntimeError("VM %s migration is not completing:%s" % (domid,err))
    
    os.close(sync_read_fd)
    os.close(sync_write_fd)
    os.unlink(read_path)
    os.unlink(write_path)
    
    try:
        free_files(mg_list)
    except Exception, err:
        logging.warning('Free regular file ignoring error: %s.' % err)
        pass
    
    try:
        libosm.free_host()
    except Exception, err:
        logging.warning('Free host ignoring error: %s.' % err)
        pass
    try:
        libosm.disconnect()
    except Exception, err:
        logging.warning('Disconnect ignoring error: %s.' % err)
        pass
        
    print_stats(mg_list)
    

def create_migrate_files_list(libosm, vmfiles):

    mg_list = []
    files = ""
    
    for item in vmfiles:
    # create regular file entries
        mg_file = {}
        mg_file['migrate_status'] = 'NO_STATUS'
        mg_file['action_state'] = 'NO_STATE'
        mg_file['infile'] = item.get('src_file_path')
        mg_file['outfile'] = item.get('dst_file_path')
        if len(files) > 0:
            files += ','
        files += "%s:%s" % (mg_file['infile'], mg_file['outfile'])
        mg_file['osm_handle'] = libosm.alloc_file(mg_file['infile'], mg_file['outfile'])
        mg_file['action_state'] = 'ALLOCATED'
        mg_list.append(mg_file)
            
    return mg_list, files
            
        
def write_exact(fd, buf, errmsg):
    
    if os.write(fd, buf) != len(buf):
        logging.debug(errmsg)
        raise RuntimeError('%s' % errmsg)


def read_exact(fd, size, errmsg):
    
    buf  = '' 
    while size != 0: 
        try:
            readstr = os.read(fd, size)
        except OSError, err:
            if err.errno == errno.EAGAIN:
                return None
            logging.debug(errmsg)
            raise RuntimeError('%s' % errmsg)
        if not len(readstr):
            return None
        size = size - len(readstr)
        buf  = buf + readstr
    return buf


def complete_sync_all_migrates(libosm, flags):
    
    libosm.complete_migration_all(flags)


def abort_migrates(mg_list, flags): 

    for mg_file in mg_list:
        if mg_file['action_state'] == 'MIGRATED':
            try:
                mg_file['osm_handle'].abort_migration(flags)
            except Exception, err:
                logging.warning('Aborting migration ignoring error: %s.' % err)
                pass     
        

def free_files(mg_list):
    
    for mg_file in mg_list:
        if mg_file['action_state'] == 'ALLOCATED' or mg_file['action_state'] == 'MIGRATED' or \
           mg_file['action_state'] == 'CLOSED':
            try:
                mg_file['osm_handle'].free_file()
            except Exception, err:
                logging.warning('Freeing files ignoring error: %s.' % err)
                pass        

     
def close_migrates(mg_list, flags):  
    
    for mg_file in mg_list:  
        if mg_file['action_state'] == 'MIGRATED':
            mg_file['action_state'] = 'CLOSED'
            try:
               mg_file['osm_handle'].close_migration(flags)
            except Exception, err:
                logging.warning('Close migrates ignoring error: %s.' % err)
                pass
        
   
def migrate_converged(mg_list):
    
    done = True
    
    for mg_file in mg_list:
        # We don't ask for updated status if it is already at the status we want.
        if mg_file['migrate_status'] == 'CONVERGED':
            continue

        mg_file['osm_status_info'] = mg_file['osm_handle'].migrate_status()
        
        if mg_file['osm_status_info']["state"] == agent.extension.osm.OSML_MIG_STATE_ERROR or \
           mg_file['osm_status_info']["state"] == agent.extension.osm.OSML_MIG_STATE_ERR_REMOTE:
            raise RuntimeError("File %s shows a migration error" % mg_file['infile'])
         
    
        if mg_file['osm_status_info']['state'] == agent.extension.osm.OSML_MIG_STATE_IN_PROGRESS:
            
            # For debug purposes. Will be removed later
            if mg_file['osm_status_info']['snap_count'] > 1:
                logging.debug('\nFile %s' % mg_file['infile'])
                logging.debug('snap count: %lu' % mg_file['osm_status_info']['snap_count'])
                logging.debug('last bytes: %lu' % mg_file['osm_status_info']['last_bytes'])
                logging.debug('avg_byte: %lu' % mg_file['osm_status_info']['avg_byte'])
                logging.debug('bytes_migrated: %lu' % mg_file['osm_status_info']['bytes_migrated'])
            
            
            if  mg_file['osm_status_info']['snap_count'] >= 20:
                mg_file['migrate_status'] = 'CONVERGED'
                logging.debug('File %s has CONVERGED' % mg_file['infile'])
                
            elif  mg_file['osm_status_info']['snap_count'] >= 2 and \
                  mg_file['osm_status_info']['last_bytes'] < 1048576:
                mg_file['migrate_status'] = 'CONVERGED' 
                logging.debug('File %s has CONVERGED' % mg_file['infile'])
                
            else:
                done = False
            
            # For debug purposes. Will be removed later
            if mg_file['migrate_status'] == 'CONVERGED':
                mg_file['time_stop'] = time.time() * 1000
            # End for debug

        else:
            raise RuntimeError('File %s not migrating. State is %d' % (mg_file['infile'],mg_file['osm_status_info']["state"]))

    return mg_list, done


def print_stats(mg_list):
    
    for mg_file in mg_list:
        through_put = mg_file['osm_status_info']['bytes_migrated'] / (mg_file['time_stop'] - mg_file['time_start'])
        logging.debug('*** File %s Average Migrate Rate at Convergence was: %.2f (bytes/sec)' % (mg_file['infile'], through_put * 1000))
        
         
def reflink_for_compare_checksums(mg_list, dest):
    import datetime
    
    tag = ".ref-%s" % datetime.datetime.now().strftime("%y-%m-%d-%H-%M")
  
    for mg_file in mg_list:
        cmd = ['reflink', mg_file['infile'], mg_file['infile'] + tag]
        proc = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, close_fds=True)
        (stdoutdata, stderrdata) = proc.communicate()
        if proc.returncode != 0:
            logging.error('Could not reflink IN File %s)' % mg_file['infile'])
        
        cmd = ['ssh', dest, 'reflink', mg_file['outfile'], mg_file['outfile'] + tag]
        proc = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, close_fds=True)
        (stdoutdata, stderrdata) = proc.communicate()
        if proc.returncode != 0:
            logging.error('Could not reflink OUT File %s)' % mg_file['outfile'])

      

def setup(vmfiles):
    
    for file in vmfiles:
        # create special file entries. These files prevent the agent from reporting their
        # corresponding regular files at the destination to the manager during the migrate.
        
        (filepath, filename) = os.path.split(file)
        # Given that the names are uuids we should not have to worry about collisions
        temp_file = filepath + '/' + 'tmp_dest.' + filename
        fd = os.open(temp_file, os.O_RDWR | os.O_CREAT)
        os.close(fd)
        
    
def cleanup(vmfiles, delete_vmfiles = False):
    
    # This always deletes the tempfiles.
    # If delete_vmfiles is true then it deletes the actual vmfiles as well.
    
    for file in vmfiles:
        (filepath, filename) = os.path.split(file)
        temp_file = filepath + '/' + 'tmp_dest.' + filename
        if os.path.exists(temp_file):
            os.unlink(temp_file)
        if delete_vmfiles and os.path.exists(file):
            os.unlink(file)
            

def cfgfile_setup(repo_uuid, vm_uuid):

    temp_file = '/OVS/Repositories/' + repo_uuid + '/VirtualMachines/' + 'tmp_dest.' + vm_uuid
    fd = os.open(temp_file, os.O_RDWR | os.O_CREAT)
    os.close(fd)


def cfgfile_cleanup(repo_uuid, vm_uuid):

    temp_file = '/OVS/Repositories/' + repo_uuid + '/VirtualMachines/' + 'tmp_dest.' + vm_uuid
    if os.path.exists(temp_file):
        os.unlink(temp_file)
        
def delete_vm(repo_id, vm_id):
    
    lock = filelock.FileLock(get_vm_lock_file(repo_id, vm_id))
    lock.acquire(wait=120)
    try:
        shutil.rmtree(get_vm_path(repo_id, vm_id), True)
    finally:
        lock.release()

