# Copyright (C) 2006-2016 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.

import fnmatch
import os
import random
import string
import shutil

from agent.lib import filelock
from agent.lib.files import (ovs_download_file, rmtree, unpack_file, get_uuid,
                             begin_transaction, end_transaction)
from agent.lib.settings import (get_vm_path, get_template_path, get_iso_path,
                                get_virtualdisk_path, get_vm_lock_file)
from agent.lib.xenvm import (create_cfg, configure_cfg, unconfigure_cfg,
                             get_cfg_file, read_cfg, write_cfg)
from agent.target.host import set_nice


def create_template(repo_id, template_id, params):
    lock = filelock.FileLock(get_vm_lock_file(repo_id, template_id))
    lock.acquire(wait=120)
    try:
        path = get_template_path(repo_id, template_id)
        if not os.path.exists(path):
            os.makedirs(path)
            try:
                return create_cfg(get_cfg_file(path), params)
            except Exception:
                rmtree(path)
                raise
    finally:
        lock.release()


def delete_template(repo_id, template_id):
    lock = filelock.FileLock(get_vm_lock_file(repo_id, template_id))
    lock.acquire(wait=120)
    try:
        rmtree(get_template_path(repo_id, template_id), raise_exception=True)
    finally:
        lock.release()


def import_xen_vm(repo_id, vm_id, url_list, option=None, is_template=False):
    set_nice()
    if not url_list:
        raise RuntimeError('No URL is provided.')
    if is_template:
        vm_path = get_template_path(repo_id, vm_id)
    else:
        vm_path = get_vm_path(repo_id, vm_id)
    tmp_path = os.path.join(vm_path, 'tmp-%s' % get_uuid())
    txn = begin_transaction(tmp_path)
    try:
        os.makedirs(tmp_path)
        unpacked = []
        obj_total = len(url_list)
        obj_current = 1
        for url in url_list:
            tmp_name = make_filename(url)
            filename = os.path.join(tmp_path, tmp_name)
            filename = ovs_download_file(url, filename, option, 'template',
                                         obj_current, obj_total)
            unpack_files(tmp_path, unpacked, 'template', obj_current, obj_total)
            obj_current += 1
        disks, isos = deploy_xen_vm(vm_id, repo_id, tmp_path, vm_path)
        return {'VirtualDisk': disks, 'ISO': isos}
    except Exception:
        rmtree(vm_path)
        raise
    finally:
        end_transaction(txn)


def deploy_xen_vm(vm_id, repo_id, tmp_path, vm_path):
    tmp_cfg_files = find_file('vm.cfg', tmp_path)
    if not tmp_cfg_files:
        raise RuntimeError('No VM configure file found.')
    if len(tmp_cfg_files) > 1:
        raise RuntimeError('Multiple VM configure files found.')
    tmp_cfg_file = tmp_cfg_files[0]
    vm_cfg_file = get_cfg_file(vm_path)
    create_cfg(vm_cfg_file, {})
    vm_cfg = read_cfg(tmp_cfg_file)
    virtualdisk_deployed = {}
    iso_deployed = {}
    try:
        new_disk = []
        for disk in vm_cfg['disk']:
            old_disk_filepath = disk.split(',')[0].split(':')[-1]
            if old_disk_filepath == '':
                # ',hdc:cdrom,r'
                new_disk.append(disk)
                continue
            old_disk_filename = os.path.basename(old_disk_filepath)
            old_disk_files = find_file(old_disk_filename, tmp_path)
            if not old_disk_files:
                raise RuntimeError('Cannot find disk file: %s' %
                                   old_disk_filename)
            is_cdrom = disk.split(',')[1].endswith(':cdrom')
            if is_cdrom:
                disk_id = '%s.iso' % get_uuid()
                disk_path = get_iso_path(repo_id, disk_id)
                shutil.move(old_disk_files[0], disk_path)
                iso_deployed[disk_id] = old_disk_filename
            else:
                disk_id = '%s.img' % get_uuid()
                disk_path = get_virtualdisk_path(repo_id, disk_id)
                shutil.move(old_disk_files[0], disk_path)
                virtualdisk_deployed[disk_id] = old_disk_filename
            new_disk.append(disk.replace(old_disk_filepath, disk_path))
        vm_cfg['disk'] = new_disk
        vm_cfg['name'] = vm_id
        vm_cfg['uuid'] = vm_id
        write_cfg(vm_cfg_file, vm_cfg)
        rmtree(tmp_path)
        return (virtualdisk_deployed, iso_deployed)
    except Exception:
        for virtualdisk_id in virtualdisk_deployed:
            try:
                os.unlink(get_virtualdisk_path(repo_id, virtualdisk_id))
            except Exception:
                pass
        for iso_id in iso_deployed:
            try:
                os.unlink(get_iso_path(repo_id, iso_id))
            except Exception:
                pass
        raise


def import_vm(repo_id, vm_id, url_list, option=None):
    return import_xen_vm(repo_id, vm_id, url_list, option)


def import_template(repo_id, template_id, url_list, option=None):
    return import_xen_vm(repo_id, template_id, url_list, option, True)


def configure_template(repo_id, template_id, params):
    lock = filelock.FileLock(get_vm_lock_file(repo_id, template_id))
    lock.acquire(wait=120)
    try:
        template_path = get_template_path(repo_id, template_id)
        return configure_cfg(get_cfg_file(template_path), params)
    finally:
        lock.release()


def unconfigure_template(repo_id, template_id, params):
    lock = filelock.FileLock(get_vm_lock_file(repo_id, template_id))
    lock.acquire(wait=120)
    try:
        template_path = get_template_path(repo_id, template_id)
        return unconfigure_cfg(get_cfg_file(template_path), params)
    finally:
        lock.release()


def get_template_config(repo_id, template_id):
    lock = filelock.FileLock(get_vm_lock_file(repo_id, template_id))
    lock.acquire(wait=120)
    try:
        template_path = get_template_path(repo_id, template_id)
        return read_cfg(get_cfg_file(template_path))
    finally:
        lock.release()


def set_template_config(repo_id, template_id, params):
    lock = filelock.FileLock(get_vm_lock_file(repo_id, template_id))
    lock.acquire(wait=120)
    try:
        template_path = get_template_path(repo_id, template_id)
        return write_cfg(get_cfg_file(template_path), params)
    finally:
        lock.release()


def unpack_files(root, unpacked, obj, obj_current, obj_total):
    need_unpack = True
    while need_unpack:
        need_unpack = False
        for (path, dirs, files) in os.walk(root):
            for filename in files:
                filepath = os.path.join(path, filename)
                if filepath in unpacked:
                    continue
                unpack_file(filepath, path, obj, obj_current, obj_total)
                unpacked.append(filepath)
                need_unpack = True


def make_filename(filename=''):
    valid_chars = '_-.' + string.digits + string.ascii_letters
    filename = os.path.basename(filename)
    filename = ''.join([x for x in filename if x in valid_chars])
    filename = filename.lstrip('.')
    if not filename:
        filename = ''.join(random.sample(string.ascii_letters, 6))
    return filename


def find_file(pattern, root=os.curdir):
    result = []
    abs_root = os.path.abspath(root)
    real_root = os.path.realpath(abs_root)
    for (path, dirs, files) in os.walk(real_root):
        for filename in files:
            if fnmatch.fnmatch(filename, pattern):
                abs_path = os.path.join(path, filename)
                real_path = os.path.realpath(abs_path)
                if real_path.startswith(real_root):
                    result.append(real_path)
    return result
