# Copyright (C) 2006-2016 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.

import logging
import os
import re
import signal
import time
import uuid
from xen.xend import sxp
from xen.util.blkif import blkdev_name_to_number
from xen.xend.xenstore.xstransact import xstransact
from xen.xend.xenstore.xsutil import GetDomainPath
from xen.xend.XendConstants import MINIMUM_RESTART_TIME

from agent.lib import filelock
from agent.lib.settings import format_uuid
from agent.lib.xenvm import get_cfg_path, read_cfg, cfg_exists
from agent.target.dlm import dlm_acquire_lock, dlm_release_lock
from agent.target.host import run_cmd
from agent.lib.storage_vm import migrate, cleanup, setup, cfgfile_setup, cfgfile_cleanup, delete_vm

def vm_exists(repo_id, vm_id):
    return cfg_exists(get_cfg_path(repo_id, vm_id))


def parse_domain_info(info_sxp):
    dom_info = {}
    dom_info['device'] = {}
    dom_info['image'] = {}
    for (k, v) in info_sxp:
        if k == 'device':
            if v[0] not in dom_info['device']:
                dom_info['device'][v[0]] = []
            d = {}
            for item in v[1:]:
                if len(item) == 2:
                    d[item[0]] = item[1]
                elif len(item) > 2:
                    d[item[0]] = item[1:]
            dom_info['device'][v[0]].append(d)
        elif k == 'image':
            dom_info['builder'] = v[0]
            for item in v[1:]:
                if len(item) == 2:
                    dom_info['image'][item[0]] = item[1]
                elif len(item) > 2:
                    dom_info['image'][item[0]] = item[1:]
        else:
            dom_info[k] = v
    return dom_info


def get_domid(domain):
    return int(run_cmd(['xm', 'domid', domain]))


def is_vm_running_locally(vm_uuid):
    try:
        if xstransact.List('/vm/%s' % uuid.UUID(vm_uuid)):
            return True
        else:
            return False
    except Exception, err:
        # assume VM is running when xenstore breaks
        logging.error('xenstore list error: %s', err)
        return True


def is_vm_running(repo_id, vm_id):
    try:
        path = get_cfg_path(repo_id, vm_id)
        cfg = read_cfg(path)
        vm_uuid = format_uuid(cfg["uuid"])
        try:
            dlm_acquire_lock(vm_uuid)
            dlm_release_lock(vm_uuid)
            return False
        except:
            return True
    except Exception, e:
        logging.debug("Error checking if VM is running: %s", e)
        # assume it is not running
        return False


def get_hv_info():
    xen_info = {}
    output = run_cmd(['xm', 'info'])
    for line in output.splitlines():
        (attr, value) = line.split(':', 1)
        xen_info[attr.strip()] = value.strip()
    return xen_info


def get_vm(domain):
    lock = filelock.FileLock('/var/run/ovs-agent/vm-list.lock')
    lock.acquire(wait=120)
    try:
        info = run_cmd(['xm', 'list', '--long', domain])
        info_sxp = sxp.from_string(info)
        return parse_domain_info(info_sxp[1:])
    finally:
        lock.release()


def get_vms():
    lock = filelock.FileLock('/var/run/ovs-agent/vm-list.lock')
    lock.acquire(wait=120)
    try:
        info = run_cmd(['xm', 'list', '--long'])
        info_sxp = sxp.all_from_string(info)
        result = []
        for dom_info_sxp in info_sxp:
            dom_info = parse_domain_info(dom_info_sxp[1:])
            result.append(dom_info)
        return result
    finally:
        lock.release()


def create_vm(path):
    lock = filelock.FileLock('/var/run/ovs-agent/vm-start.lock')
    lock.acquire(wait=300)
    try:
        state_file = os.path.join(os.path.dirname(path), 'state')
        if os.path.exists(state_file):
            logging.debug('Starting saved VM will remove the state file.')
            os.unlink(state_file)
        run_cmd(['xm', 'create', path])
    finally:
        lock.release()


def shutdown_vm(domain):
    run_cmd(['xm', 'shutdown', domain])


def destroy_vm(domain):
    run_cmd(['xm', 'destroy', domain])


def reboot_vm(domain, wait=True):
    # Leave at least MINIMUM_RESTART_TIME seconds between two reboots,
    # otherwise xend will destroy the VM to prevent loops.
    try:
        domain_path = GetDomainPath(get_domid(domain))
        vm_path = xstransact.Read('%s/vm' % domain_path)
        start_time = xstransact.Read('%s/start_time' % vm_path)
        elapse = time.time() - float(start_time)
        if elapse < MINIMUM_RESTART_TIME:
            time.sleep(MINIMUM_RESTART_TIME - elapse + 1)
    except Exception, err:
        logging.warn('Error getting start time for VM %s: %s', domain, err)
    if wait:
        run_cmd(['xm', 'reboot', '-w', domain])
    else:
        run_cmd(['xm', 'reboot', domain])


def pause_vm(domain):
    run_cmd(['xm', 'pause', domain])


def unpause_vm(domain):
    run_cmd(['xm', 'unpause', domain])


def save_vm(domain, state_file, checkpoint=False):
    if checkpoint:
        run_cmd(['xm', 'save', '--checkpoint', domain, state_file])
    else:
        run_cmd(['xm', 'save', domain, state_file])


def restore_vm(state_file, paused=False):
    if paused:
        run_cmd(['xm', 'restore', '--paused', state_file])
    else:
        run_cmd(['xm', 'restore', state_file])


def migrate_vm(domain, dest, live=True, ssl=True):
    cmd = ['xm', 'migrate']
    if live:
        cmd += ['--live']
    if ssl:
        cmd += ['--ssl']
    cmd += [domain, dest]
    run_cmd(cmd)


def notify_vm_vdisk(virtual_disk):
    output = run_cmd(['losetup', '-j', virtual_disk])
    for line in output.splitlines():
        loopdev = line.split(':', 1)
        run_cmd(['losetup', '-c', loopdev[0]])


def migrate_vm_with_storage(domain, dest, vmfiles, vmcfg, live=True, ssl=True):
    domid = get_domid(domain)
   # It is not necessary to lock the migrate. The domain is immediately renamed
   # to migrating-domain_name so it will not be available for any other calls.
    migrate(domid, domain, dest, vmfiles, vmcfg, live, ssl)


def storage_migration_setup(vmfiles):
    setup(vmfiles)


def storage_migration_cleanup(vmfiles, delete_vmfiles):
    cleanup(vmfiles, delete_vmfiles)


def storage_migration_cfgfile_setup(repo_uuid, vm_uuid):
    cfgfile_setup(repo_uuid, vm_uuid)


def storage_migration_cfgfile_cleanup(repo_uuid, vm_uuid):
    cfgfile_cleanup(repo_uuid, vm_uuid)


def storage_migration_delete_vm(repo_uuid, vm_uuid):
    delete_vm(repo_uuid, vm_uuid)


def trigger_vm(domain, name, vcpu=None):
    cmd = ['xm', 'trigger', domain, name]
    if vcpu:
        cmd += [vcpu]
    run_cmd(cmd)


def sysrq_vm(domain, letter):
    run_cmd(['xm', 'sysrq', domain, letter])


def cleanup_xc_restore(dom_id):
    try:
        if not dom_id:
            return
        proclist = run_cmd(['ps', '-o', 'pid=,command=', '-C', 'xc_restore'])
        for proc in proclist.splitlines():
            fields = proc.split()
            if fields[1] != '/usr/lib64/xen/bin/xc_restore' or int(fields[3]) != dom_id:
                continue
            xc_restore_pid = int(fields[0])
            logging.info('Found xc_restore process for domain=%s. Kill pid: %s', dom_id, xc_restore_pid)
            os.kill(xc_restore_pid, signal.SIGKILL)
            break
    except Exception, err:
        logging.info('Unable to kill xc_restore for domain: %s. %s', dom_id, err)


def dump_core(domain, core_file, live=False, crash=False, reset=False):
    cmd = ['xm', 'dump-core']
    if live:
        cmd += ['--live']
    if crash:
        cmd += ['--crash']
    if reset:
        cmd += ['--reset']
    cmd += [domain, core_file]
    run_cmd(cmd)


def set_vm_params(domain, params):
    for option in sorted(params.keys()):
        value = params[option]
        if option in VM_PARAMS:
            setter = VM_PARAMS[option]
            try:
                setter(domain, value)
            except Exception, err:
                raise RuntimeError('%s:setup_error:%s' % (option, err))


def set_name(domain, value):
    run_cmd(['xm', 'rename', domain, value])


def set_memory(domain, value):
    run_cmd(['xm', 'mem-set', domain, value])


def set_maxmem(domain, value):
    run_cmd(['xm', 'mem-max', domain, value])


def set_vcpus(domain, value):
    run_cmd(['xm', 'vcpu-set', domain, value])

def set_cpus(domain, value):
    if isinstance(value, str):
        if value:
            run_cmd(['xm', 'vcpu-pin', domain, 'all', value])
        else:
            run_cmd(['xm', 'vcpu-pin', domain, 'all', 'all'])
    else:
        for (i, v) in enumerate(value):
            if v:
                run_cmd(['xm', 'vcpu-pin', domain, i, v])
            else:
                run_cmd(['xm', 'vcpu-pin', domain, i, 'all'])


def set_cpu_cap(domain, value):
    run_cmd(['xm', 'sched-credit', '-d', domain, '-c', value])


def set_cpu_weight(domain, value):
    run_cmd(['xm', 'sched-credit', '-d', domain, '-w', value])


def set_disk(domain, value):
    vbds_old = set()
    cdroms_old = set()
    cdroms_info_old = {}
    info = run_cmd(['xm', 'list', '--long', domain])
    info_sxp = sxp.from_string(info)
    domain_info = parse_domain_info(info_sxp[1:])
    for vbd in domain_info['device']['vbd']:
        front_dev = re.sub(r'^/dev/(.*)', r'\1', vbd['dev'])
        if front_dev.endswith(':cdrom'):
            cdroms_old.add(front_dev)
            cdroms_info_old[front_dev] = [vbd.get('uname', ''), front_dev,
                                          vbd['mode']]
        else:
            vbds_old.add((vbd['uname'], front_dev.split(':')[0], vbd['mode']))

    vbds_new = set()
    cdroms_new = set()
    cdroms_info_new = {}
    for entry in value:
        backend, front_dev, mode = entry.split(',')
        front_dev = re.sub(r'^/dev/(.*)', r'\1', front_dev)
        if front_dev.endswith(':cdrom'):
            cdroms_new.add(front_dev)
            cdroms_info_new[front_dev] = [backend, front_dev, mode]
        else:
            vbds_new.add((backend, front_dev, mode))

    vbds_add = vbds_new - vbds_old
    vbds_remove = vbds_old - vbds_new
    cdroms_add = cdroms_new - cdroms_old
    cdroms_remove = cdroms_old - cdroms_new
    cdroms_changed = cdroms_old & cdroms_new

    for entry in vbds_remove:
        devid = blkdev_name_to_number(entry[1].split(':')[0])[1]
        run_cmd(['xm', 'block-detach', domain, devid])
    for entry in cdroms_remove:
        devid = blkdev_name_to_number(entry.split(':')[0])[1]
        run_cmd(['xm', 'block-detach', domain, devid])
    for entry in vbds_add:
        run_cmd(['xm', 'block-attach', domain] + list(entry))
    for entry in cdroms_add:
        run_cmd(['xm', 'block-attach', domain] + cdroms_info_new[entry])
    for entry in cdroms_changed:
        run_cmd(['xm', 'block-configure', domain] + cdroms_info_new[entry])


def set_vif(domain, value):
    output = run_cmd(['xm', 'network-list', '--long', domain])
    vifs_sxpr = sxp.all_from_string(output)
    vifs_old = set()
    if vifs_sxpr:
        for (vif_id, vif_desc) in vifs_sxpr:
            for (k, v) in vif_desc:
                if k == 'mac':
                    vifs_old.add(v)

    vifs_new = set()
    vifs_new_info = {}
    for entry in value:
        # FIXME: when there's no 'mac' in the entry? Just continue for now.
        m = re.search(r'mac\s*=\s*([a-fA-F0-9:]+)', entry)
        if not m:
            continue
        mac = m.group(1)
        vifs_new.add(mac)
        vifs_new_info[mac] = entry

    add_to = list(vifs_new - vifs_old)
    remove_from = list(vifs_old - vifs_new)
    for entry in remove_from:
        run_cmd(['xm', 'network-detach', domain, entry])
    for entry in add_to:
        run_cmd(['xm', 'network-attach', domain] +
                vifs_new_info[entry].split(','))


def set_disk_other_config(domain, value):
    output = run_cmd(['xm', 'block-list', '--long', domain])
    vbds_sxpr = sxp.all_from_string(output)
    vbds_info = {}
    for (vbd_id, vbd_desc) in vbds_sxpr:
        vbds_info[vbd_id] = {}
        for (k, v) in vbd_desc:
            vbds_info[vbd_id][k] = v
    for (front_end, algorithm, params) in value:
        front_id = blkdev_name_to_number(front_end)[1]
        if front_id in vbds_info and 'backend' in vbds_info[front_id]:
            run_cmd(['xenstore-write', '%s/qos_algorithm_type' %
                     vbds_info[front_id]['backend'], algorithm])
            run_cmd(['xenstore-write', '%s/qos_algorithm_params' %
                     vbds_info[front_id]['backend'], params])


def set_vif_other_config(domain, value):
    output = run_cmd(['xm', 'network-list', '--long', domain])
    vifs_sxpr = sxp.all_from_string(output)
    vifs_info = {}
    if vifs_sxpr:
        for (vif_id, vif_desc) in vifs_sxpr:
            vifs_info[vif_id] = {}
            for (k, v) in vif_desc:
                vifs_info[vif_id][k] = v
    for (mac, algorithm, params) in value:
        for vif in vifs_info.values():
            if 'mac' in vif and 'backend' in vif and vif['mac'] == mac.lower():
                run_cmd(['xenstore-write', '%s/qos_algorithm_type' %
                         vif['backend'], algorithm])
                run_cmd(['xenstore-write', '%s/qos_algorithm_params' %
                         vif['backend'], params])


VM_PARAMS = {
    # option: setter
    'cpu_cap': set_cpu_cap,
    'cpus': set_cpus,
    'cpu_weight': set_cpu_weight,
    'disk_other_config': set_disk_other_config,
    'disk': set_disk,
    'maxmem': set_maxmem,
    'memory': set_memory,
    'name': set_name,
    'vcpus': set_vcpus,
    'vif_other_config': set_vif_other_config,
    'vif': set_vif
}
