#!/usr/bin/env python
#
# Copyright (C) 2011-2016 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.

"""Utilities for acquiring block device and disk information
"""

import os
import re
import shlex
import stat
import logging

from agent.storage.st_tech.blockDevice import (get_empty_blkDevInfo,
                                       fill_blkDevInfo_CCISS_device,
                                       fill_blkDevInfo_scsi_device_basics,
                                       fill_blkDevInfo_scsi_device)
from agent.storage.st_tech.SG_device import do_sg_inq
from agent.target.host import (run_cmd)
from agent.storage.st_tech.multipath import get_multipath_info
from agent.storage.st_tech.openISCSI import get_openiscsi_session_info_dict

__all__ = ['global_numOfDisksSetupAtOnce',
           'get_blk_dev_list',
           'filter_blk_dev_list',
           'filter_blk_dev_list_emc',
           'populate_blk_dev_info_dict',
           'populate_blk_dev_info_dict_emc',
           'fill_blk_dev_info_dict____basic',
           'fill_blk_dev_info_dict____multipath',
           'fill_blk_dev_info_dict____emc']

global_numOfDisksSetupAtOnce = 2000


class DiskInfo_CurState_OBJ:

    def __init__(self):
        self.mWWID = None

    def __repr__(self):
        return "mWWID: %s " % (self.mWWID)


class DiskInfo_OBJ:

    def __init__(self):
        self.mWWID = None
        self.mIsSCSIdevice = None
        self.mCurState = None  # DiskInfo_CurState_OBJ
        self.mBlkDevInfo_LIST = []  # [BlkDevInfo_OBJ]

    def __repr__(self):
        return ("mWWID: %s mIsSCSIdevice: %s mCurState: %s "
                "mBlkDevInfo_LIST: %s" % (
                self.mWWID, self.mIsSCSIdevice, self.mCurState,
                self.mBlkDevInfo_LIST))


def get_blk_dev_list():
    blkdev_LIST = []
    for blkdev in os.listdir('/dev'):
        blkdev_path = "/dev/%s" % (blkdev)
        try:
            mode = os.stat(blkdev_path)[stat.ST_MODE]
            if not stat.S_ISBLK(mode):
                continue
            if os.path.islink(blkdev_path):
                # NOTE: S_ISLNK() cannot be used in this case
                continue
        except StandardError:
            continue
        if not os.path.exists("/sys/block/%s" % (blkdev)):
            # cciss devices fail here but are picked up below
            continue
        blkdev_LIST.append(blkdev)

    if os.path.exists('/dev/cciss'):
        for blkdev in os.listdir('/dev/cciss'):
            blkdev_path = "/dev/cciss/%s" % (blkdev)
            try:
                mode = os.stat(blkdev_path)[stat.ST_MODE]
                if not stat.S_ISBLK(mode):
                    continue
                if os.path.islink(blkdev_path):
                    # NOTE: S_ISLNK() can not be used in this case
                    continue
            except StandardError:
                continue
            if not os.path.exists("/sys/block/cciss!%s" % (blkdev)):
                continue
            blkdev_cciss = "cciss!%s" % (blkdev)
            blkdev_LIST.append(blkdev_cciss)
    return blkdev_LIST


def filter_blk_dev_list(blkdev_LIST, mpConf_blacklistInfo):
    filtered_blkdev_LIST = []
    for blkdev in blkdev_LIST:
        for mpConf_blkLstDevNode in mpConf_blacklistInfo.mDevNode_LIST:
            if mpConf_blkLstDevNode.mDevNode_cmpldRegEx.search(blkdev):
                # match() assumes '^' anchor in pattern
                blkdev = None
                break
        if not blkdev:
            continue
        filtered_blkdev_LIST.append(blkdev)
    return filtered_blkdev_LIST


def filter_blk_dev_list_emc(blkdev_LIST, filter_List):

    filtered_blkdev_LIST = []
    for blkdev in blkdev_LIST:
        found = False
        for item in filter_List:
            if blkdev.startswith(item):
                found = True
                break
        if found:
            filtered_blkdev_LIST.append(blkdev)
    return filtered_blkdev_LIST


def populate_blk_dev_info_dict(diskInfo_DICT,
                               blkdev_LIST, mpConf_default_callout,
                               mpConf_blacklistInfo,
                               mpConf_blacklistExcepInfo,
                               mpConf_devSpec_LIST):
    cciss_blkDev_cmpldRegEx = re.compile('^cciss!')
    for blkdev in blkdev_LIST:
        blkDevInfo = get_empty_blkDevInfo(blkdev)
        if not blkDevInfo:
            continue
        if not cciss_blkDev_cmpldRegEx.match(blkdev):
            if fill_blkDevInfo_scsi_device_basics(blkDevInfo):
                continue
            blkDevInfo.mIsSCSIdevice = True
            blkDevInfo.mDevPath = "/dev/%s" % (blkdev)
        else:
            blkDevInfo.mIsSCSIdevice = False
            blkDevInfo.mDevPath = "/dev/cciss/%s" % (blkdev[6:])
            cciss_sg_inq = do_sg_inq(blkDevInfo.mDevPath)
            if cciss_sg_inq.mSCSI_vendor:
                blkDevInfo.mVendor = cciss_sg_inq.mSCSI_vendor
            else:
                blkDevInfo.mVendor = 'HP'    # only known cciss vendor
            if cciss_sg_inq.mSCSI_model:
                blkDevInfo.mProduct = cciss_sg_inq.mSCSI_model
            else:
                blkDevInfo.mProduct = 'SMART ARRAY'
            if cciss_sg_inq.mSCSI_rev:
                blkDevInfo.mRev = cciss_sg_inq.mSCSI_rev
            else:
                blkDevInfo.mRev = '0.0'
        is_blacklisted = False
        for mpConf_blkLstDev in mpConf_blacklistInfo.mDevice_LIST:
            if not mpConf_blkLstDev.mVendor_cmpldRegEx.match(
                                                blkDevInfo.mVendor):
                continue
            if not mpConf_blkLstDev.mProduct_cmpldRegEx.match(
                                                blkDevInfo.mProduct):
                continue
            is_blacklisted = True
            break
        if is_blacklisted:
            continue
        mpConf_devSpec = None
        for mpConf_devSpec in mpConf_devSpec_LIST:
            if mpConf_devSpec.mVendor_cmpldRegEx:
                if not mpConf_devSpec.mVendor_cmpldRegEx.search(
                                                blkDevInfo.mVendor):
                    # match() assumes '^' anchor in pattern
                    continue
            if mpConf_devSpec.mProduct_cmpldRegEx:
                if not mpConf_devSpec.mProduct_cmpldRegEx.search(
                                                    blkDevInfo.mProduct):
                    # match() assumes '^' anchor in pattern
                    continue
                break
        else:
            mpConf_devSpec = None

        callout = mpConf_default_callout
        if mpConf_devSpec:
            if mpConf_devSpec.mBlacklist:
                if re.search(mpConf_devSpec.mBlacklist, blkDevInfo.mProduct):
                    # match() assumes '^' anchor in pattern
                    continue
            if mpConf_devSpec.mCallout:
                callout = mpConf_devSpec.mCallout
            blkDevInfo.mHasExplicitMPconf = mpConf_devSpec.mIsExplicit
            blkDevInfo.mIsWWIDgenerated = mpConf_devSpec.mIsCalloutGenerated
        if not cciss_blkDev_cmpldRegEx.match(blkdev):
            blkdev_callout = blkdev
        else:
            blkdev_callout = "cciss/%s" % (blkdev[6:])
        final_callout = re.sub('%n', blkdev_callout, callout)
        cmd_LIST = shlex.split(final_callout)
        stdout_str = run_cmd(cmd_LIST, (0, 1))
        if len(stdout_str) > 0:
            wwid = stdout_str.strip()
            if len(wwid) < 1:
                continue
        else:
            continue

        is_blacklisted = False
        for mpConf_blkLstWWID in mpConf_blacklistInfo.mWWID_LIST:
            if (mpConf_blkLstWWID.mWWID_cmpldRegEx and
                not mpConf_blkLstWWID.mWWID_cmpldRegEx.search(wwid)):
                    # match() assumes '^' anchor in pattern
                continue
            for mpConf_blkLstExWWID in mpConf_blacklistExcepInfo.mWWID_LIST:
                if mpConf_blkLstExWWID.mWWID_cmpldRegEx.search(wwid):
                    break
                else:
                    is_blacklisted = True
            break
        if is_blacklisted:
            continue
        blkDevInfo.mWWID = wwid
        try:
            diskInfo = diskInfo_DICT[wwid]
            diskInfo.mBlkDevInfo_LIST.append(blkDevInfo)
        except KeyError:
            diskInfo = DiskInfo_OBJ()
            diskInfo.mWWID = wwid
            diskInfo.mIsSCSIdevice = blkDevInfo.mIsSCSIdevice
            diskInfo.mCurState = DiskInfo_CurState_OBJ()
            diskInfo.mCurState.mWWID = diskInfo.mWWID
            diskInfo.mBlkDevInfo_LIST = [blkDevInfo]
            diskInfo_DICT[wwid] = diskInfo


def populate_blk_dev_info_dict_emc(diskInfo_DICT, blkdev_LIST):
    if not os.path.exists("/dev/emcpower"):
        return
    for blkdev in blkdev_LIST:
        blkDevInfo = get_empty_blkDevInfo(blkdev)
        if not blkDevInfo:
            continue
        if fill_blkDevInfo_scsi_device_basics(blkDevInfo):
            continue
        blkDevInfo.mIsSCSIdevice = True
        blkDevInfo.mDevPath = "/dev/%s" % (blkdev)
        callout = ("/sbin/scsi_id --whitelisted "
                   "--replace-whitespace --device=/dev/%s" % blkdev)
        cmd_LIST = shlex.split(callout)
        stdout_str = run_cmd(cmd_LIST, (0, 1))
        if len(stdout_str) > 0:
            wwid = stdout_str.strip()
            if len(wwid) < 1:
                continue
        else:
            continue
        blkDevInfo.mWWID = wwid
        try:
            diskInfo = diskInfo_DICT[wwid]
            diskInfo.mBlkDevInfo_LIST.append(blkDevInfo)
        except KeyError:
            diskInfo = DiskInfo_OBJ()
            diskInfo.mWWID = wwid
            diskInfo.mIsSCSIdevice = blkDevInfo.mIsSCSIdevice
            diskInfo.mCurState = DiskInfo_CurState_OBJ()
            diskInfo.mCurState.mWWID = diskInfo.mWWID
            diskInfo.mBlkDevInfo_LIST = [blkDevInfo]
            diskInfo_DICT[wwid] = diskInfo


def fill_blk_dev_info_dict____basic(diskInfo_DICT):
    """
    This routine does everything that does not require accessing
    the disk itself
    """
    wwidsToDelete_LIST = []
    openiscsi_session_DICT = get_openiscsi_session_info_dict(False)
    for wwid, diskInfo in diskInfo_DICT.iteritems():
        # Be more efficient than always making a slice copy
        blkDevIdxsToDelete_LIST = []
        if diskInfo.mIsSCSIdevice:
            for idx, blkDevInfo in enumerate(diskInfo.mBlkDevInfo_LIST):
                if fill_blkDevInfo_scsi_device(blkDevInfo, openiscsi_session_DICT, False):
                    logging.warning("Expected disk : %s %s",
                                  blkDevInfo.mDevPath, wwid)
                    blkDevIdxsToDelete_LIST.append(idx)
                    continue
                blkDevInfo.mOperations = None
        else:    # can only be CCISS for now
            for idx, blkDevInfo in enumerate(diskInfo.mBlkDevInfo_LIST):
                if fill_blkDevInfo_CCISS_device(blkDevInfo):
                    logging.warning("Expected CCISS disk : %s %s",
                                  blkDevInfo.mDevPath, wwid)
                    blkDevIdxsToDelete_LIST.append(idx)
                    continue
                blkDevInfo.mOperations = None
        if len(blkDevIdxsToDelete_LIST) < 1:
            continue
        if len(blkDevIdxsToDelete_LIST) == len(diskInfo.mBlkDevInfo_LIST):
            wwidsToDelete_LIST.append(wwid)
            continue
        for idx in reversed(blkDevIdxsToDelete_LIST):
            del(diskInfo.mBlkDevInfo_LIST[idx])
    for wwid in wwidsToDelete_LIST:
        del(diskInfo_DICT[wwid])
    return


def fill_blk_dev_info_dict____multipath(diskInfo_DICT):
    sorted_wwid_LIST = diskInfo_DICT.keys()
    sorted_wwid_LIST.sort()
    wwid_LIST = []
    wwid_LIST_LIST = []
    for wwid in sorted_wwid_LIST:
        wwid_LIST.append(wwid)
        if len(wwid_LIST) > global_numOfDisksSetupAtOnce:
            wwid_LIST_LIST.append(wwid_LIST)
            wwid_LIST = []
    if len(wwid_LIST) > 0:
        wwid_LIST_LIST.append(wwid_LIST)
    for wwid_LIST in wwid_LIST_LIST:
        multipathInfo_DICT = get_multipath_info()
        for wwid in wwid_LIST:  # Set mp state (UP vs DOWN) info
            diskInfo = diskInfo_DICT[wwid]
            for blkDevInfo in diskInfo.mBlkDevInfo_LIST[:]:
                try:
                    blkDev_DICT = multipathInfo_DICT[wwid]
                    mpBlkDevInfo = blkDev_DICT.get(blkDevInfo.mBlkDev, None)
                    if mpBlkDevInfo == None:
                        # Race condition. The devmon calls the same
                        # script to delete from the cache entries about
                        # this block device (and multipath -ll already
                        # has it deleted), but we are still using old
                        # cache data
                        diskInfo.mBlkDevInfo_LIST.remove(blkDevInfo)
                        continue
                    blkDevInfo.mCurState.mPathDMmapState = mpBlkDevInfo.mPathDMmapState
                    blkDevInfo.mCurState.mCheckerState = mpBlkDevInfo.mCheckerState
                    blkDevInfo.mDevMapNum = ("/dev/dm-%s" %
                                              mpBlkDevInfo.mDevMapNum)
                except KeyError:
                    if diskInfo.mIsSCSIdevice:
                        # (unless multipath died) The HWtable inside
                        # multipath blacklisted all of the found blkDevs
                        # for this disk
                        del(diskInfo_DICT[wwid])
                        break
        del(multipathInfo_DICT)


def fill_blk_dev_info_dict____emc(diskInfo_DICT):
    for wwid, diskInfo in diskInfo_DICT.items():
        for blkDevInfo in diskInfo.mBlkDevInfo_LIST[:]:
            blkDevInfo.mDevMapNum = "unavailable"
            cmd_LIST = ['powermt', 'display',
                        'dev=%s' % blkDevInfo.mBlkDev]
            #  pass (0, 1) to run_cmd; allow non-powermt device
            #  to return here without RuntimeError exception
            stdout_str = run_cmd(cmd_LIST, (0, 1))
            if len(stdout_str) > 0:
                stdout_LIST = stdout_str.splitlines(True)
                power_lines_LIST = map(str.strip, stdout_LIST)
                for line in power_lines_LIST:
                    if line.find('Pseudo name=') >= 0:
                        blkDevInfo.mDevMapNum = ("/dev/%s" %
                                                 line.split('=')[1].strip())
                    items = line.split()
                    for idx in range(len(items)):
                        if items[idx] == blkDevInfo.mBlkDev:
                            blkDevInfo.mCurState.mCheckerState = items[idx + 3]
                            blkDevInfo.mCurState.mPathDMmapState = items[idx + 4]
                del(stdout_LIST)
            if blkDevInfo.mDevMapNum == "unavailable":
                del diskInfo_DICT[wwid]
