#!/usr/bin/env python
#
# Copyright (C) 2011-2016 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.

"""Utilities for generating storage XML
"""

import os
import logging
import glob
import re

from agent.storage.st_utils.st_readAndWrite import (read_text_value,
                                                    read_text_lines)
from agent.storage.st_tech.openISCSI import iscsiadm_exe_sync
from agent.storage.st_xml.printXMLutils import (print_beg_XML_tag,
                                                print_end_XML_tag,
                                                print_XML,
                                                print_XML_data,
                                                sys_item_XML,
                                                copy_XML_to_str,
                                                delete_XML)
from agent.storage.st_xml.domHelper import DOMdocHelper_OBJ

__all__ = ['output_disk_XML',
           'io_host_XML']


def output_disk_scan_issues_section_XML(indentXMLlvl, numOfErr=0, numOfWrn=0):
    print_beg_XML_tag(indentXMLlvl, 'DiskScanIssues', None)
    indentXMLlvl = indentXMLlvl + 2
    print_beg_XML_tag(indentXMLlvl,
                      'Errors', "Num=\"%s\"" % (numOfErr), True)
    print_end_XML_tag(indentXMLlvl, 'Errors', True)
    print_beg_XML_tag(indentXMLlvl,
                      'Warnings', "Num=\"%s\"" % (numOfWrn), True)
    print_end_XML_tag(indentXMLlvl, 'Warnings', True)
    indentXMLlvl = indentXMLlvl - 2
    print_end_XML_tag(indentXMLlvl, 'DiskScanIssues')


def disk_info_XML(indentXMLlvl, wwid, diskInfo):
    if len(diskInfo.mBlkDevInfo_LIST) < 1:
        logging.warn("disk '%s' has no blkDevs !?!", wwid)
        return
    sorted_blkDevInfo_LIST = diskInfo.mBlkDevInfo_LIST
    sorted_blkDevInfo_LIST.sort()
    # To get device info: skip passive "ghost" devices; the size will be zero
    for blkDevInfo in sorted_blkDevInfo_LIST:
        if blkDevInfo.mCurState.mSizeInBytes != 0:
            break
    domDocHelper = DOMdocHelper_OBJ(indentXMLlvl, 2)
    domDocHelper.push_level('Disk', {'WWID': wwid, 'TYPE': blkDevInfo.mKind})
    domDocHelper.add_data_elem('Transport', blkDevInfo.mTransport)
    domDocHelper.add_data_elem('Vendor', blkDevInfo.mVendor)
    domDocHelper.add_data_elem('Product', blkDevInfo.mProduct)
    domDocHelper.add_data_elem('Rev', blkDevInfo.mRev)
    domDocHelper.add_data_elem('Size', blkDevInfo.mCurState.mSizeInBytes,
                               {'units': 'bytes'})
    domDocHelper.add_data_elem('ExplicitMPdevCnfg',
                               str(blkDevInfo.mHasExplicitMPconf))
    domDocHelper.add_data_elem('WWIDisGenerated',
                               str(blkDevInfo.mIsWWIDgenerated))
    domDocHelper.add_data_elem('Target', blkDevInfo.mTarget)
    domDocHelper.push_level('Multipath',
                            {'DeviceMap': "%s" % blkDevInfo.mDevMapNum})
    for blkDevInfo in sorted_blkDevInfo_LIST:
        domDocHelper.push_level('Path', {'Dev': blkDevInfo.mBlkDev})
        domDocHelper.add_data_elem('DevPath', blkDevInfo.mDevPath)
        domDocHelper.add_data_elem('Device', blkDevInfo.mDevNums)
        if blkDevInfo.mSCSIdevID:
            domDocHelper.add_data_elem('SCSIdeviceID',
                                       str(blkDevInfo.mSCSIdevID))
        else:
            domDocHelper.add_data_elem('SCSIdeviceID', '')
        domDocHelper.add_data_elem('Target', blkDevInfo.mTarget)
        domDocHelper.add_data_elem('TargetIP', blkDevInfo.mTargetIP)
        domDocHelper.add_data_elem('TargetPort', blkDevInfo.mTargetPort)
        domDocHelper.add_data_elem('TargetIQN', blkDevInfo.mTargetIQN)
        domDocHelper.add_data_elem('TargetTPGT', blkDevInfo.mTargetTPGT)
        domDocHelper.add_data_elem('Path_State',
                                   blkDevInfo.mCurState.mPathDMmapState)
        domDocHelper.add_data_elem('Check_State',
                                   blkDevInfo.mCurState.mCheckerState)
        domDocHelper.pop_level()
    print_XML(0, '\n'.join(domDocHelper.finalize()))


def iscsi_node_info_XML(indentXMLlvl, initiator_name):
    domDocHelper = DOMdocHelper_OBJ(indentXMLlvl, 2)
    domDocHelper.push_level('ISCSI_Node')

    domDocHelper.add_data_elem("Initiatorname", initiator_name)
    print_XML(0, '\n'.join(domDocHelper.finalize()))


def scsi_host_info_XML_single_host(indentXMLlvl, scsi_host_num):
    try:
        scsi_host = "host%s" % (scsi_host_num)
        scsi_host_path = "/sys/class/scsi_host/%s" % (scsi_host)
        sorted_devInfo_LIST = os.listdir(scsi_host_path)
        #NEWKERNEL: lrwxrwxrwx /sys/class/scsi_host/host0 ->
        #    ../../devices/pci0000:00/0000:00:1f.2/host0/scsi_host/host0
        #OLDKERNEL: lrwxrwxrwx /sys/class/scsi_host/host0/device ->
        #    ../../../devices/pci0000:00/0000:00:1f.2/host0
        if os.path.islink(scsi_host_path):
            scsi_host_devLink = os.readlink(scsi_host_path)
            scsi_host_devLink = re.sub(
                        "/%s/scsi_host/%s" % (scsi_host, scsi_host),
                        '', scsi_host_devLink)
            scsi_host_devLink = re.sub('^\.\./\.\./devices', '/sys/devices',
                                       scsi_host_devLink)
        else:
            scsi_host_devLink = os.readlink(
                                "/sys/class/scsi_host/%s/device" % (scsi_host))
            scsi_host_devLink = re.sub(
                                "/%s" % (scsi_host), '', scsi_host_devLink)
            scsi_host_devLink = re.sub('^\.\./\.\./\.\./devices',
                                       '/sys/devices', scsi_host_devLink)
        sorted_devLinkInfo_LIST = os.listdir(scsi_host_devLink)
    except StandardError:
        return

    domDocHelper = DOMdocHelper_OBJ(indentXMLlvl, 2)
    domDocHelper.push_level('SCSI_Host', {'Num': scsi_host_num})
    exclude_LIST = ['ctlreg', 'resource', 'unchecked_isa_dma', 'max_sectors']
    sorted_devInfo_LIST.sort()
    for sys_item in sorted_devInfo_LIST:
        sys_item_XML(domDocHelper, scsi_host_path, sys_item, exclude_LIST)
    domDocHelper.push_level('SysFSDev')
    exclude_LIST = ['config', 'local_cpus', 'modalias',
                    'pools', 'resource', 'vpd']
    sorted_devLinkInfo_LIST.sort()
    for sys_item in sorted_devLinkInfo_LIST:
        sys_item_XML(domDocHelper, scsi_host_devLink, sys_item, exclude_LIST)
    print_XML(0, '\n'.join(domDocHelper.finalize()))


def fc_host_info_XML_single_host(indentXMLlvl, fc_host_num):
    fc_host = "host%s" % (fc_host_num)
    fc_host_path = "/sys/class/fc_host/%s" % (fc_host)
    try:
        sorted_fc_host_LIST = os.listdir(fc_host_path)
    except StandardError:
        return
    domDocHelper = DOMdocHelper_OBJ(indentXMLlvl, 2)
    domDocHelper.push_level('FC_Host', {'Num': fc_host_num})
    exclude_LIST = ['active_fc', 'modalias', 'supported_fc',
                    'tgtid_bind_type', 'max_npiv_vports',
                    'npiv_vports_inuse', 'uevent']
    sorted_fc_host_LIST.sort()
    for sys_item in sorted_fc_host_LIST:
        sys_item_XML(domDocHelper, fc_host_path, sys_item, exclude_LIST)
    print_XML(0, '\n'.join(domDocHelper.finalize()))


def fc_target_info_XML_single_host(indentXMLlvl, fc_target,
                                   fc_target_num, wwpn):
    fc_target_path = "/sys/class/fc_transport/%s" % (fc_target)
    fc_target_devInfo_path = "%s/device/%s" % (fc_target_path, fc_target_num)
    try:
        sorted_fc_target_LIST = os.listdir(fc_target_path)
        sorted_devInfo_LIST = os.listdir(fc_target_devInfo_path)
    except StandardError:
        return

    domDocHelper = DOMdocHelper_OBJ(indentXMLlvl, 2)
    domDocHelper.push_level('FC_Target',
                            {'Target': fc_target_num, 'WWPN': wwpn})
    exclude_LIST = ['active_fc', 'modalias', 'supported_fc',
                    'tgtid_bind_type', 'uevent']
    sorted_fc_target_LIST.sort()
    for sys_item in sorted_fc_target_LIST:
        sys_item_XML(domDocHelper, fc_target_path, sys_item, exclude_LIST)
    domDocHelper.push_level('Device_Info')
    exclude_LIST = [
        'delete', 'dh_state', 'evt_media_change', 'iocounterbits', 'ioerr_cnt',
        'modalias', 'queue_type', 'rev', 'type', 'unload_heads',
        'device_blocked', 'iodone_cnt', 'iorequest_cnt', 'queue_depth',
        'queue_ramp_up_period', 'rescan', 'state', 'timeout', 'uevent',
    ]
    sorted_devInfo_LIST.sort()
    for sys_item in sorted_devInfo_LIST:
        sys_item_XML(domDocHelper, fc_target_devInfo_path,
                     sys_item, exclude_LIST)
    print_XML(0, '\n'.join(domDocHelper.finalize()))


def iscsi_target_info_XML_single_host(indentXMLlvl,
                                      iscsi_session, iscsi_session_num,
                                      iscsi_targetname, iscsi_target_ip):

    iscsi_target_path = "/sys/class/iscsi_session/%s" % (iscsi_session)
    try:
        sorted_iscsi_target_LIST = os.listdir(iscsi_target_path)
    except StandardError:
        return

    domDocHelper = DOMdocHelper_OBJ(indentXMLlvl, 2)
    if iscsi_target_ip:
        domDocHelper.push_level('ISCSI_Target',
            {'Session': iscsi_session_num, 'Target_Ip': iscsi_target_ip})
    else:
        domDocHelper.push_level('ISCSI_Target',
            {'Session': iscsi_session_num, 'WWNN': iscsi_targetname})
    exclude_LIST = ['password', 'username']
    include_LIST = ['tpgt', 'targetname', 'initiatorname']
    sorted_iscsi_target_LIST.sort()
    for sys_item in sorted_iscsi_target_LIST:
        if sys_item in include_LIST:
            sys_item_XML(domDocHelper, iscsi_target_path,
                         sys_item, exclude_LIST)
    print_XML(0, '\n'.join(domDocHelper.finalize()))


def iscsi_host_info_XML_single_host(indentXMLlvl, iscsi_host_num,
                                    initiator_name):
    iscsi_host = "host%s" % (iscsi_host_num)
    iscsi_host_path = "/sys/class/iscsi_host/%s" % (iscsi_host)
    try:
        sorted_iscsi_host_LIST = os.listdir(iscsi_host_path)
    except StandardError:
        return

    domDocHelper = DOMdocHelper_OBJ(indentXMLlvl, 2)
    domDocHelper.push_level('ISCSI_Host', {'Num': iscsi_host_num})
    exclude_LIST = ['hwaddress', 'netdev', 'initiatorname']
    sorted_iscsi_host_LIST.sort()
    for sys_item in sorted_iscsi_host_LIST:
        sys_item_XML(domDocHelper, iscsi_host_path, sys_item, exclude_LIST)
    domDocHelper.add_data_elem("Initiatorname", initiator_name)
    print_XML(0, '\n'.join(domDocHelper.finalize()))


def fc_target_DICT__compare(fc_target_DICT__A, fc_target_DICT__B):
    if fc_target_DICT__A['fc_target'] > fc_target_DICT__A['fc_target']:
        return  1
    if fc_target_DICT__A['fc_target'] < fc_target_DICT__A['fc_target']:
        return -1
    fc_target__A__digits_LIST = fc_target_DICT__A['fc_target_num'].split(':')
    fc_target__B__digits_LIST = fc_target_DICT__B['fc_target_num'].split(':')
    if int(fc_target__A__digits_LIST[0]) > int(fc_target__B__digits_LIST[0]):
        return  1
    if int(fc_target__A__digits_LIST[0]) < int(fc_target__B__digits_LIST[0]):
        return -1
    if int(fc_target__A__digits_LIST[1]) > int(fc_target__B__digits_LIST[1]):
        return  1
    if int(fc_target__A__digits_LIST[1]) < int(fc_target__B__digits_LIST[1]):
        return -1
    if int(fc_target__A__digits_LIST[2]) > int(fc_target__B__digits_LIST[2]):
        return  1
    if int(fc_target__A__digits_LIST[2]) < int(fc_target__B__digits_LIST[2]):
        return -1
    return 0


def sas_target_info_XML_single_host(indentXMLlvl, end_device_str,
                                    target, address):
    sas_target_path = "/sys/class/sas_device/%s" % (end_device_str)
    try:
        sorted_sas_target_LIST = os.listdir(sas_target_path)
    except StandardError:
        return

    domDocHelper = DOMdocHelper_OBJ(indentXMLlvl, 2)
    domDocHelper.push_level('SAS_Target',
                            {'Target': target, 'Address': address})
    exclude_LIST = ['uevent', 'device', ]
    sorted_sas_target_LIST.sort()
    for sys_item in sorted_sas_target_LIST:
        sys_item_XML(domDocHelper, sas_target_path, sys_item, exclude_LIST)
    print_XML(0, '\n'.join(domDocHelper.finalize()))


def ide_cdrom_info_XML_single_drive(indentXMLlvl, ideCDrom_num):
    ideCDrom_path = "/sys/bus/ide/drivers/ide-cdrom/%s" % (ideCDrom_num)
    ideCDrom_blkDev, read_failed = read_text_value(
                                            "%s/drivename" % (ideCDrom_path))
    if read_failed:
        return
    ideCDrom_proc_path = "/proc/ide/%s" % (ideCDrom_blkDev)
    domDocHelper = DOMdocHelper_OBJ(indentXMLlvl, 2)
    domDocHelper.push_level('CD-ROM', {'Name': ideCDrom_blkDev})
    domDocHelper.add_data_elem('IDE_Num', ideCDrom_num)
    sys_item_XML(domDocHelper, ideCDrom_proc_path, 'model', [])
    print_XML(0, '\n'.join(domDocHelper.finalize()))


def sr_scsi_cdrom_info_XML_single_drive(indentXMLlvl, srCDrom_path):
    try:
        sorted_srCDromPossible_LIST = os.listdir(srCDrom_path)
    except StandardError:
        return

    srCDrom_name = None
    srCDrom_cmpldRegEx = re.compile('^block:(sr\d+)$')
    sorted_srCDromPossible_LIST.sort()
    for srCDromPossible in sorted_srCDromPossible_LIST:
        srCDrom_match = srCDrom_cmpldRegEx.match(srCDromPossible)
        if not srCDrom_match:
            continue
        srCDrom_name = srCDrom_match.group(1)
        break
    if not srCDrom_name:
        return
    vendor, read_failed = read_text_value("%s/vendor" % (srCDrom_path))
    if read_failed:
        logging.warn("sr cdrom vendor not found for %s (VANISHED ???)",
                     srCDrom_path)
        return
    model, read_failed = read_text_value("%s/model" % (srCDrom_path))
    if read_failed:
        logging.warn("sr cdrom model not found for %s (VANISHED ???)",
                     (srCDrom_path))
        return
    domDocHelper = DOMdocHelper_OBJ(indentXMLlvl, 2)
    domDocHelper.push_level('CD-ROM', {'Name': srCDrom_name})
    domDocHelper.add_data_elem('Model', "%s %s" % (vendor, model))
    print_XML(0, '\n'.join(domDocHelper.finalize()))


def output_disk_XML(indentXMLlvl, numOfErr, diskInfo_DICT, item_LIST):
    print_beg_XML_tag(indentXMLlvl, 'SCSI', None)
    indentXMLlvl = indentXMLlvl + 2
    if is_default_node_configuration():
        print_XML_data(indentXMLlvl, 'NodeConfiguration', 'DEFAULT')
    wwid_LIST = diskInfo_DICT.keys()
    if item_LIST:
        sorted_wwid_LIST = list(set(item_LIST).intersection(set(wwid_LIST)))
        unknown_wwid = list(set(item_LIST).difference(set(sorted_wwid_LIST)))
        for unknown in unknown_wwid:
            logging.warn("disk does not exist '%s'", unknown)
    else:
        sorted_wwid_LIST = wwid_LIST
    sorted_wwid_LIST.sort()
    for wwid in sorted_wwid_LIST:
        diskInfo = diskInfo_DICT[wwid]
        if diskInfo.mIsSCSIdevice != True:
            continue
        disk_info_XML(indentXMLlvl, wwid, diskInfo)
    indentXMLlvl = indentXMLlvl - 2
    print_end_XML_tag(indentXMLlvl, 'SCSI')
    if os.path.exists('/sys/class/iscsi_session'):
        cmd_LIST = ['iscsiadm', '-m', 'session']
        (stdout_LIST, dummy_stderr_LIST,
         dummy_exitStatus, dummy_deltaTime) = iscsiadm_exe_sync(cmd_LIST)
        adm_session_DICT = {}
        for i, dummy_entry in enumerate(stdout_LIST):
            session = stdout_LIST[i].split()[1]
            session = session.strip(']')
            session = session.strip('[')
            ip = stdout_LIST[i].split()[2]
            ip = ip.split(',')
            adm_session_DICT[session] = ip[0]
        iscsi_session_LIST = os.listdir('/sys/class/iscsi_session')
        iscsi_session_DICT_LIST = []
        for iscsi_session in iscsi_session_LIST:
            iscsi_session_num = iscsi_session.replace('session', '')
            iscsi_targetname, read_failed = read_text_value(
                "/sys/class/iscsi_session/%s/targetname" % (iscsi_session))
            if read_failed:
                continue
            iscsi_session_DICT_LIST.append({'iscsi_session': iscsi_session,
                        'iscsi_session_num': int(iscsi_session_num),
                        'iscsi_targetname': iscsi_targetname,
                        'target_ip': adm_session_DICT.get(iscsi_session_num)})

        iscsi_session_DICT_LIST.sort(iscsi_session_DICT__compare)
        for iscsi_session_DICT in iscsi_session_DICT_LIST:
            iscsi_target_info_XML_single_host(indentXMLlvl,
                                  iscsi_session_DICT['iscsi_session'],
                                  iscsi_session_DICT['iscsi_session_num'],
                                  iscsi_session_DICT['iscsi_targetname'],
                                  iscsi_session_DICT['target_ip'])

    # output disk_scan_issues_section_XML is not longer used
    # but is still printed to keep the meet the requirement that
    # the XML output the same.
    output_disk_scan_issues_section_XML(indentXMLlvl, numOfErr)


def iscsi_session_DICT__compare(iscsi_session_DICT__A, iscsi_session_DICT__B):
    if (iscsi_session_DICT__A['iscsi_targetname'] >
        iscsi_session_DICT__B['iscsi_targetname']):
        return  1
    if (iscsi_session_DICT__A['iscsi_targetname'] <
        iscsi_session_DICT__B['iscsi_targetname']):
        return -1
    if (iscsi_session_DICT__A['iscsi_session_num'] >
        iscsi_session_DICT__B['iscsi_session_num']):
        return  1
    if (iscsi_session_DICT__A['iscsi_session_num'] <
        iscsi_session_DICT__B['iscsi_session_num']):
        return -1
    return 0


def sas_target_DICT__compare(sas_target_DICT__A, sas_target_DICT__B):
    """
    This routine does the comparison for sorting.
    """
    if (sas_target_DICT__A['sas_target_num'] >
        sas_target_DICT__B['sas_target_num']):
        return  1
    if (sas_target_DICT__A['sas_target_num'] <
        sas_target_DICT__B['sas_target_num']):
        return -1
    return 0


def is_default_node_configuration():
    if not os.path.exists('/var/log/DEFAULT_NODE_CONFIGURATION'):
        return False
    return True


def io_host_XML(indentXMLlvl):
    print_beg_XML_tag(indentXMLlvl, 'SCSI', None)
    indentXMLlvl = indentXMLlvl + 2
    if is_default_node_configuration():
        print_XML_data(indentXMLlvl, 'NodeConfiguration', 'DEFAULT')
    if os.path.exists('/sys/class/scsi_host'):
        scsi_host_LIST = os.listdir('/sys/class/scsi_host')
        scsi_host_num_LIST = []
        for scsi_host in scsi_host_LIST:
            scsi_host_num_LIST.append(int(scsi_host.replace('host', '')))
        for scsi_host_num in sorted(scsi_host_num_LIST):
            scsi_host_info_XML_single_host(indentXMLlvl, scsi_host_num)
    if os.path.exists('/sys/class/fc_host'):
        fc_host_LIST = os.listdir('/sys/class/fc_host')
        fc_host_num_LIST = []
        for fc_host in fc_host_LIST:
            fc_host_num_LIST.append(int(fc_host.replace('host', '')))
        for fc_host_num in sorted(fc_host_num_LIST):
            fc_host_info_XML_single_host(indentXMLlvl, fc_host_num)
    if os.path.exists('/sys/class/fc_transport'):
        fc_target_LIST = os.listdir('/sys/class/fc_transport')
        fc_target_DICT_LIST = []
        for fc_target in fc_target_LIST:
            fc_target_wild = fc_target.replace('target', '') + ':*'
            fc_target_dir = glob.glob(
                    "/sys/class/fc_transport/%s/device/%s*" % (fc_target,
                                                               fc_target_wild))
            if not fc_target_dir:
                continue
            fc_target_num = fc_target_dir[0].replace(
                        "/sys/class/fc_transport/%s/device/" % fc_target, "")

            wwpn, read_failed = read_text_value(
                "/sys/class/fc_transport/%s/port_name" % (fc_target))
            if read_failed:
                continue
            wwpn = re.sub('^0x', '', wwpn)
            fc_target_DICT_LIST.append({'fc_target': fc_target,
                                        'fc_target_num': fc_target_num,
                                        'wwpn': wwpn})
        fc_target_DICT_LIST.sort(fc_target_DICT__compare)
        for fc_target_DICT in fc_target_DICT_LIST:
            fc_target_info_XML_single_host(indentXMLlvl,
                fc_target_DICT['fc_target'], fc_target_DICT['fc_target_num'],
                fc_target_DICT['wwpn'])

    (initiator_lines_LIST,
     read_failed) = read_text_lines('/etc/iscsi/initiatorname.iscsi')
    if read_failed:
        initiator_lines_LIST[0] = ''
    initiator_name = initiator_lines_LIST[0].replace('InitiatorName=', '')
    iscsi_node_info_XML(indentXMLlvl, initiator_name)

    if os.path.exists('/sys/class/iscsi_host'):
        (initiator_lines_LIST,
         read_failed) = read_text_lines('/etc/iscsi/initiatorname.iscsi')
        if read_failed:
            initiator_lines_LIST[0] = ''
        initiator_name = initiator_lines_LIST[0].replace('InitiatorName=', '')
        iscsi_host_LIST = os.listdir('/sys/class/iscsi_host')
        iscsi_host_num_LIST = []
        for iscsi_host in iscsi_host_LIST:
            iscsi_host_num_LIST.append(int(iscsi_host.replace('host', '')))
        for iscsi_host_num in sorted(iscsi_host_num_LIST):
            iscsi_host_info_XML_single_host(indentXMLlvl,
                                            iscsi_host_num, initiator_name)

    if os.path.exists('/sys/class/iscsi_session'):
        cmd_LIST = ['iscsiadm', '-m', 'session']
        (stdout_LIST, dummy_stderr_LIST,
         dummy_exitStatus, dummy_deltaTime) = iscsiadm_exe_sync(cmd_LIST)
        adm_session_DICT = {}
        for i, dummy_entry in enumerate(stdout_LIST):
            session = stdout_LIST[i].split()[1]
            session = session.strip(']')
            session = session.strip('[')
            ip = stdout_LIST[i].split()[2]
            ip = ip.split(',')
            adm_session_DICT[session] = ip[0]

        iscsi_session_LIST = os.listdir('/sys/class/iscsi_session')
        iscsi_session_DICT_LIST = []
        for iscsi_session in iscsi_session_LIST:
            iscsi_session_num = iscsi_session.replace('session', '')
            (iscsi_targetname, read_failed) = read_text_value(
                    "/sys/class/iscsi_session/%s/targetname" % (iscsi_session))
            if read_failed:
                continue
            iscsi_session_DICT_LIST.append({'iscsi_session': iscsi_session,
                        'iscsi_session_num': int(iscsi_session_num),
                        'iscsi_targetname': iscsi_targetname,
                        'target_ip': adm_session_DICT.get(iscsi_session_num)})
        iscsi_session_DICT_LIST.sort(iscsi_session_DICT__compare)
        for iscsi_session_DICT in iscsi_session_DICT_LIST:
            iscsi_target_info_XML_single_host(indentXMLlvl,
                iscsi_session_DICT['iscsi_session'],
                iscsi_session_DICT['iscsi_session_num'],
                iscsi_session_DICT['iscsi_targetname'],
                iscsi_session_DICT['target_ip'])
    # <SAS_Host> is the same as <SCSI_Host>
    # <SAS_Target>
    if os.path.exists('/sys/class/sas_device'):
        sas_target_DICT_LIST = []
        sas_device_LIST = os.listdir('/sys/class/sas_device')
        for sas_device in sas_device_LIST:
            scsi_is_sas_expander_match = re.match(
                                            "^end_device-(\d+):(\d+):(\d+)$",
                                            sas_device)
            if scsi_is_sas_expander_match:
                sas_address, read_failed = read_text_value(
                        "/sys/class/sas_device/%s/sas_address" % sas_device)
                sas_address = re.sub('^0x', '', sas_address)
                if read_failed:
                    continue
                sas_target_LIST = os.listdir(
                            '/sys/class/sas_device/%s/device/' % (sas_device))
                for sas_target in sas_target_LIST:
                    if 'target' in sas_target:
                        sas_target_num = sas_target.replace(
                                                        'target', '') + ':0'
                        sas_target_DICT_LIST.append({'sas_device':  sas_device,
                                'sas_target_num': sas_target_num,
                                'sas_address': sas_address})

        sas_target_DICT_LIST.sort(sas_target_DICT__compare)
        for sas_target_DICT in sas_target_DICT_LIST:
            sas_target_info_XML_single_host(indentXMLlvl,
                                            sas_target_DICT['sas_device'],
                                            sas_target_DICT['sas_target_num'],
                                            sas_target_DICT['sas_address'])

    if os.path.exists('/sys/bus/scsi/drivers/sr'):
        scsiCDrom_cmpldRegEx = re.compile('^\d+:\d+:\d+:\d+$')
        try:
            sorted_srCDromPossible_LIST = os.listdir(
                                                    '/sys/bus/scsi/drivers/sr')
            sorted_srCDromPossible_LIST.sort()
            raise StandardError

        except StandardError:
            sorted_srCDromPossible_LIST = []
        for scsiCDromPossible in sorted_srCDromPossible_LIST:
            if not scsiCDrom_cmpldRegEx.match(scsiCDromPossible):
                continue
            srCDrom_path = "/sys/bus/scsi/drivers/sr/%s" % (scsiCDromPossible)
            typeValue, read_failed = read_text_value(
                                                    "%s/type" % (srCDrom_path))
            if read_failed:
                continue
            if int(typeValue) != 5:
                continue
            sr_scsi_cdrom_info_XML_single_drive(indentXMLlvl,
                                                srCDrom_path)

    indentXMLlvl = indentXMLlvl - 2
    print_end_XML_tag(indentXMLlvl, 'SCSI')
    print_beg_XML_tag(indentXMLlvl, 'IDE', None)
    indentXMLlvl = indentXMLlvl + 2
    if os.path.exists('/sys/bus/ide/drivers/ide-cdrom'):
        ideCDrom_cmpldRegEx = re.compile('^\d+.\d+$')
        try:
            sorted_ideCDromPossible_LIST = os.listdir(
                                            '/sys/bus/ide/drivers/ide-cdrom')
            sorted_ideCDromPossible_LIST.sort()
        except StandardError:
            sorted_ideCDromPossible_LIST = []
        ideCDrom_num_LIST = []
        for ideCDromPossible in sorted_ideCDromPossible_LIST:
            if ideCDrom_cmpldRegEx.match(ideCDromPossible):
                ideCDrom_num_LIST.append(float(ideCDromPossible))
        for ideCDrom_num in sorted(ideCDrom_num_LIST):
            ide_cdrom_info_XML_single_drive(indentXMLlvl, ideCDrom_num)

    indentXMLlvl = indentXMLlvl - 2
    print_end_XML_tag(indentXMLlvl, 'IDE')

    doc_str = copy_XML_to_str()
    delete_XML()

    return doc_str
