#!/usr/bin/env python
#
# Copyright (C) 2011-2016 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.

"""Utilities for dealing with the SCSI Generic device.
"""

import re

from agent.target.host import (run_cmd)

__all__ = ['do_sg_inq']


class SG_device_State_OBJ:

    mSCSI_device_id = None

    def __init__(self):
        self.mDeviceName = None
        # Optional. If set, this SCSI device has a block device.
        self.mBlkName = None

        self.mPQual = -1    # Values obtained from SCSI INQ.
        self.mTGPS = -1
        self.mRDAC = -1
        self.mMultiP = -1

        self.mSCSI_vendor = None  # Values from SCSI INQ cmd
        self.mSCSI_model = None
        self.mSCSI_rev = None
        self.mSCSI_type = -1
        self.mSCSI_size = -1
        self.mSysFS_model = None  # vs values from sysfs
        self.mSysFS_vendor = None
        self.mSysFS_rev = None
        self.mSysFS_type = -1
        self.mSysFS_size = -1

        self.mSCSI_device_id = None  # SCSI_device_ID_OBJ

    def __repr__(self):
        return ("mDeviceName: %s mBlkName: %s mPQual: %d mTGPS: %d mRDAC: %d "
                "mMultiP: %d (SCSI: mVendor: %s mModel: %s, mRev: %s, "
                "mType: %d,mSize: %d;SysFS: mVendor: %s, mModel: %s, "
                "mRev: %s, mType: %d, mSize: %d) mSCSI_Dev_ID: %s: ") % (
                self.mDeviceName,
                self.mBlkName,
                self.mPQual,
                self.mTGPS,
                self.mRDAC,
                self.mMultiP,
                self.mSCSI_vendor,
                self.mSCSI_model,
                self.mSCSI_rev,
                self.mSCSI_type,
                self.mSCSI_size,
                self.mSysFS_vendor,
                self.mSysFS_model,
                self.mSysFS_rev,
                self.mSysFS_type,
                self.mSysFS_size,
                self.mSCSI_device_id.as_str())

    def __str__(self):
        """
        Output similar to what 'lsscsi' produces.
        """

        def figure_out_type(scsi_type):
            device_type_LIST = ['disk', 'tape', 'printer', 'process',
                                'worm', 'cd/dvd', 'scanner', 'optical',
                                'mediumx', 'comms', '(0xa)', '(0xb)',
                                'storage', 'enclosu', 'sim dsk', 'opti rd',
                                'bridge', 'osd', 'adi', '(0x13)', '(0x14)',
                                '(0x15)', '(0x16)', '(0x17)', '(0x18)',
                                '(0x19)', '(0x1a)', '(0x1b)', '(0x1c)',
                                '(0x1e)', 'wlun', 'no dev']
            if scsi_type > len(device_type_LIST):
                return "unknown"
            return device_type_LIST[scsi_type].strip()

        def if_same_pick_second(first, second):
            if first == None:
                return second
            if second == None:
                return first
            if first in second:
                return second
            else:
                return "mismatch"

        #[2:0:0:0]    disk    ATA      WDC WD1600JD-75H 08.0  /dev/sda
        if self.mSCSI_device_id:
            string = "[%s]" % (self.mSCSI_device_id.as_str())
            string = "%-14s" % (str)
        else:
            string = ''
        # And now the rest of the values.
        if self.mSysFS_type == -1:
            string += " %-7s" % (figure_out_type(self.mSCSI_type))
        else:
            if self.mSCSI_type == self.mSysFS_type:
                string += " %-7s" % (figure_out_type(self.mSCSI_type))
            else:
                string += "mismatch"
        # COMPELNT
        string += (" %-8s" %
                (if_same_pick_second(self.mSysFS_vendor, self.mSCSI_vendor)))
        # WDC WD2500SD-01K
        string += (" %-16s" %
                   (if_same_pick_second(self.mSysFS_model, self.mSCSI_model)))
        string += (" %-4s" %
                   (if_same_pick_second(self.mSysFS_rev, self.mSCSI_rev)))
        if self.mBlkName:
            string += " /dev/%s" % (self.mBlkName)
        else:
            string += " /dev/%s" % (self.mDeviceName)
        return string


def extract_from_sg_inq_data(sg_device, lines):
    """
    Extract from the SCSI INQ command data information.
    """
    def __handle_PQual(sg_device, val=None):
        if  val:
            sg_device.mPQual = int(val)

    def __handle_Type(sg_device, val=None):
        if  val:
            sg_device.mSCSI_type = int(val)

    def __handle_TGPS(sg_device, val=None):
        if  val:
            sg_device.mTGPS = int(val)

    def __handle_MultiP(sg_device, val=None):
        if  val:
            sg_device.mMultiP = int(val)

    def __handle_Vendor(sg_device, val=None):
        if  val:
            sg_device.mSCSI_vendor = val.strip()

    def __handle_Model(sg_device, val=None):
        if val:
            sg_device.mSCSI_model = val.strip()

    def __handle_Rev(sg_device, val=None):
        if val:
            sg_device.mSCSI_rev = val.strip()

    def __handle_RDAC(sg_device, val=None):
        if val:
            sg_device.mRDAC = 1

    def __handle_Size(sg_device, val=None):
        if val:
            if (int(val) >= 512):
                sg_device.mSCSI_size = int(val) / 512
            else:
                sg_device.mSCSI_size = 0

    pattern_DICT = {".*PQual=(\d+).*": __handle_PQual,
                    ".*Device_type=(\d+).*": __handle_Type,
                    ".*TGPS=(\d+).*": __handle_TGPS,
                    ".*MultiP=(\d+).*": __handle_MultiP,
                    ".*Vendor identification: (.*)": __handle_Vendor,
                    ".*Product identification: (.*)": __handle_Model,
                    ".*Product revision level: (.*)": __handle_Rev,
                    ".*Volume Access Control (RDAC)": __handle_RDAC,
                    ".*Device size: (\d+) bytes.*": __handle_Size,
                   }
    # Construct the patterns
    pats_LINE = []
    for pat, handler in pattern_DICT.iteritems():
        pats_LINE.append((re.compile(pat), handler))

    # Start the match.
    for line in lines:
        for pat, handler in pats_LINE:
            m = pat.match(line)
            if m:
                if callable(handler):
                    data = m.group(1)
                    handler(sg_device, data)


def do_sg_inq(blkDevPath):
    """
    Extract the device information from a SCSI Generic INQ
    command and return a SG_device_State_OBJ with that device information.

    Returns:
    SG_device_State_OBJ or None
    """
    cmd_LIST = ['sg_inq', blkDevPath]
    stdout_str = run_cmd(cmd_LIST)
    stdout_LIST = stdout_str.splitlines(True)
    if len(stdout_LIST) < 1:
        return None
    sg_dev_state = SG_device_State_OBJ()
    extract_from_sg_inq_data(sg_dev_state, stdout_LIST)
    return sg_dev_state
