#!/usr/bin/env python
#
# Copyright (C) 2011-2016 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.

"""Utilities for dealing with block devices
"""

import os
import re
import traceback
import logging

from agent.storage.st_utils.st_readAndWrite import read_text_value
from agent.storage.st_tech.CCISS import (find_particular_cciss_disk,
                                         CCISS_BlkDev_Operations)
from agent.storage.st_tech.scsiDeviceID import gen_SCSI_device_ID_OBJ_from_str
from agent.storage.st_tech.openISCSI import openISCSI_BlkDev_Operations
from agent.storage.st_tech.QLogicHBA import get_qla4xxx_info_dict
from agent.storage.st_tech.blockDeviceOperations import Std_BlkDev_Operations

__all__ = ['get_empty_blkDevInfo',
           'fill_blkDevInfo_scsi_device',
           'fill_blkDevInfo_scsi_device_basics',
           'fill_blkDevInfo_CCISS_device',
           'fill_blkDevInfo_AOE_device']


class BlkDev_CurState_OBJ:

    def __init__(self):
        self.mBlkDev = None
        self.mSizeInBytes = None
        self.mPathDMmapState = None
        self.mCheckerState = None

    def is_valid(self):
        if type(self.mBlkDev) == None:
            return False
        if type(self.mSizeInBytes) == None:
            return False
        if type(self.mPathDMmapState) == None:
            return False
        if type(self.mCheckerState) == None:
            return False
        return True

    def __repr__(self):
        return ("mBlkDev: %s mSizeInBytes: %s mPathDMmapState: %s "
                "mCheckerState: %s") % (self.mBlkDev,
                                        self.mSizeInBytes,
                                        self.mPathDMmapState,
                                        self.mCheckerState)


class BlkDevInfo_OBJ:

    def __init__(self):
        self.mBlkDev = None
        self.mDevPath = None
        self.mWWID = None
        self.mVendor = None
        self.mProduct = None
        self.mRev = None
        self.mHasExplicitMPconf = False
        self.mIsWWIDgenerated = False
        self.mIsSCSIdevice = None
        self.mDevNums = None
        self.mSCSIdevID = None  # optional SCSI_device_ID_OBJ()
        self.mDriver = None
        self.mKind = None
        self.mTransport = None
        # for FC this is the HBA nr: '0x12...' for iSCSI this is the target IQN
        self.mTarget = None
        self.mTargetIP = None
        # for FC this is the HBA's port ID. For iSCSI target's
        # port (usually 3260)
        self.mTargetPort = None
        self.mTargetIQN = None  # the same as the mTarget for iSCSI
        self.mTargetTPGT = -1  # 'int' value.
        self.mSpecificTransport = None
        self.mCurState = None  # BlkDev_CurState_OBJ()
        self.mOperations = None  # optional BlkDev_Operations subclass
        self.mSGOperations = None  # optional SG_device_Operations class.
        self.mDevMapNum = None

    def is_valid(self):
        if type(self.mBlkDev) == None:
            return False
        if type(self.mDevPath) == None:
            return False
        if type(self.mDevNums) == None:
            return False
        if type(self.mWWID) == None:
            return False
        if type(self.mVendor) == None:
            return False
        if type(self.mProduct) == None:
            return False
        if type(self.mRev) == None:
            return False
        if type(self.mIsSCSIdevice) == None:
            return False
        if self.mIsSCSIdevice and not self.mSCSIdevID.is_valid():
            return False
        if type(self.mDriver) == None:
            return False
        if type(self.mKind) == None:
            return False
        if type(self.mTransport) == None:
            return False
        if type(self.mTarget) == None:
            return False
        if type(self.mTargetIP) == None:
            return False
        if type(self.mTargetPort) == None:
            return False
        if type(self.mTargetIQN) == None:
            return False
        if type(self.mTargetTPGT) == None:
            return False
        if type(self.mSpecificTransport) == None:
            return False
        if type(self.mCurState) == None:
            return False
        if not self.mCurState.is_valid():
            return False
        return True

    def __repr__(self):
        return ("mBlkDev: %s mDevPath: %s mWWID: %s mDevNums: %s mVendor: %s "
               "mProduct: %s mRev: %s mIsSCSIdevice: %s mSCSIdevID: %s "
               "mDriver: %s mKind: %s mTransport: %s mTarget: %s "
               "mTargetIP: %s mTargetPort: %s mTargetIQN: %s "
               "mTargetTPGT: %d mSpecificTransport: %s mCurState: %s "
               "mOperations: %s mSGOperations: %s") % (self.mBlkDev,
                self.mDevPath,
                self.mWWID,
                self.mDevNums,
                self.mVendor,
                self.mProduct,
                self.mRev,
                self.mIsSCSIdevice,
                self.mSCSIdevID,
                self.mDriver,
                self.mKind,
                self.mTransport,
                self.mTarget,
                self.mTargetIP,
                self.mTargetPort,
                self.mTargetIQN,
                self.mTargetTPGT,
                self.mSpecificTransport,
                self.mCurState,
                self.mOperations,
                self.mSGOperations)

    def __cmp__(self, other):
        if not isinstance(other, BlkDevInfo_OBJ):
            return -1
        if not self.mSCSIdevID or not other.mSCSIdevID:
            if self.mSCSIdevID:
                return 1
            if other.mSCSIdevID:
                return -1
            return cmp(self.mBlkDev, other.mBlkDev)
        if self.mSCSIdevID.mHost > other.mSCSIdevID.mHost:
            return  1
        if self.mSCSIdevID.mHost < other.mSCSIdevID.mHost:
            return -1
        if self.mSCSIdevID.mChannel > other.mSCSIdevID.mChannel:
            return  1
        if self.mSCSIdevID.mChannel < other.mSCSIdevID.mChannel:
            return -1
        if self.mSCSIdevID.mSCSIid > other.mSCSIdevID.mSCSIid:
            return  1
        if self.mSCSIdevID.mSCSIid < other.mSCSIdevID.mSCSIid:
            return -1
        if self.mSCSIdevID.mLUN > other.mSCSIdevID.mLUN:
            return  1
        if self.mSCSIdevID.mLUN < other.mSCSIdevID.mLUN:
            return -1
        return 0


def fill_blkDevInfo_CCISS_device(blkDevInfo):
    cciss_disk = find_particular_cciss_disk(blkDevInfo.mBlkDev)
    if not cciss_disk:
        return -1

    blkDevInfo.mDevPath = cciss_disk.mDevPath
    blkDevInfo.mProduct = cciss_disk.mModel
    blkDevInfo.mRev = cciss_disk.mFirmwareRev
    blkDevInfo.mIsSCSIdevice = False
    blkDevInfo.mSCSIdevID = ''
    blkDevInfo.mDriver = 'cciss'
    blkDevInfo.mKind = 'LOCAL'
    blkDevInfo.mTransport = 'SAS_cciss'
    blkDevInfo.mTarget = ''
    blkDevInfo.mTargetPort = ''
    blkDevInfo.mTargetIP = ''
    blkDevInfo.mTargetIQN = ''
    blkDevInfo.mTargetTPGT = -1
    blkDevInfo.mSpecificTransport = cciss_disk.mInterface
    blkDevInfo.mOperations = CCISS_BlkDev_Operations(blkDevInfo, cciss_disk)
    return 0


def fill_blkDevInfo_scsi_device_basics(blkDevInfo):

    blkdev_info_path = "/sys/block/%s/device/vendor" % (blkDevInfo.mBlkDev)
    blkDevInfo.mVendor, dummy_read_failed = read_text_value(blkdev_info_path)

    blkdev_info_path = "/sys/block/%s/device/model" % (blkDevInfo.mBlkDev)
    blkDevInfo.mProduct, dummy_read_failed = read_text_value(blkdev_info_path)

    blkdev_info_path = "/sys/block/%s/device/rev" % (blkDevInfo.mBlkDev)
    blkDevInfo.mRev, dummy_read_failed = read_text_value(blkdev_info_path)

    return 0


def fill_blkDevInfo_scsi_device(blkDevInfo, openiscsi_session_info_DICT, refresh):
    blkDevInfo.mIsSCSIdevice = True
    blkdev_info_path = "/sys/block/%s" % (blkDevInfo.mBlkDev)
    try:
        blkdev_devLink = os.readlink(blkdev_info_path)
    except StandardError:
        logging.warn("blkdev '%s' vanished [%s] !?!",
                      blkDevInfo.mBlkDev, blkdev_info_path)
        return 1

    try:
        devLink = blkdev_devLink.split("/block")
        blkdev_devLink = devLink[0]
        theSCSIid = re.sub('^.*/', '', blkdev_devLink)
        blkDevInfo.mSCSIdevID = gen_SCSI_device_ID_OBJ_from_str(theSCSIid)
        if not blkDevInfo.mSCSIdevID:
            logging.error(("blkdev '%s' bad devLink [%s] !?!",
                           (blkDevInfo.mBlkDev, blkdev_devLink)))
            return 1

        targetID = ("target%s:%s:%s" %
                    (blkDevInfo.mSCSIdevID.mHost,
                     blkDevInfo.mSCSIdevID.mChannel,
                     blkDevInfo.mSCSIdevID.mSCSIid))
        devLink_match = re.match("[\./]*(.*)/host%s/(.*)%s/%s$" % (
                                 blkDevInfo.mSCSIdevID.mHost,
                                 targetID,
                                 blkDevInfo.mSCSIdevID),
                                 blkdev_devLink)

        blkdev_info_path = '/sys/' + devLink_match.group(1) + '/driver'
        if os.path.exists(blkdev_info_path):
            blkdev_drvrLink = os.readlink(blkdev_info_path)
            blkDevInfo.mDriver = re.sub('^.*/', '', blkdev_drvrLink)
        else:
            blkDevInfo.mDriver = '???'

        blkDevInfo.mKind = 'LOCAL'
        blkDevInfo.mTransport = 'LOCAL'
        blkDevInfo.mTarget = ''
        blkDevInfo.mTargetIP = ''
        blkDevInfo.mTargetPort = ''
        blkDevInfo.mTargetIQN = ''
        blkDevInfo.mTargetTPGT = -1
        blkDevInfo.mSpecificTransport = ''
        specificTransport = devLink_match.group(2).rstrip('/')

        if len(specificTransport) > 0:
            specificTransport_match = re.match('^session(\d+)$',
                                               specificTransport)
            if specificTransport_match:
                iscsi_sessionNum = int(specificTransport_match.group(1))
                blkDevInfo.mDriver = 'iscsi_tcp'
                blkDevInfo.mKind = 'SAN'
                blkDevInfo.mTransport = 'iSCSI'
                blkDevInfo.mSpecificTransport = specificTransport
                openISCSIstate = openiscsi_session_info_DICT[iscsi_sessionNum]
                blkDevInfo.mTargetPort = str(openISCSIstate.mTargetPort)
                blkDevInfo.mTarget = openISCSIstate.mIQN
                blkDevInfo.mTargetIP = openISCSIstate.mTargetIP
                blkDevInfo.mTargetIQN = openISCSIstate.mIQN
                blkDevInfo.mTargetTPGT = openISCSIstate.mTpgt
                blkDevInfo.mOperations = openISCSI_BlkDev_Operations(
                                                    blkDevInfo, openISCSIstate)

            specificTransport_match = re.match("^rport-%s:%s-(\d+)$" % (
                                               blkDevInfo.mSCSIdevID.mHost,
                                               blkDevInfo.mSCSIdevID.mChannel),
                                               specificTransport)
            if specificTransport_match:
                blkDevInfo.mKind = 'SAN'
                blkDevInfo.mTransport = 'FC'
                blkDevInfo.mSpecificTransport = specificTransport
                blkDevInfo.mTargetIQN = ''
                blkDevInfo.mTargetTPGT = -1
                blkdev_info_path = ("/sys/class/fc_transport/%s/node_name" %
                                    (targetID))
                blkDevInfo.mTarget, read_failed = read_text_value(
                                                            blkdev_info_path)
                if read_failed:
                    logging.warn("blkdev '%s' vanished [%s] !?!",
                                    blkDevInfo.mBlkDev, blkdev_info_path)
                    return 1
                blkdev_info_path = ("/sys/class/fc_transport/%s/port_name" %
                                    (targetID))
                blkDevInfo.mTargetPort, read_failed = read_text_value(
                                                            blkdev_info_path)
                if read_failed:
                    logging.warn("blkdev '%s' vanished [%s] !?!",
                                 blkDevInfo.mBlkDev, blkdev_info_path)
                    return 1

            # HBA->SAS disk are 'end_device-<SCSI Host Nr>:<Port ID>'.
            # This is non-SAS SAN setup
            # HBA->SAS Expander->SAS disk are
            # 'end_device-<SCSI Host Nr>:<SCSI Id>:<Port Id>'.
            # This is SAS SAN setup.
            specificTransport_match = re.match(
                                        '.*end_device-(\d+):(\d+):(\d+)$',
                                        specificTransport)
            if specificTransport_match:
                end_device = re.sub('^.*end_device-', 'end_device-',
                                    specificTransport)
                blkdev_info_path = (
                            "/sys/class/sas_device/%s/enclosure_identifier" %
                            (end_device))
                blkDevInfo.mTarget, read_failed = read_text_value(
                                                            blkdev_info_path)
                if not read_failed:
                    # Local HBA's address:
                    blkdev_info_path = (
                                    "/sys/class/sas_device/%s/sas_address" %
                                    (end_device))
                    blkDevInfo.mTargetPort, read_failed = read_text_value(
                                                            blkdev_info_path)
                    if not read_failed:
                        # If we fail with setting those values,
                        # we will just fail-back to the LOCAL storage view.
                        blkDevInfo.mKind = 'SAN'
                        blkDevInfo.mTransport = 'SAS'
                        blkDevInfo.mSpecificTransport = specificTransport
                        blkDevInfo.mTargetIQN = ''
                        blkDevInfo.mTargetTPGT = -1
        else:
            if blkDevInfo.mDriver == 'qla4xxx':
                blkDevInfo.mKind = 'SAN'
                blkDevInfo.mTransport = 'iSCSI'
                blkDevInfo.mSpecificTransport = 'HW_iSCSI'
                try:
                    qla4xxx_state_info_DICT = get_qla4xxx_info_dict(refresh)
                    qla4xxx_state = qla4xxx_state_info_DICT[
                                                blkDevInfo.mSCSIdevID.mSCSIid]
                    blkDevInfo.mTargetPort = str(qla4xxx_state.mTargetPort)
                    blkDevInfo.mTarget = qla4xxx_state.mIQN
                    blkDevInfo.mTargetIP = qla4xxx_state.mTargetIP
                    blkDevInfo.mTargetIQN = qla4xxx_state.mIQN
                    blkDevInfo.mTargetTPGT = qla4xxx_state.mTpgt
                except KeyError:
                    pass
    except StandardError:
        logging.error("blkdev '%s' bad devLink [%s] !?!\n%s",
                blkDevInfo.mBlkDev, blkdev_devLink, traceback.format_exc())
        return 1

    if not blkDevInfo.mOperations:
        blkDevInfo.mOperations = Std_BlkDev_Operations(blkDevInfo)
    return 0


def get_empty_blkDevInfo(blkdev):
    blkdev_info_path = "/sys/block/%s" % (blkdev)
    if not os.path.exists(blkdev_info_path):
        return None
    blkDevInfo = BlkDevInfo_OBJ()
    blkDevInfo.mBlkDev = blkdev
    blkDevInfo.mDevPath = "/dev/%s" % (blkdev)
    blkDevInfo.mCurState = BlkDev_CurState_OBJ()
    blkDevInfo.mCurState.mBlkDev = blkdev
    blkdev_info_path = "/sys/block/%s/dev" % (blkdev)
    blkDevInfo.mDevNums, read_failed = read_text_value(blkdev_info_path)
    if read_failed:
        logging.warn("blkdev '%s' vanished [%s] !?!",
                     blkdev, blkdev_info_path)
        return None
    blkdev_info_path = "/sys/block/%s/size" % (blkdev)
    blkdev_sizeInSectors, read_failed = read_text_value(blkdev_info_path)
    if read_failed:
        logging.warn("blkdev '%s' vanished [%s] !?!",
                     blkdev, blkdev_info_path)
        return None
    blkDevInfo.mCurState.mSizeInBytes = int(blkdev_sizeInSectors) * 512
    return blkDevInfo
