#!/usr/bin/env python
#
# Copyright (C) 2011-2016 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.

"""Utilities for dealing with block device operations
"""

import logging
import os
import time

from agent.target.host import (run_cmd)
from agent.storage.st_utils.st_readAndWrite import (read_text_value,
                                                    write_text_value)

__all__ = ['BlkDev_Operations',
           'Std_BlkDev_Operations',
           'std_disconnect_block_device',
           'std_revalidate_block_device']


class BlkDev_Operations:

    """ABSTRACT class for transport specific subclasses
    """

    def __repr__(self):
        logging.error("BlkDev_Operations __repr__ not implemented")
        raise Exception("BlkDev_Operations __repr__ not implemented")

    def disconnect_block_device(self):
        logging.error(
                "BlkDev_Operations disconnect_block_device not implemented")
        raise Exception(
               "BlkDev_Operations disconnect_block_device not implemented")

    def revalidate_block_device(self):
        logging.error(
                "BlkDev_Operations revalidate_block_device not implemented")
        raise Exception(
                "BlkDev_Operations revalidate_block_device not implemented")


class Std_BlkDev_Operations(BlkDev_Operations):

    """Standard class for BlkDev_Operations
    """

    def __init__(self, blkDevInfo):

        if (not blkDevInfo.mIsSCSIdevice or
            not blkDevInfo.mSCSIdevID.is_valid()):
            raise Exception("given blkDevInfo %s is SCSIDevID is bad !?!" %
                               (blkDevInfo.mBlkDev))

        self.mBlkDevInfo = blkDevInfo

    def __repr__(self):

        return "Std_BlkDev_Operations[%s]" % (self.mBlkDevInfo.mBlkDev)

    def disconnect_block_device(self):

        return std_disconnect_block_device(self.mBlkDevInfo)

    def revalidate_block_device(self):

        return std_revalidate_block_device(self.mBlkDevInfo)


def std_disconnect_block_device(blkDevInfo):
    """
    Disconnects a standard block device by using the sysfs filesystem.
    To acquire the block device back the user would need to pass "- - -"
    to /sys/class/scsi_host/hostX/scan.
    """
    blkDev_delete_path = ("/sys/bus/scsi/drivers/sd/%s/delete" %
                          (blkDevInfo.mSCSIdevID.as_str()))
    if not os.path.exists(blkDev_delete_path):
        return 0

    # Flush out the dirty blocks. This is imperative otherwise we
    # would have I/O in the dirty cache which hasn't yet reached the SCSI ml.
    cmd_LIST = ['sync']
    run_cmd(cmd_LIST)

    # The deletion will flush out any outstanding blocks that have hit the
    # SCSI middle layer. It will also invoke an SCSI SYNC command.
    write_text_value(blkDev_delete_path, '1', False)

    run_cmd(cmd_LIST)

    # This is the kicker.. /dev/<something> doesn't immediately
    # disappear. Try up to 10 seconds
    mRetries = 10
    while (mRetries > 0):
        if not os.path.exists(blkDevInfo.mDevPath):
            break
        time.sleep(1)
        mRetries -= 1

    if os.path.exists(blkDevInfo.mDevPath):
        logging.error("Block device disconnect failed : %s",
                      blkDevInfo.mDevPath)
        return 1

    return 0


def std_revalidate_block_device(blkDevInfo):
    """
    Updates a standard block device to the latest sysfs values
    """
    blkdev_info_path = "/sys/block/%s" % (blkDevInfo.mBlkdev)
    if not os.path.exists(blkdev_info_path):
        return 1

    blkdev_info_path = "/sys/block/%s/dev" % (blkDevInfo.mBlkdev)
    blkDevInfo.mDevNums, read_failed = read_text_value(blkdev_info_path)
    if read_failed:
        logging.warn("blkdev '%s' vanished [%s] !?!",
                     blkDevInfo.mBlkdev, blkdev_info_path)
        return 1

    blkdev_info_path = "/sys/block/%s/size" % (blkDevInfo.mBlkdev)
    blkdev_sizeInSectors, read_failed = read_text_value(blkdev_info_path)
    if read_failed:
        logging.warn("blkdev '%s' vanished [%s] !?!",
                     blkDevInfo.mBlkdev, blkdev_info_path)
        return 1

    blkDevInfo.mCurState.mSizeInBytes = int(blkdev_sizeInSectors) * 512

    blkdev_info_path = "/sys/block/%s/device/vendor" % (blkDevInfo.mBlkdev)
    blkDevInfo.mVendor, read_failed = read_text_value(blkdev_info_path)
    if read_failed:
        logging.warn("blkdev '%s' vanished [%s] !?!",
                     blkDevInfo.mBlkdev, blkdev_info_path)
        return 1

    blkdev_info_path = "/sys/block/%s/device/model" % (blkDevInfo.mBlkdev)
    blkDevInfo.mProduct, read_failed = read_text_value(blkdev_info_path)
    if read_failed:
        logging.warn("blkdev '%s' vanished [%s] !?!",
                     blkDevInfo.mBlkdev, blkdev_info_path)
        return 1

    blkdev_info_path = "/sys/block/%s/device/rev" % (blkDevInfo.mBlkdev)
    blkDevInfo.mRev, read_failed = read_text_value(blkdev_info_path)
    if read_failed:
        logging.warn("blkdev '%s' vanished [%s] !?!",
                     blkDevInfo.mBlkdev, blkdev_info_path)
        return 1

    return 0
