#!/usr/bin/env python
#
# Copyright (C) 2011-2016 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.

"""Utilities for dealing with the multipath
"""

import logging
import re
import itertools

from itertools import chain
from agent.target.host import (run_cmd)
from agent.storage.st_utils.st_readAndWrite import (read_text_lines,
                                                    read_text_value)

__all__ = ['get_multipath_conf',
           'parse_multipath_conf_file',
           'get_multipath_info']


class multipathBlkDevInfo_OBJ:

    def __init__(self):
        self.mWWID = None
        self.mDevMapNum = None
        self.mSCSIdevIDstr = None
        self.mBlkDev = None
        self.mDevNums = None
        self.mPathDMmapState = None
        self.mCheckerState = None
        self.mSizeInSect = -1

    def __repr__(self):
        return ("mWWID: %s mDevMapNum: %s mSCSIdevIDstr: %s mBlkDev: %s "
                "mDevNums: %s mPathDMmapState: %s mCheckerState: %s "
                "mSizeInSect:%d") % (self.mWWID,
                                     self.mDevMapNum,
                                     self.mSCSIdevIDstr,
                                     self.mBlkDev,
                                     self.mDevNums,
                                     self.mPathDMmapState,
                                     self.mCheckerState,
                                     self.mSizeInSect)


class multipathConfDeviceSpec:

    def __init__(self):
        self.mIsExplicit = False
        self.mIsCalloutGenerated = False
        self.mVendor = None
        self.mVendor_cmpldRegEx = None
        self.mProduct = None
        self.mProduct_cmpldRegEx = None
        self.mCallout = None
        self.mBlacklist = None

    def __repr__(self):
        if self.mIsExplicit == False:
            return "-v: %s p: %s c: %s b: %s" % (
                        self.mVendor,
                        self.mProduct,
                        self.mCallout,
                        self.mBlacklist)
        else:
            return "+v: %s p: %s c: %s b: %s" % (
                        self.mVendor,
                        self.mProduct,
                        self.mCallout,
                        self.mBlacklist)

    def __cmp__(self, other):
        if not isinstance(other, multipathConfDeviceSpec):
            return -1

        cmpVal = cmp(self.mVendor, other.mVendor)
        if cmpVal != 0:
            return cmpVal

        cmpVal = cmp(self.mProduct, other.mProduct)
        if cmpVal != 0:
            return cmpVal

        cmpVal = cmp(self.mCallout, other.mCallout)
        if cmpVal != 0:
            return cmpVal

        cmpVal = cmp(self.mBlacklist, other.mBlacklist)
        if cmpVal != 0:
            return cmpVal

        return 0


class multipathConfBlacklistDevNode:

    def __init__(self, theDevNode):
        self.mDevNode = theDevNode
        self.mDevNode_cmpldRegEx = None

    def __repr__(self):
        return "%s" % (self.mDevNode)


class multipathConfBlacklistDevice:

    def __init__(self):
        self.mVendor = None
        self.mVendor_cmpldRegEx = None
        self.mProduct = None
        self.mProduct_cmpldRegEx = None

    def __repr__(self):
        return "v: %s p: %s" % (self.mVendor, self.mProduct)

    def __cmp__(self, other):
        if not isinstance(other, multipathConfBlacklistDevice):
            return -1
        if self.mVendor < other.mVendor:
            return -1
        if self.mVendor > other.mVendor:
            return  1
        if self.mProduct < other.mProduct:
            return -1
        if self.mProduct > other.mProduct:
            return  1
        return 0


class multipathConfBlacklistWWID:

    def __init__(self, theWWID):
        self.mWWID = theWWID
        self.mWWID_cmpldRegEx = None

    def __repr__(self):
        return "%s" % (self.mWWID)


class multipathConfBlacklistInfo:

    def __init__(self):
        self.mDevNode_LIST = []
        self.mDevice_LIST = []
        self.mWWID_LIST = []

    def __repr__(self):
        return "devNode_L: %s device_L: %s wwid_L: %s" % (
               self.mDevNode_LIST, self.mDevice_LIST, self.mWWID_LIST)


global_multipathConf_default_callout = ('scsi_id --whitelisted '
                                    '--replace-whitespace --device=/dev/%n')


def get_block_device_size(name):
    """
    This routine returns the size of the block device in sector size.
    """
    if not name:
        return -1
    sysfs_attr = "/sys/block/%s/size" % (name)
    str_size, read_failed = read_text_value(sysfs_attr)
    if read_failed:
        return -1
    return int(str_size)


def get_multipath_info():
    multipathInfo_DICT = {}
    # 3600805f3001321c02f8962adb414003c dm-11 COMPAQ,MSA1000 VOLUME
    # Alias:
    # yellow (360a98000433468714a6f7757464a4f6d) dm-11 NETAPP,LUN
    mp_wwid_line_cmpldRegEx = re.compile('^(.*)\sdm-(\d+)\s+')
    mp_alias_wwid_line_cmpldRegEx = re.compile('^(\w+)\s*\((\w+)\)\sdm-(\d+)')
    mp_path_line_cmpldRegEx_1 = re.compile('^\`-\s+(\d+:\d+:\d+:\d+)\s+(\S*)'
                                           '\s+(\d+:\d+)\s+(\w+)\s+(\w+)'
                                           '\s+(\w+)$')
    mp_path_line_cmpldRegEx_2 = re.compile('^\|-\s+(\d+:\d+:\d+:\d+)'
                                           '\s+(\S*)\s+(\d+:\d+)\s+(\w+)'
                                           '\s+(\w+)\s+(\w+)$')
    mp_path_line_cmpldRegEx_3 = re.compile('\| `-\s+(\d+:\d+:\d+:\d+)\s+'
                                           '(\S*)\s+(\d+:\d+)\s+(\w+)\s+'
                                           '(\w+)\s+(\w+)$')
    mp_path_line_cmpldRegEx_4 = re.compile('\| \|-\s+(\d+:\d+:\d+:\d+)'
                                           '\s+(\S*)\s+(\d+:\d+)\s+(\w+)'
                                           '\s+(\w+)\s+(\w+)$')

    stdout_str = run_cmd(["multipathd", "list", "topology"])
    stdout_LIST = stdout_str.splitlines(True)
    multipath_lines_LIST = map(str.strip, stdout_LIST)
    del(stdout_LIST)
    try:
        cur_wwid = ''
        cur_dev_map_num = -1
        multipath_line = ''
        for multipath_line in multipath_lines_LIST:
            if len(multipath_line) < 1:
                continue
            if multipath_line.startswith("size"):
                continue
            if multipath_line.startswith("create:"):
                multipath_line = multipath_line.lstrip("create:").strip()

            # Old format had '\' in front of block devices. New has the `
            if multipath_line[0] != '|' and multipath_line[0] != '`':
                mp_wwid_line_match = \
                    mp_alias_wwid_line_cmpldRegEx.match(multipath_line)
                if mp_wwid_line_match:
                    # an alias
                    # store the wwid and dev_num; alias name is not saved.
                    cur_wwid = mp_wwid_line_match.group(2)
                    cur_dev_map_num = int(mp_wwid_line_match.group(3))
                else:  # recompile for non-alias format
                    mp_wwid_line_match = \
                        mp_wwid_line_cmpldRegEx.match(multipath_line)
                    if not mp_wwid_line_match:
                        continue
                    cur_wwid = mp_wwid_line_match.group(1)
                    cur_dev_map_num = int(mp_wwid_line_match.group(2))
                continue
            mp_path_line_match = \
                mp_path_line_cmpldRegEx_1.match(multipath_line)
            if not mp_path_line_match:
                mp_path_line_match = \
                    mp_path_line_cmpldRegEx_2.match(multipath_line)
                if not mp_path_line_match:
                    mp_path_line_match = \
                        mp_path_line_cmpldRegEx_3.match(multipath_line)
                    if not mp_path_line_match:
                        mp_path_line_match = \
                            mp_path_line_cmpldRegEx_4.match(multipath_line)
                        if not mp_path_line_match:
                            continue

            mpBlkDevInfo = multipathBlkDevInfo_OBJ()
            mpBlkDevInfo.mWWID = cur_wwid
            mpBlkDevInfo.mDevMapNum = cur_dev_map_num
            mpBlkDevInfo.mSCSIdevIDstr = mp_path_line_match.group(1)
            mpBlkDevInfo.mBlkDev = mp_path_line_match.group(2)
            mpBlkDevInfo.mDevNums = mp_path_line_match.group(3)
            mpBlkDevInfo.mPathDMmapState = mp_path_line_match.group(4)
            mpBlkDevInfo.mCheckerState = mp_path_line_match.group(5)
            mpBlkDevInfo.mSizeInSect = \
                get_block_device_size(mpBlkDevInfo.mBlkDev)

            blkDev_DICT = multipathInfo_DICT.get(cur_wwid, None)
            if not blkDev_DICT:
                multipathInfo_DICT[cur_wwid] = {mpBlkDevInfo.mBlkDev:
                                                mpBlkDevInfo}
            else:
                blkDev_DICT[mpBlkDevInfo.mBlkDev] = mpBlkDevInfo

        del(multipath_lines_LIST)
    except StandardError, Err:
        raise Exception("multipathd list topology format %s '%s' %s" % \
                        (cur_wwid, multipath_line, Err))

    return multipathInfo_DICT


def parse_multipath_conf_defaults(mpc_token_itr, mpc_token_cntr):
    mpConf_default_callout = global_multipathConf_default_callout
    while True:
        try:
            mpc_key_token = mpc_token_itr.next()
            mpc_token_cntr.next()
            if  '}' == mpc_key_token:
                return mpConf_default_callout
            mpc_val_token = mpc_token_itr.next()
            mpc_token_cntr.next()
        except StopIteration, Err:
            raise Exception(
                "multipathCnfLines unexpected early end [%s] %s" % \
                (mpc_token_cntr, Err))

        if 'getuid_callout' == mpc_key_token:
            mpConf_default_callout = mpc_val_token


def parse_multipath_conf_blacklist_device(mpc_token_itr, mpc_token_cntr):
    mpConf_blkLstDev = multipathConfBlacklistDevice()
    while True:
        try:
            mpc_key_token = mpc_token_itr.next()
            mpc_token_cntr.next()
            if  '}' == mpc_key_token:
                break
            mpc_val_token = mpc_token_itr.next()
            mpc_token_cntr.next()
        except StopIteration, Err:
            raise Exception(
                "multipathCnfLines unexpected early end [%s] %s" % \
                (mpc_token_cntr, Err))

        if  'vendor' == mpc_key_token:
            mpConf_blkLstDev.mVendor = mpc_val_token
        if  'product' == mpc_key_token:
            mpConf_blkLstDev.mProduct = mpc_val_token
    if not mpConf_blkLstDev.mVendor:
        return None
    if not mpConf_blkLstDev.mProduct:
        mpConf_blkLstDev.mProduct = '*'
    return mpConf_blkLstDev


def parse_multipath_conf_blacklist(mpc_token_itr, mpc_token_cntr):
    mpConf_blacklistInfo = multipathConfBlacklistInfo()
    while True:
        try:
            mpc_key_token = mpc_token_itr.next()
            mpc_token_cntr.next()
            if  '}' == mpc_key_token:
                return mpConf_blacklistInfo
            mpc_val_token = mpc_token_itr.next()
            mpc_token_cntr.next()
        except StopIteration, Err:
            raise Exception(
                "multipathCnfLines unexpected early end [%s] %s" % \
                (mpc_token_cntr, Err))

        if 'devnode' == mpc_key_token:
            if mpc_val_token[0] == '*':
                mpc_val_token = '.' + mpc_val_token
            mpConf_blacklistInfo.mDevNode_LIST.append(
                multipathConfBlacklistDevNode(mpc_val_token))
            continue
        if 'device' == mpc_key_token:
            mpConf_blkLstDev = parse_multipath_conf_blacklist_device(
                                                      mpc_token_itr,
                                                      mpc_token_cntr)
            if mpConf_blkLstDev:
                if not mpConf_blkLstDev in mpConf_blacklistInfo.mDevice_LIST:
                    mpConf_blacklistInfo.mDevice_LIST.append(mpConf_blkLstDev)
            continue
        if 'wwid' == mpc_key_token:
            mpConf_blacklistInfo.mWWID_LIST.append(
                                    multipathConfBlacklistWWID(mpc_val_token))
            continue
        if '%include' == mpc_key_token:
            if mpc_val_token:
                mpc_include_LIST, read_failed = read_text_lines(mpc_val_token)
                if read_failed:
                    continue
                mpc_include_itr = iter(mpc_include_LIST)
                mpc_token_itr = chain(mpc_token_itr, mpc_include_itr)
                continue
        raise Exception(
            "multipathCnfLines unexpected format '%s' => '%s' [%s]" % \
            (mpc_key_token, mpc_val_token, mpc_token_cntr))


def parse_multipath_conf_blacklist_exceptions(mpc_token_itr, mpc_token_cntr):
    return parse_multipath_conf_blacklist(mpc_token_itr, mpc_token_cntr)


def parse_multipath_conf_device(mpc_token_itr, mpc_token_cntr):
    mpConf_devSpec = multipathConfDeviceSpec()
    while True:
        try:
            mpc_key_token = mpc_token_itr.next()
            mpc_token_cntr.next()
            if  '}' == mpc_key_token:
                break
            mpc_val_token = mpc_token_itr.next()
            mpc_token_cntr.next()
        except StopIteration:
            return None

        if  'vendor' == mpc_key_token:
            mpConf_devSpec.mVendor = mpc_val_token
        if  'product' == mpc_key_token:
            mpConf_devSpec.mProduct = mpc_val_token
        if  'getuid_callout' == mpc_key_token:
            mpConf_devSpec.mCallout = mpc_val_token
            if mpc_val_token.find(' --force_uniq ') >= 0:
                mpConf_devSpec.mIsCalloutGenerated = True
        if  'product_blacklist' == mpc_key_token:
            mpConf_devSpec.mBlacklist = mpc_val_token

    if not mpConf_devSpec.mVendor:
        return None
    if not mpConf_devSpec.mProduct:
        mpConf_devSpec.mProduct = '*'

    return mpConf_devSpec


def parse_multipath_conf_devices(mpc_token_itr, mpc_token_cntr):
    mpConf_devSpec_LIST = []
    while True:
        try:
            mpc_token = mpc_token_itr.next()
            mpc_token_cntr.next()
            if  '}' == mpc_token:
                return mpConf_devSpec_LIST
            if 'device' != mpc_token:
                raise Exception(
                    "multipath.conf unexpected format '%s' [%s]" % \
                    (mpc_token, mpc_token_cntr))
            mpc_token = mpc_token_itr.next()
            mpc_token_cntr.next()
            if '{' != mpc_token:
                raise Exception(
                    "multipath.conf unexpected format '%s' [%s]" % \
                    (mpc_token, mpc_token_cntr))
        except StopIteration, Err:
            raise Exception("multipath.conf unexpected early end [%s] %s" % \
                            (mpc_token_cntr, Err))

        mpConf_devSpec = parse_multipath_conf_device(mpc_token_itr,
                                                     mpc_token_cntr)
        if not mpConf_devSpec:
            continue
        mpConf_devSpec_LIST.append(mpConf_devSpec)
    return mpConf_devSpec_LIST


def parse_multipath_conf_multipath(mpc_token_itr, mpc_token_cntr):
    """
    This routine will skip over multipath tokens. It does not
    do anything with them.

    Note: alias disks are missing from discover_physical_luns output
          Multipath -ll displays alias disks with a different syntax.
          Note the alias name and the wwid is enclosed in ().

          # multipath -ll
            yellow (360a98000433468714a6f7757464a4f6d) dm-1 NETAPP,LUN
            360a98000433468714a6f775746486849 dm-0 NETAPP,LUN
    """
    while True:
        try:
            mpc_key_token = mpc_token_itr.next()
            mpc_token_cntr.next()
            if  '}' == mpc_key_token:
                return
            mpc_token_itr.next()
            mpc_token_cntr.next()
        except StopIteration, Err:
            raise Exception("multipath.conf unexpected early end [%s] %s" % \
                            (mpc_token_cntr, Err))


def parse_multipath_conf_multipaths(mpc_token_itr, mpc_token_cntr):
    while True:
        try:
            mpc_key_token = mpc_token_itr.next()
            mpc_token_cntr.next()
            if  '}' == mpc_key_token:
                return
            if 'multipath' != mpc_key_token:
                raise Exception(
                    "multipath.conf contains unexpected format '%s' [%s]" % \
                    (mpc_key_token, mpc_token_cntr))
            mpc_key_token = mpc_token_itr.next()
            mpc_token_cntr.next()
            if '{' != mpc_key_token:
                raise Exception(
                    "multipath.conf contains unexpected format '%s' [%s]" % \
                    (mpc_key_token, mpc_token_cntr))
        except StopIteration, Err:
            raise Exception("multipath.conf unexpected early end [%s] %s" % \
                            (mpc_token_cntr, Err))

#        Do not allow multipath alias for now
#        parse_multipath_conf_multipath(mpc_token_itr, mpc_token_cntr)

        raise Exception("found unsupported multipath '%s' %s" % \
                        (mpc_key_token, mpc_token_cntr))
    return


def parse_multipath_conf_lines_list(mpConfLines_LIST):
    # Tokenize based on spaces and "quoted" strings
    #     AND if not quoted all tokens after the first in a line
    #     are considered part of the same token
    mpc_tokens = []
    cur_token = ''
    is_quoted = False
    for line in mpConfLines_LIST:
        conf_line = (line.split('#', 1))[0].lstrip().rstrip()
        if not conf_line:
            continue
        is_first_token_in_line = True
        was_quoted = False
        for mpc_char in conf_line:
            if mpc_char == '"':
                if is_quoted:
                    is_quoted = False
                    was_quoted = True
                else:
                    is_quoted = True
                continue
            if not mpc_char.isspace() or is_quoted:
                cur_token += mpc_char
                continue
            if len(cur_token) < 1:
                continue
            if not is_first_token_in_line and not was_quoted:
                cur_token += mpc_char
                continue
            mpc_tokens.append(cur_token)
            cur_token = ''
            is_first_token_in_line = False
        if len(cur_token) > 0:
            mpc_tokens.append(cur_token)
            cur_token = ''

    mpConf_default_callout = global_multipathConf_default_callout
    mpConf_blacklistInfo = multipathConfBlacklistInfo()
    mpConf_blacklistExceptionsInfo = multipathConfBlacklistInfo()
    mpConf_devSpec_LIST = []

    mpc_token_itr = iter(mpc_tokens)
    mpc_token_cntr = itertools.count(0)
    while True:
        try:
            mpc_tok_start = mpc_token_itr.next()
            mpc_token_cntr.next()
            if not re.match('^\w+$', mpc_tok_start):
                raise Exception("multipathCnfLines unexpected format [%s]" % \
                                (mpc_token_cntr))
        except StopIteration:
            break

        try:
            mpc_token = mpc_token_itr.next()
            mpc_token_cntr.next()
            if '{' != mpc_token:
                raise Exception(
                    "multipathCnfLines unexpected format [%s] %s" % \
                     (mpc_token_cntr, mpc_token))
        except StopIteration, Err:
            raise Exception(
                        "multipathCnfLines unexpected early end [%s] %s" % \
                        (mpc_token_cntr, Err))

        if 'defaults' == mpc_tok_start:
            mpConf_default_callout = \
                parse_multipath_conf_defaults(mpc_token_itr, mpc_token_cntr)
            continue
        if 'blacklist' == mpc_tok_start:
            mpConf_blacklistInfo = parse_multipath_conf_blacklist(
                                        mpc_token_itr, mpc_token_cntr)
            continue
        if 'blacklist_exceptions' == mpc_tok_start:
            mpConf_blacklistExceptionsInfo = \
                parse_multipath_conf_blacklist_exceptions(mpc_token_itr,
                                                          mpc_token_cntr)
            continue
        if 'devices' == mpc_tok_start:
            mpConf_devSpec_LIST = parse_multipath_conf_devices(mpc_token_itr,
                                                               mpc_token_cntr)
            continue
        if 'multipaths' == mpc_tok_start:
            parse_multipath_conf_multipaths(mpc_token_itr, mpc_token_cntr)
            continue
        raise Exception("multipathCnfLines unexpected section '%s' [%s]" % \
                        (mpc_tok_start, mpc_token_cntr))

    for mpConf_blkLstDevNode in mpConf_blacklistInfo.mDevNode_LIST:
        mpConf_blkLstDevNode.mDevNode_cmpldRegEx = \
            re.compile(mpConf_blkLstDevNode.mDevNode)
    for mpConf_blkLstDev in mpConf_blacklistInfo.mDevice_LIST:
        try:
            mpConf_blkLstDev.mVendor_cmpldRegEx = \
                re.compile(mpConf_blkLstDev.mVendor)
        except re.error:
            mpConf_blkLstDev.mVendor_cmpldRegEx = re.compile('.*?')
        try:
            mpConf_blkLstDev.mProduct_cmpldRegEx = \
                re.compile(mpConf_blkLstDev.mProduct)
        except StandardError:
            mpConf_blkLstDev.mProduct_cmpldRegEx = re.compile('.*?')
    for mpConf_blkLstWWID in mpConf_blacklistInfo.mWWID_LIST:
        try:
            mpConf_blkLstWWID.mWWID_cmpldRegEx = \
                re.compile(mpConf_blkLstWWID.mWWID)
        except StandardError:
            pass
    for mpConf_blkLstDevNode in mpConf_blacklistExceptionsInfo.mDevNode_LIST:
        try:
            mpConf_blkLstDevNode.mDevNode_cmpldRegEx = \
                re.compile(mpConf_blkLstDevNode.mDevNode)
        except StandardError:
            pass
    for mpConf_blkLstDev in mpConf_blacklistExceptionsInfo.mDevice_LIST:
        try:
            mpConf_blkLstDev.mVendor_cmpldRegEx = \
                re.compile(mpConf_blkLstDev.mVendor)
            mpConf_blkLstDev.mProduct_cmpldRegEx = \
                re.compile(mpConf_blkLstDev.mProduct)
        except StandardError:
            pass
    for mpConf_blkLstWWID in mpConf_blacklistExceptionsInfo.mWWID_LIST:
        try:
            mpConf_blkLstWWID.mWWID_cmpldRegEx = \
                re.compile(mpConf_blkLstWWID.mWWID)
        except StandardError:
            pass
    for mpConf_devSpec in mpConf_devSpec_LIST:
        try:
            mpConf_devSpec.mVendor_cmpldRegEx = \
                re.compile(mpConf_devSpec.mVendor)
            mpConf_devSpec.mProduct_cmpldRegEx = \
                re.compile(mpConf_devSpec.mProduct)
        except StandardError:
            pass
    return (mpConf_default_callout, mpConf_blacklistInfo,
            mpConf_blacklistExceptionsInfo, mpConf_devSpec_LIST)


def parse_multipath_conf_file(multipathConfFilePath):
    if multipathConfFilePath == None:
        multipathConfFilePath = '/etc/multipath.conf'
    mpc_contents_LIST, read_failed = read_text_lines(multipathConfFilePath)
    if read_failed:
        logging.warning(
            "Warning:  %s could not be read !?!", multipathConfFilePath)
        return None
    (dummy_mpConf_default_callout, dummy_mpConf_blacklistInfo,
     dummy_mpConf_blacklistExceptionsInfo,
     mpConf_devSpec_LIST) = parse_multipath_conf_lines_list(mpc_contents_LIST)

    return mpConf_devSpec_LIST


def get_multipath_conf():
    cmd_LIST = ['multipathd', '-kshow config']
    # pass (0, 1) to run_cmd to allow for EMC powermnt
    # which does not start multipath
    stdout_str = run_cmd(cmd_LIST, (0, 1))
    multipath_lines_LIST = stdout_str.splitlines(True)
    mpc_contents_LIST = map(str.strip, multipath_lines_LIST)
    del(multipath_lines_LIST)
    (mpConf_default_callout, mpConf_blacklistInfo,
     mpConf_blacklistExceptionsInfo, mpConf_devSpec_LIST) = \
         parse_multipath_conf_lines_list(mpc_contents_LIST)

    etc_devSpec_LIST = parse_multipath_conf_file(None)
    for devSpec in mpConf_devSpec_LIST:
        if devSpec in etc_devSpec_LIST:
            devSpec.mIsExplicit = True

    return (mpConf_default_callout, mpConf_blacklistInfo,
            mpConf_blacklistExceptionsInfo, mpConf_devSpec_LIST)
