#!/usr/bin/env python
#
# Copyright (C) 2011-2016 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.

"""Utilities for dealing with the open-iscsi program
"""

import logging
import errno
import time
import os
import re

from agent.storage.st_utils.st_readAndWrite import read_text_value
from agent.storage.st_tech.blockDeviceOperations import (BlkDev_Operations,
                                                  std_disconnect_block_device,
                                                  std_revalidate_block_device)
from agent.storage.st_utils.st_process import (simple_process_checker,
                                               terminate_process,
                                               exe_async_with_checker,
                                               pidof)

__all__ = ['get_openiscsi_session_info_dict',
           'openISCSI_BlkDev_Operations',
           'iscsiadm_exe_sync']


class openISCSIstate_OBJ:

    INACTIVE_SESSION_NUM = -1

    def __init__(self):
        self.mSessionNum = -1
        self.mSCSIHostNum = -1
        self.mTargetPort = -1
        self.mTpgt = -1
        self.mTargetIP = None
        self.mIQN = None
        self.mBlkDev_LIST = []
        self.mSGDev_LIST = []

    def __repr__(self):
        return ("mSessionNum: %d mSCSIHostNum: %d mTargetPort: %d "
            "mTpgt: %d mTargetIP: %s mIQN: %s mBlkDevs: %s mSGDevs: %s") % (
            self.mSessionNum,
            self.mSCSIHostNum,
            self.mTargetPort,
            self.mTpgt,
            self.mTargetIP,
            self.mIQN,
            self.mBlkDev_LIST,
            self.mSGDev_LIST)

    def __str__(self):
        return ("%s:%d,%d %s" %
                (self.mTargetIP, self.mTargetPort, self.mTpgt, self.mIQN))

    def __cmp__(self, other):
        if self.mTargetIP < other.mTargetIP:
            return -1
        if self.mTargetPort < other.mTargetPort:
            return -1
        if self.mTpgt < other.mTpgt:
            return -1
        if self.mIQN < other.mIQN:
            return -1
        if self.mTargetIP > other.mTargetIP:
            return 1
        if self.mTargetPort > other.mTargetPort:
            return 1
        if self.mTpgt > other.mTpgt:
            return 1
        if (self.mIQN > other.mIQN):
            return 1
        return 0


class openISCSI_BlkDev_Operations(BlkDev_Operations):
    """openISCSI class for BlkDev_Operations
    """

    def __init__(self, blkDevInfo, openISCSIstate):
        if blkDevInfo.mDriver != 'iscsi_tcp':
            raise ValueError(
                   "given blkDevInfo %s is not openISCSI [%s] !?!" %
                    (blkDevInfo.mBlkDev, blkDevInfo.mDriver))
        self.mBlkDevInfo = blkDevInfo
        self.mOpenISCSIstate = openISCSIstate

    def __repr__(self):
        return "openISCSI_BlkDev_Operations[%s]" % (self.mBlkDevInfo.mBlkDev)

    def disconnect_block_device(self):
        std_disconnect_block_device(self.mBlkDevInfo)
        return 0

    def revalidate_block_device(self):
        """
        Updates a standard block device to the latests sysfs values
        """
        retVal = std_revalidate_block_device(self.mBlkDevInfo)
        if retVal:
            return retVal
        return 0


def iscsiadm_exe_sync(args, inputData_LIST=None):
    """
    This routine manages a safe iscsiadm execution
    """
    checkerData = (600, "iscsiadm -m node")
    checkerFunc = simple_process_checker
    itr = 0
    while True:
        (stdout_LIST, stderr_LIST,
         exitStatus, deltaTime) = exe_async_with_checker(args, checkerFunc,
                                                         checkerData,
                                                         inputData_LIST)
        if exitStatus != errno.ETIMEDOUT:
            break
        itr += 1
        if itr > 3:
            logging.warn("iscsiadm_exe_sync() failed to restart iSCSI daemon")
            break
        # If there are other iscsiadm commands wedged they will retry
        # their IPC calls with the new iSCSI daemon. This might led back
        # to the wedge state so we kill those before we restart iSCSI daemon.
        pid_LIST = pidof("iscsiadm")
        for pid in pid_LIST:
            terminate_process(pid, True)
        pid_LIST = pidof("iscsid")
        for pid in pid_LIST:
            terminate_process(pid, True)
        logging.warn("iscsiadm_exe_sync() restarting iSCSI daemon")
        os.system("iscsid &")    # system() launches the exe in a shell
        time.sleep(5)            # Give iscsid a chance to start running

    return (stdout_LIST, stderr_LIST, exitStatus, deltaTime)


def get_openiscsi_session_info_dict(dummy_refresh):
    """
    This routine extracts the Open-iSCSI session data. The returned
    values are in the dictionary wherein the key is the session number
    and the value the session data (as the openISCSIstate_OBJ)
    """
    openiscsi_session_info_DICT = {}
    # Directory format of ../iscsi_session is as follow:
    #    session<X>:
    #        -targetname, recovery_tmo, username, etc..
    #        + device
    #            -uevent
    #            +connection<Y>:0
    #                + iscsi_connection:connection<Y>:0 (a sym-link)
    #                    + device (symlink to connection<Y>:0
    #                    - address, port, persistentaddress, etc
    #            +power
    #            +target<Z>:0:0
    #                + Z:A:B:C (the SCSI Host:Bus:Channel:LUN ID)
    #                   + block:<sdM> (symlink to /sys/block/sdM) - optional
    #                   + scsi_generic:<sgN> (symlink to /sys/scsi_generic/sgN)
    #                   + and more dirs...
    #                   - delete, type, vendor, model, rescan, etc..
    #
    session_root = '/sys/class/iscsi_session/'
    if not os.path.exists(session_root):
        return {}
    for session in os.listdir(session_root):
        openISCSIstate = openISCSIstate_OBJ()
        openISCSIstate.mSessionNum = int(re.sub('^.*session', '', session))
        tpgt, read_failed = read_text_value(session_root + session + "/tpgt")
        if not read_failed:
            openISCSIstate.mTpgt = int(tpgt)
        openISCSIstate.mIQN, read_failed = read_text_value(session_root +
                                                           session +
                                                           "/targetname")
        device_root = "/sys/class/iscsi_session/%s/device" % (session)
        # ../../../devices/platform/host5/session1
        link = os.readlink(device_root)
        host_match = re.match("^\S*host(\d+)\S*$", link)
        if host_match:
            openISCSIstate.mSCSIHostNum = int(host_match.group(1))
        for dummy_root, dirs, dummy_files in os.walk(device_root):
            for directory in dirs:
                # Extract the block device and SCSI Generic name
                if 'target' in directory:
                    # Silly sym-links means that we have to re-do a
                    # walk on the .../target<Z>:0:0 directory:
                    target_root = (session_root +
                                   "%s/device/%s/" % (session, directory))
                    for dummy_root, dirs, dummy_files in os.walk(target_root):
                        # 12:0:0:1, 12:0:0:3, etc..
                        for directory in dirs:
                            if 'scsi_generic:' in directory:
                                sgname = re.sub('scsi_generic:', '', directory)
                                openISCSIstate.mSGDev_LIST.append(sgname)
                            if 'block:' in directory:
                                blkName = re.sub('block:', '', directory)
                                openISCSIstate.mBlkDev_LIST.append(blkName)
                # Extract the address and port
                if 'connection' in directory:
                    iscsi_connection_root = (session_root +
                                    "%s/device/%s/" % (session, directory))
                    for dummy_root, dirs, dummy_files in os.walk(
                                                        iscsi_connection_root):
                        for dir2 in dirs:
                            if 'iscsi_connection' in dir2:
                                port, read_failed = read_text_value(
                                     iscsi_connection_root + dir2 +
                                     "/persistent_port")
                                if not read_failed:
                                    openISCSIstate.mTargetPort = int(port)
                                else:
                                    port, read_failed = read_text_value(
                                             iscsi_connection_root + dir2 +
                                             "/" + directory +
                                             "/persistent_port")
                                    if not read_failed:
                                        openISCSIstate.mTargetPort = int(port)
                                    else:
                                        openISCSIstate.mTargetPort = 3260
                                (openISCSIstate.mTargetIP,
                                 read_failed) = read_text_value(
                                                iscsi_connection_root + dir2 +
                                                "/persistent_address")
                                if read_failed:
                                    (openISCSIstate.mTargetIP,
                                     read_failed) = read_text_value(
                                                        iscsi_connection_root +
                                                        dir2 +
                                                        "/" + directory +
                                                        "/persistent_address")
                                if read_failed:
                                    (openISCSIstate.mTargetIP,
                                     read_failed) = read_text_value(
                                                        iscsi_connection_root +
                                                        dir2 + "/address")
                                if read_failed:
                                    (openISCSIstate.mTargetIP,
                                     read_failed) = read_text_value(
                                                iscsi_connection_root + dir2 +
                                                "/" + directory + "/address")

        openiscsi_session_info_DICT[openISCSIstate.mSessionNum] = openISCSIstate

    return openiscsi_session_info_DICT
