#!/usr/bin/env python
#
# Copyright (C) 2011-2016 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.

"""Utilities for dealing with processes
"""

import os
import re  # simpleXML.py
import time
import errno
import fcntl
import select
import signal
import subprocess
import shlex
import logging

from agent.storage.st_utils.st_readAndWrite import read_text_lines

__all__ = ['Process_OBJ',
           'simple_process_checker',
           'exe_async_with_checker',
           'pidof',
           'terminate_process']


class Process_OBJ:
    """Class for creating and managing a child process

    Arguments:  runWithinBashShell     Run the cmd within a bash shell
                args                   Either a Tuple/List or str
                relativeTimeoutInSecs  When to give up and kill the
                                       child process
                inputData_LIST         List of input data

    NOTE:    I truly hope that python implements this more directly
             in the future !!!
    """
    def __init__(self, runWithinBashShell,
                 args, relativeTimeoutInSecs=0, inputData_LIST=None):
        if type(runWithinBashShell) != bool:
            raise ValueError("bad runWithinBashShell type")
        if type(relativeTimeoutInSecs) != int:
            raise ValueError("bad relativeTimeoutInSecs type")

        if type(args) == list:
            args_LIST = args[:]
        elif type(args) == tuple:
            args_LIST = list(args[:])
        elif type(args) == str:
            args_LIST = shlex.split(args)
        else:
            raise ValueError("bad args type")

        if runWithinBashShell:
            shellExe = '/bin/bash'
        else:
            shellExe = None

        try:
            # Start the child proc ASAP
            self.mChildProc = subprocess.Popen(
                # args :: the command and arguments to execute
                args_LIST,
                # bufsize :: 0 == unbuffered, 1 == line buffered, <0 sys
                # default, >0 fully buffered
                0,
                # executable :: only useful to specify the shell
                shellExe,
                # stdin :: fd to connect to or PIPE to create a pipe
                subprocess.PIPE,
                # stdout :: fd to connect to or PIPE to create a pipe
                subprocess.PIPE,
                # stderr :: fd to connect to or PIPE to create a pipe
                subprocess.PIPE,
                # preexec_fn :: fcn to run in child before exec
                None,
                # close_fds :: True ==> close all FDs except 0, 1 and 2
                True,
                # shell :: True ==> execute the cmomand through the
                # shell specified by executable
                runWithinBashShell,
                # cwd :: dir to begin command execution in
                None,
                # env :: specified env for the command
                None,
                # universal_newlines :: True ==> open stdout and
                # stdout as text files instead of pipes
                False,
                # startupinfo  :: Windows only CreateProcess flag
                None,
                # creationflags :: Windows only CreateProcess flag
                0)
        except:
            logging.error(
                    "Child Process creation failed : %s :\n", args_LIST)
            raise

        self.mStartTimeInSecs = time.time()
        self.mArgsStr = str(args_LIST)
        self.mRelTimeoutInSecs = relativeTimeoutInSecs
        self.mDeltaRunTimeInSecs = None
        self.mExitStatus = None
        self.mDidTimeout = False
        self.mPID = self.mChildProc.pid
        self.mChildIO_POLL = select.poll()
        self.mNumInChildIOpoll = 0
        self.mInputDataItr = None
        self.mCurInputData = None
        self.mPrev_childStdoutData = ''
        self.mPrev_childStderrData = ''
        self.mChildStdoutLines_LIST = []
        self.mChildStderrLines_LIST = []

        if inputData_LIST:
            self.mInputDataItr = iter(inputData_LIST)
            self.mCurInputData = self.mInputDataItr.next()
#            _INTERNAL__setblocking(self.mChildProc.stdin, False)
            self.mChildIO_POLL.register(self.mChildProc.stdin, select.POLLOUT)
            self.mNumInChildIOpoll = self.mNumInChildIOpoll + 1

        _INTERNAL__setblocking(self.mChildProc.stdout, False)
        self.mChildIO_POLL.register(self.mChildProc.stdout, select.POLLIN)
        self.mNumInChildIOpoll = self.mNumInChildIOpoll + 1

        _INTERNAL__setblocking(self.mChildProc.stderr, False)
        self.mChildIO_POLL.register(self.mChildProc.stderr, select.POLLIN)
        self.mNumInChildIOpoll = self.mNumInChildIOpoll + 1

    def __repr__(self):
        return str(self.mPID)

    def terminate_child(self, maximumKill=True):
        """
        This method will terminate the child if it is still running
        """
        if self.mExitStatus != None:
            return False
        nowTimeInSecs = time.time()
        try:
            os.kill(self.mChildProc.pid, signal.SIGHUP)
            for dummy_itr in range(1, 5):
                self.mExitStatus = self.mChildProc.poll()
                if self.mExitStatus != None:
                    break
                time.sleep(0.1)
            if self.mExitStatus != None and maximumKill:
                os.kill(self.mChildProc.pid, signal.SIGKILL)
                for dummy_itr in range(1, 10):
                    self.mExitStatus = self.mChildProc.poll()
                    if self.mExitStatus != None:
                        break
                    time.sleep(0.1)
        except StandardError:
            pass

        if self.mPrev_childStdoutData:
            self.mChildStdoutLines_LIST.append(self.mPrev_childStdoutData)
        if self.mPrev_childStderrData:
            self.mChildStderrLines_LIST.append(self.mPrev_childStderrData)

        if self.mExitStatus != None:
            if nowTimeInSecs >= self.mStartTimeInSecs:
                self.mDeltaRunTimeInSecs = (nowTimeInSecs -
                                            self.mStartTimeInSecs)
            else:
                self.mDeltaRunTimeInSecs = (self.mStartTimeInSecs -
                                            nowTimeInSecs)
        else:
            self.mDeltaRunTimeInSecs = -1
            logging.warn("Process_OBJ(%s) child would not die [%s]",
                         self.mArgsStr, maximumKill)

        del(self.mChildProc)
        del(self.mStartTimeInSecs)
        del(self.mRelTimeoutInSecs)
        del(self.mChildIO_POLL)
        del(self.mNumInChildIOpoll)
        del(self.mInputDataItr)
        del(self.mCurInputData)
        del(self.mPrev_childStdoutData)
        del(self.mPrev_childStderrData)

        if self.mExitStatus != None:
            return False

        return True        # Child failed to exit

    def monitor_child_for_moment(self):
        """
        This method will do some child IO processing before returning

        @return     isDone, wasChildIOprogressMade
        """
        if self.mExitStatus != None:
            return True, False
        wasChildIOprogressMade = False
        progressCnt = 0
        while self.mNumInChildIOpoll > 0:
            if self._INTERNAL__monitor_child_io_poll(0.5) and progressCnt < 2:
                # Keep making progress up to 3 times in a row
                progressCnt = progressCnt + 1
                continue
            if progressCnt > 0:
                wasChildIOprogressMade = True
            timeLeftInSecs = self._INTERNAL__monitor_child_consider_timeout()
            if timeLeftInSecs < 0:
                return True, wasChildIOprogressMade
            return False, wasChildIOprogressMade

        if progressCnt > 0:
            wasChildIOprogressMade = True

        time.sleep(0.1)
        self.mExitStatus = self.mChildProc.poll()
        if self.mExitStatus == None:
            time.sleep(0.1)
            self.mExitStatus = self.mChildProc.poll()
            if self.mExitStatus == None:
                if self._INTERNAL__monitor_child_consider_timeout():
                    return True, wasChildIOprogressMade
                return False, wasChildIOprogressMade

        nowTimeInSecs = time.time()
        if nowTimeInSecs >= self.mStartTimeInSecs:
            self.mDeltaRunTimeInSecs = nowTimeInSecs - self.mStartTimeInSecs
        else:
            self.mDeltaRunTimeInSecs = self.mStartTimeInSecs - nowTimeInSecs

        del(self.mChildProc)
        del(self.mStartTimeInSecs)
        del(self.mRelTimeoutInSecs)
        del(self.mNumInChildIOpoll)

        return True, False

    def monitor_child_until_done(self):
        """
        This method will do child IO processing until the child is done

        @return     didTimeout
        """
        if self.mExitStatus != None:
            return False
        while self.mNumInChildIOpoll > 0:
            timeLeftInSecs = self._INTERNAL__monitor_child_consider_timeout()
            if timeLeftInSecs < 0:
                return True
            self._INTERNAL__monitor_child_io_poll(timeLeftInSecs)
        if self.mRelTimeoutInSecs == 0:
            self.mExitStatus = self.mChildProc.wait()
            nowTimeInSecs = time.time()
            if nowTimeInSecs >= self.mStartTimeInSecs:
                self.mDeltaRunTimeInSecs = (nowTimeInSecs -
                                            self.mStartTimeInSecs)
            else:
                self.mDeltaRunTimeInSecs = (self.mStartTimeInSecs -
                                            nowTimeInSecs)
        else:
            while True:
                time.sleep(0.1)
                self.mExitStatus = self.mChildProc.poll()
                if self.mExitStatus != None:
                    break
                time.sleep(0.1)
                self.mExitStatus = self.mChildProc.poll()
                if self.mExitStatus != None:
                    break

                timeLeftInSecs = self._INTERNAL__monitor_child_consider_timeout()
                if timeLeftInSecs < 0:
                    return True
                time.sleep(0.1)
            nowTimeInSecs = time.time()
            if nowTimeInSecs >= self.mStartTimeInSecs:
                self.mDeltaRunTimeInSecs = (nowTimeInSecs -
                                            self.mStartTimeInSecs)
            else:
                self.mDeltaRunTimeInSecs = (self.mStartTimeInSecs -
                                            nowTimeInSecs)
        del(self.mChildProc)
        del(self.mStartTimeInSecs)
        del(self.mRelTimeoutInSecs)
        del(self.mNumInChildIOpoll)

        return False

    def get_child_results(self):
        if self.mExitStatus != None:
            if not self.mDidTimeout:
                return (self.mChildStdoutLines_LIST,
                        self.mChildStderrLines_LIST,
                        self.mExitStatus,
                        self.mDeltaRunTimeInSecs)
            return (self.mChildStdoutLines_LIST,
                    self.mChildStderrLines_LIST,
                    errno.ETIMEDOUT,
                    -1)
        return (self.mChildStdoutLines_LIST[:],
                self.mChildStderrLines_LIST[:],
                0,
                0)

    def _INTERNAL__monitor_child_consider_timeout(self):
        """
        This method checks for timeout and terminates if so
        """
        if self.mRelTimeoutInSecs == 0:
            return 0
        nowTimeInSecs = time.time()
        if nowTimeInSecs >= self.mStartTimeInSecs:
            deltaTimeInSecs = nowTimeInSecs - self.mStartTimeInSecs
        else:
            deltaTimeInSecs = self.mStartTimeInSecs - nowTimeInSecs
        if deltaTimeInSecs < self.mRelTimeoutInSecs:
            return self.mRelTimeoutInSecs - deltaTimeInSecs
        self.terminate_child(True)
        self.mDidTimeout = True

        return -1

    def _INTERNAL__monitor_child_io_poll_ingest_stdout(self, theFD,
                                                       theEventBits):
        """
        This method ingests the stdout data stdout
        """
        if theEventBits & select.POLLIN:
            childStdoutData = os.read(theFD, 262144)  # 256kB max PIPE size
            if not childStdoutData:
                theEventBits = theEventBits | select.POLLHUP
            else:
                newLines_List = (self.mPrev_childStdoutData +
                                 childStdoutData).splitlines(True)
                self.mPrev_childStdoutData = newLines_List.pop()
                self.mChildStdoutLines_LIST.extend(newLines_List)

        if theEventBits & select.POLLERR:
            logging.error("Process_OBJ(%s) stdout poll error", self.mArgsStr)
            theEventBits = theEventBits | select.POLLHUP

        if theEventBits & select.POLLHUP:
            self.mChildIO_POLL.unregister(theFD)
            self.mNumInChildIOpoll = self.mNumInChildIOpoll - 1
            if self.mPrev_childStdoutData:
                self.mChildStdoutLines_LIST.append(self.mPrev_childStdoutData)
            del(self.mPrev_childStdoutData)

    def _INTERNAL__monitor_child_io_poll_ingest_stderr(self, theFD,
                                                       theEventBits):
        """
        This method ingests the stderr data from stderr
        """
        if theEventBits & select.POLLIN:
            childStderrData = os.read(theFD, 262144)  # 256kB max PIPE size
            if not childStderrData:
                theEventBits = theEventBits | select.POLLHUP
            else:
                newLines_List = (self.mPrev_childStderrData +
                                 childStderrData).splitlines(True)
                self.mPrev_childStderrData = newLines_List.pop()
                self.mChildStderrLines_LIST.extend(newLines_List)
        if theEventBits & select.POLLERR:
            theEventBits = theEventBits | select.POLLHUP
        if theEventBits & select.POLLHUP:
            self.mChildIO_POLL.unregister(theFD)
            self.mNumInChildIOpoll = self.mNumInChildIOpoll - 1
            if self.mPrev_childStderrData:
                self.mChildStderrLines_LIST.append(self.mPrev_childStderrData)
            del(self.mPrev_childStderrData)

    def _INTERNAL__monitor_child_io_poll_send_to_stdin(self, theFD,
                                                       theEventBits):
        """
        This method send the stdin data to stdin
        """
        if theEventBits == select.POLLOUT:
            amountWritten = os.write(self.mChildProc.stdin.fileno(),
                                     self.mCurInputData)
            if amountWritten == 0:
                return False
            if amountWritten < len(self.mCurInputData):
                self.mCurInputData = self.mCurInputData[amountWritten:]
            else:
                try:
                    self.mCurInputData = self.mInputDataItr.next()
                except StopIteration:
                    self.mCurInputData = None
                    theEventBits = theEventBits | select.POLLHUP

        if theEventBits & select.POLLERR or theEventBits & select.POLLHUP:
            self.mChildIO_POLL.unregister(theFD)
            self.mNumInChildIOpoll = self.mNumInChildIOpoll - 1
            if self.mCurInputData != None:
                bytesLeftUnWritten = len(self.mCurInputData)
                nunOfItemsUnWritten = 1
                for inputData in self.mInputDataItr:
                    bytesLeftUnWritten = bytesLeftUnWritten + len(inputData)
                    nunOfItemsUnWritten = nunOfItemsUnWritten + 1
            self.mChildProc.stdin.close()
            del(self.mCurInputData)
            del(self.mInputDataItr)

    def _INTERNAL__monitor_child_io_poll(self, pollTimeoutInSecs):
        """
        This method does a select on the Child IO fds and
        processes any results
        """
        pollTimeoutInMilliSecs = pollTimeoutInSecs * 1000
        fdEvt_TPL_LIST = self.mChildIO_POLL.poll(pollTimeoutInMilliSecs)
        if len(fdEvt_TPL_LIST) == 0:
            return False    # no progress made
        for theFD, theEventBits in fdEvt_TPL_LIST:
            if theFD == self.mChildProc.stdout.fileno():
                self._INTERNAL__monitor_child_io_poll_ingest_stdout(theFD,
                                                                theEventBits)
                continue
            if theFD == self.mChildProc.stderr.fileno():
                self._INTERNAL__monitor_child_io_poll_ingest_stderr(theFD,
                                                                theEventBits)
                continue
            if theFD == self.mChildProc.stdin.fileno():
                self._INTERNAL__monitor_child_io_poll_send_to_stdin(theFD,
                                                                theEventBits)
                continue
            logging.critical("Process_OBJ(%s) unknown FD %d [0x%X]",
                             self.mArgsStr, theFD, theEventBits)
        if self.mNumInChildIOpoll != 0:
            return True
        del(self.mChildIO_POLL)

        return True


def _INTERNAL__setblocking(givenFile, newState):
    """
    This routine sets the blocking attribute on the given file

    Example:        setblocking( stdout, False )
    """
    if type(givenFile) == file:
        givenFile_FD = givenFile.fileno()
    else:
        givenFile_FD = int(givenFile)
    givenFile_flag = fcntl.fcntl(givenFile_FD, fcntl.F_GETFL)
    if not newState:
        givenFile_flag = givenFile_flag | os.O_NONBLOCK
    else:
        givenFile_flag = givenFile_flag & ~os.O_NONBLOCK
    fcntl.fcntl(givenFile_FD, fcntl.F_SETFL, givenFile_flag)


def exe_sync(args, relativeTimeoutInSecs=0, inputData_LIST=None):
    """
    This routine creates a child process and wait for it to end or timeout

    @param      args                    Either a Tuple/List or str
                relativeTimeoutInSecs   When to give up and kill
                                        the child process
                inputData_LIST          List of input data

    @return    stdout_LIST, stderr_LIST, exitStatus, deltaTime
    """
    theChild = Process_OBJ(False, args,
                           relativeTimeoutInSecs, inputData_LIST)
    theChild.monitor_child_until_done()

    return theChild.get_child_results()


def exe_developer_test(args, relativeTimeoutInSecs=0, inputData_LIST=None):
    """
    This routine creates a child process and wait for it to end or timeout

    NOTE:   THIS IS FOR DEVELOPER TESTING AND EXAMPLE PURPOSES ONLY !!!
            Use exe_sync() instead

    @param      args                    Either a Tuple/List or str
                relativeTimeoutInSecs   When to give up and kill the
                                        child process
                inputData_LIST          List of input data

    @return     stdout_LIST, stderr_LIST, exitStatus, deltaTime
    """
    theChild = Process_OBJ(False, args, relativeTimeoutInSecs, inputData_LIST)
    while True:
        isDone, wasChildIOprogressMade = theChild.monitor_child_for_moment()
        if isDone:
            break
        if not wasChildIOprogressMade:
            time.sleep(0.1)

    return theChild.get_child_results()


def simple_process_checker(process_being_monitored, chkD):
    """
    This checker continuously executes the program passed in chkD.
    It continues executing as long as the process_being_monitored
    is still running. If the program passed in chkD takes longer
    than the timeout (specified as first element of the chkD) then
    the process_being_monitored and checker are both terminated.

    @param  process_being_monitored  object of Process_OBJ type. The
                                     main application that is being
                                     checked/monitored.
            chkD                     A list with the first element being
                                     an integer timeout and the rest the
                                     checker command to execute.
    """
    ckkrTimeout = int(chkD[0])
    chkrArgs = chkD[1]
    startTime = process_being_monitored.mStartTimeInSecs
    updateTime = process_being_monitored.mStartTimeInSecs
    deltaSinceStartTime = 0

    while True:
        # Check the process_being_monitored for completion
        (isDone,
         dummy_Childprog) = process_being_monitored.monitor_child_for_moment()
        if isDone:
            break
        # Start up a checker process
        checkerProcess = Process_OBJ(False, chkrArgs)
        while True:
            # Check the process_being_monitored for completion
            (isDone, dummy_Childprog) = process_being_monitored.monitor_child_for_moment()
            if isDone:
                checkerProcess.terminate_child(True)
                break
            # Check the checkerProcess for completion
            (isCheckerDone,
             dummy_Childprog) = checkerProcess.monitor_child_for_moment()
            # Check the process_being_monitored for completion
            (isDone, dummy_Childprog) = process_being_monitored.monitor_child_for_moment()
            if isDone:
                if not isCheckerDone:
                    checkerProcess.terminate_child(True)
                break
            deltaSinceStartTime = _INTERNAL_getdelta(startTime)
            if isCheckerDone:
                break
            if deltaSinceStartTime > ckkrTimeout:
                # Give up on both the checkerProcess and the
                # process_being_monitored
                checkerProcess.terminate_child(True)
                break
            deltaSinceUpdateTime = _INTERNAL_getdelta(updateTime)
            if deltaSinceUpdateTime > 2:
                # Provide user feedback
                deltaSinceUpdateTime = time.time()
        if isDone:
            break

        if deltaSinceStartTime > ckkrTimeout:
            # Give up on the process_being_monitored
            break

        deltaSinceUpdateTime = _INTERNAL_getdelta(updateTime)
        if deltaSinceUpdateTime > 2:  # Provide user feedback
            deltaSinceUpdateTime = time.time()

    if not isDone:
        # Kill the process_being_monitored and mark as timed out
        process_being_monitored.terminate_child(True)
        process_being_monitored.mDidTimeout = 1


def wait_for_process_end(pid, timeoutInSecs):
    if pid == 0:
        logging.error("Bad pid == 0")
        return errno.EINVAL
    proc_status_path = "/proc/%s/status" % (pid)
    if not os.path.exists(proc_status_path):
        return 0  # Already gone
    state_cmpldRegEx = re.compile('^State:\s(\w)\s\((\w+)\)$')
    for dummy_cntr in range(timeoutInSecs):
        proc_status_lines_LIST, read_failed = read_text_lines(proc_status_path)
        if read_failed:
            return 0  # Already gone
        for proc_status_line in proc_status_lines_LIST:
            state_match = state_cmpldRegEx.match(proc_status_line)
            if not state_match:
                continue
            if state_match.group(1) == 'Z':
                return 0  # Dead Zombie
        time.sleep(1.0)

    return errno.ETIMEDOUT


def terminate_process(pid, maximumKill):
    """
    This routine will terminate the PID if it is still running.
    """
    if pid == 0:
        return errno.EINVAL
    proc_status_path = "/proc/%s/status" % (pid)
    if not os.path.exists(proc_status_path):
        return 0
    try:
        os.kill(pid, signal.SIGHUP)
    except OSError:
        return 0

    if wait_for_process_end(pid, 30) == 0:
        return 0
    if not maximumKill:
        return errno.ETIMEDOUT
    try:
        os.kill(pid, signal.SIGKILL)
    except OSError:
        return 0

    if wait_for_process_end(pid, 30) == 0:
        return 0

    return errno.ETIMEDOUT


def exe_async(args, inputData_LIST=None):
    """
    This routine creates an async child process and hands it to the caller.

    @param      args                  Either a Tuple/List or str
                inputData_LIST        List of input data provided to
                                      the args application.

    @return     stdout_LIST, stderr_LIST, exitStatus, exitCode, deltaTime
    """
    return Process_OBJ(False, args, 0, inputData_LIST)


def exe_async_with_checker(args, checkerFnc, checkerData, inputData_LIST=None):
    """
    This routine creates an asynchronous process, hands it to the
    checkerFnc along with the chkD for it to monitor the process.

    the checkerFnc MUST accept two arguments:
            checkerFnc(Process_OBJ, checkerData)

    wherein:
            Process_OBJ      Process object
            checkerData      list/tuple consisting of arguments
                             necessary for the checker to function.

    For an example, take a look at 'simple_process_checker'
    """
    proc = exe_async(args, inputData_LIST)
    checkerFnc(proc, checkerData)
    return proc.get_child_results()


def pidof(commandNameToFind):
    """
    This routine finds all of the process ids for the given command name.

    @param         commandNameToFind    the command name

    @return        list of pids
    """
    pid_cmpldRegEx = re.compile('^\d+$')
    cln_cmpldRegEx = re.compile('\x00.*$')
    cmd_cmpldRegEx = re.compile("^%s$" % (commandNameToFind))
    pid_LIST = []
    for process_id in os.listdir('/proc'):
        if not pid_cmpldRegEx.match(process_id):
            continue
        process_id_cmdline = ''
        try:
            process_id_cmdline_path = "/proc/%s/cmdline" % (process_id)
            process_id_cmdline_FILE = open(process_id_cmdline_path, 'r')
            process_id_cmdline_LIST = process_id_cmdline_FILE.read().splitlines()
            if process_id_cmdline_LIST:
                process_id_cmdline = cln_cmpldRegEx.sub('',
                                                process_id_cmdline_LIST[0])
            process_id_cmdline_FILE.close()
        except StandardError:
            continue  # Likely that the process terminated

        if cmd_cmpldRegEx.match(process_id_cmdline):
            pid_LIST.append(int(process_id))

    return pid_LIST


def _INTERNAL_getdelta(startTime):
    """
    Get the delta of time between the passed in time and the current.
    """
    nowTime = time.time()
    if nowTime >= startTime:
        deltaTime = nowTime - startTime
    else:
        deltaTime = startTime - nowTime
    return deltaTime
